#include <string.h>
#include <stdio.h>

#define LBDA_MARK  (-2)
#define LBDA_FRAME (-1)
#define SUBR       1
#define LIST       2
#define SYM        3
#define INT        4
#define STR        5
#define FPTR       6
#define SET        7
#define REL        8
#define FF         9
#define FUN        10
#define TTUP   12
#define CHAN       13

struct lab_type {
    char            *label ;
    struct t_table  *type ;
    struct lab_type *next ;
} ;

struct type {
    struct t_table  *type ;
    struct type *next ;
} ;

union b {
    struct lab_type *tcomp ;    /* list of tuple label-component pairs */
    struct type *acomp ;        /* list of alternative components */
    struct t_table *type ;      /* set, list or equality */
    struct {                    /* relation or finite function */
        struct t_table *dom ;   /* domain type */
        struct t_table *ran ;   /* range type */
    } rel ;
} ;

struct t_table {
    char           *tname ;     /* type name */
    int            nnumber ;    /* type code */
    int            tnumber ;    /* definition code */
    union b        def ;        /* type definition */
    struct t_table *next ;
} ;


#define LABEL      258
#define TYPE       259
#define MACTYPE    260
/* types >= 1000 are reserved for user defined */
#define USERDEFTYPE    1000

/* string types */
#define DYNAMIC	0
#define STATIC	1


/* symbol structure */
struct xsym {
    char *xsy_name;		/* symbol name */
    struct node *xsy_value;	/* the current value */
};

/* subr node structure */
struct xsubr {
    struct node *(*xsu_subr)();	/* pointer to an internal routine */
};

/* list node structure */
struct xlist {
    struct node *xl_value;	/* value at this node */
    struct node *xl_next;	/* next node */
};

/* integer node structure */
struct xint {
    int xi_int;			/* integer value */
};

/* string node structure */
struct xstr {
    int xst_type;		/* string type */
    char *xst_str;		/* string pointer */
};

/* file pointer node structure */
struct xfptr {
    FILE *xf_fp;		/* the file pointer */
};

struct xchan {
    struct node  *q_name ;
    int          qid ;
};
/* TYPE structure */
struct xtype {
    struct t_table *n_tp ;
} ;

/* symbol node */
#define n_symname	n_info.n_xsym.xsy_name
#define n_symvalue	n_info.n_xsym.xsy_value

/* subr node */
#define n_subr		n_info.n_xsubr.xsu_subr

/* list (and binding and REC) node */
#define n_head	n_info.n_xlist.xl_value
#define n_tail	n_info.n_xlist.xl_next
#define n_listvalue	n_info.n_xlist.xl_value
#define n_listnext	n_info.n_xlist.xl_next
#define n_bndsym	n_info.n_xlist.xl_value
#define n_bndvalue	n_info.n_xlist.xl_next
#define n_left		n_info.n_xlist.xl_value
#define n_right		n_info.n_xlist.xl_next
#define n_ptr		n_info.n_xlist.xl_value
#ifdef LAZY
#define n_recexp        n_info.n_xlist.xl_value
#define n_recenv        n_info.n_xlist.xl_next
#endif

/* integer node */
#define n_int		n_info.n_xint.xi_int

/* string node */
#define n_str		n_info.n_xstr.xst_str
#define n_strtype	n_info.n_xstr.xst_type

/* file pointer node */
#define n_fp		n_info.n_xfptr.xf_fp

#ifdef COMM
/* CHAN node    */
#define n_qname        n_info.n_xchan.q_name
#define n_qid          n_info.n_xchan.qid
#endif

#define n_typedef      n_info.n_xtype.n_tp
#define n_macdef       n_info.n_xsym.xsy_value

/* node structure */
struct node {
    int  n_type ;		/* type of node */
    char n_flags ;		/* flag bits */
    union {			/* value */
	struct xsym n_xsym;	/*     symbol node */
	struct xsubr n_xsubr;	/*     subr node */
	struct xlist n_xlist;	/*     list node */
	struct xint n_xint;	/*     integer node */
	struct xstr n_xstr;	/*     string node */
	struct xfptr n_xfptr;	/*     file pointer node */
	struct xchan n_xchan;	/*     chan node */
        struct xtype n_xtype;  /*     type node */
    } n_info ;
} ;

typedef struct node * SEXP;

/* external procedure declarations */
extern SEXP xlread();		/* read an expression */
extern SEXP xleval();		/* evaluate an expression */
#ifdef LAZY
extern SEXP delay();            /* recipe an expression */
#endif
extern SEXP xlarg();		/* fetch an argument */
extern SEXP xlevarg();		/* fetch and evaluate an argument */
extern SEXP xljustevarg();	/* fetch and evaluate an argument */
extern SEXP userevarg();        /* xlevarg() using labeval() */
extern SEXP labeval();          /* xleval() ignoring
                                           prefix type labels */
extern SEXP xlmatch();		/* fetch an typed argument */
extern SEXP xlevmatch();	/* fetch and evaluate a typed arg */
extern void        xllastarg();         /* ensures there are no more args */
extern SEXP xlenter();		/* enter a symbol into the oblist */
extern void        xlsubr();		/* enter a builtin function */
extern SEXP xlsave();		/* generate a stack frame */
extern SEXP newnode();		/* allocate a new node */
extern char        *xmstralloc();		/* allocate string space */
extern char        *strsave();		/* make a safe copy of a string */
/*extern void free() ;                    /* frees space allocated by strsave() */
extern void        xlfail();            /* error handling routine */

/* ---------------------------------------------------------------------*/


#define get1(l) getele(1,l)
#define get2(l) getele(2,l)
#define get3(l) getele(3,l)
#define get4(l) getele(4,l)
#define get5(l) getele(5,l)
#define get6(l) getele(6,l)
#define forall(nptr,x,l) if(l->n_type == LIST || \
                            l->n_type == SET  ) \
   for(nptr=l,x = get1(l); nptr != NULL; nptr = nptr->n_tail,x=get1(nptr))

#define forallpair(nptr,x,y,l) if( l->n_type == FF || \
                                   l->n_type == REL  )\
   for(nptr=l,x = get1(l),y=get2(x),x=get1(x); nptr != NULL; nptr = nptr->n_listnext,x=get1(nptr),y=get2(x),x=get1(x))


#define bool(S) ((S->n_type == SYM) && strcmp(S->n_symname,"true")==0)
#define f0(name) l1sexp(xlenter(name)) 
#define f1(name,s) l2sexp(xlenter(name),s) 
#define f2(name,s1,s2) l3sexp(xlenter(name),s1,s2) 
#define f3(name,s1,s2,s3) l4sexp(xlenter(name),s1,s2,s3) 
#define f4(name,s1,s2,s3,s4) l5sexp(xlenter(name),s1,s2,s3,s4) 
#define symsexp(name) xlenter(name)
 
#define evf0(name)  libxmeval(f0(name))
#define evf1(name,s)  libxmeval(f1(name,s))
#define evf2(name,s1,s2)  libxmeval(f2(name,s1,s2))
#define evf3(name,s1,s2,s3)  libxmeval(f3(name,s1,s2,s3))
#define evf4(name,s1,s2,s3,s4)  libxmeval(f4(name,s1,s2,s3,s4))

extern SEXP getele();
extern SEXP l1sexp();  /* faz uma lista de 1 SEXP */
extern SEXP l2sexp();  /* faz uma lista de 2 SEXP */
extern SEXP l3sexp();  /* faz uma lista de 3 SEXP */
extern SEXP l4sexp();  /* faz uma lista de 4 SEXP */
extern SEXP conssexp();  /* faz uma lista de cons(Sexp , Sexp-list) */
extern SEXP intsexp();  /* faz uma SEXP com um inteiro */
extern SEXP fpsexp();  /* faz uma SEXP com um FILE * */
extern SEXP strsexp();  /* faz uma SEXP com uma string */
extern SEXP libxmeval();
extern SEXP libxmevalstr();
extern SEXP arg_n();
extern SEXP arg_n_opt();
extern int arg_number();

#define XM_ERROR ((SEXP) -1)
