/* xmbool.c - boolean algebra */

#include <stdio.h>
#include "xmetoo.h"


/* external variables */
extern struct node *xlstack ;
extern struct node *xlenv ;


/* external procedures */
extern int strcmp() ;
extern char *strcat() ;
extern void xlbind() ;
extern void xlunbind() ;
extern void xlfixbindings() ;
extern void subst_name() ;


/* global variables */
struct node *true ;
struct node *false ;


/*. true - metoo symbol .*/
/* (implemented by symbol "true") */


/*. false - metoo symbol .*/
/* (implemented by symbol "false") */


/*. and - builtin function and (implements "andif") .*/
/*. and : expbool expbool -> bool 
.*/
static struct node *and( args ) 
struct node *args ;
{
   struct node arg ;
   struct node *oldstk, *val1, *val2 ;

   oldstk = xlsave( &arg, NULL ) ;

   arg.n_ptr = args ;
   val1 = xlevarg( &arg.n_ptr ) ;
   val2 = xlarg( &arg.n_ptr ) ;
   xllastarg( arg.n_ptr ) ;

   if( (val1 != false) && (val1 != true) ) xlfail( "bad argument type" ) ;

   if( val1 == false)
      val2 = false ;
    else {
      val2 = xleval( val2 ) ;
      if( (val2 != false) && (val2 != true) ) xlfail( "bad argument type" ) ; }

   xlstack = oldstk ;

   return( val2 ) ;
}


/*. or - builtin function or (implements "orif") .*/
/*. or : expbool expbool -> bool 
.*/
static struct node *or( args )
struct node *args ;
{
   struct node arg ;
   struct node *oldstk, *val1, *val2 ;

   oldstk = xlsave( &arg, NULL ) ;

   arg.n_ptr = args ;
   val1 = xlevarg( &arg.n_ptr ) ;
   val2 = xlarg( &arg.n_ptr ) ;
   xllastarg( arg.n_ptr ) ;

   if( (val1 != false) && (val1 != true) ) xlfail( "bad argument type" ) ;

   if( val1 == true)
      val2 = true ;
    else {
      val2 = xleval( val2 ) ;
      if( (val2 != false) && (val2 != true) ) xlfail( "bad argument type" ) ; }

   xlstack = oldstk ;

   return( val2 ) ;
}
 

/*. not - builtin function not .*/
/*. not : expbool -> bool .*/
static struct node *not( args )
struct node *args ;
{
   struct node *val1 ;

   val1 = xlevarg( &args ) ;
   xllastarg( args ) ;

   if( (val1 != false) && (val1 != true) ) xlfail( "bad argument type" ) ;

   return( (val1 == false) ? true : false ) ;
}

 
/*. eq - builtin function eq .*/
/*. eq : exp exp -> bool .*/

/* xeq - internal eq function */
static int xeq( arg1, arg2 )
struct node *arg1, *arg2 ;
{
    /* compare the arguments */
    if( (arg1 == NULL) && (arg2 == NULL) ) return( TRUE ) ;
    if( (arg1 == NULL) || (arg2 == NULL) ) return( FALSE ) ;
    if( (arg1->n_type == INT) && (arg2->n_type == INT) )
        return( arg1->n_int == arg2->n_int ) ;
    if( (arg1->n_type == STR) && (arg2->n_type == STR) )
        return( strcmp( arg1->n_str, arg2->n_str ) == 0 ) ;
    if( (arg1->n_type == FPTR) && (arg2->n_type == FPTR) )
        return( arg1->n_fp == arg2->n_fp ) ;
#ifdef COMM
    if( (arg1->n_type == CHAN) && (arg2->n_type == CHAN) )
        return( arg1->n_qid == arg2->n_qid ) ;
#endif
    if( (arg1->n_type == SUBR) && (arg2->n_type == SUBR) )
        return( arg1->n_subr == arg2->n_subr ) ;
    return( arg1 == arg2 ) ;
}

static struct node *eq( args )
struct node *args ;
{
    struct node arg, arg1, arg2 ;
    struct node *oldstk, *val ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &arg1, &arg2, NULL ) ;

    /* initialize */
    arg.n_ptr = args ;
    arg1.n_ptr = xlevarg( &arg.n_ptr ) ;
    arg2.n_ptr = xlevarg( &arg.n_ptr ) ;
    xllastarg( arg.n_ptr ) ;

    /* compare the arguments */
    val = (xeq( arg1.n_ptr, arg2.n_ptr )) ? true : false ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return the result */
    return( val ) ;
}


/*. lequal - compares two non empty lists .*/
static int lequal( l1, l2 )
struct node *l1, *l2 ;
{
    int xequal() ;
    int ok ;

    for( ok = TRUE ; ok && (l1 != NULL) && (l2 != NULL) ; 
                     l1 = l1->n_listnext, l2 = l2->n_listnext )
        ok = xequal( l1->n_listvalue, l2->n_listvalue ) ;

    return( ok && (l1 == NULL) && (l2 == NULL) ) ;
}


/*. sequal - compares two non empty sets .*/
static int sequal( s1, s2 )
struct node *s1, *s2 ;
{
    int xequal() ;
    struct node *aux1, *aux2 ; 
    int marked, found, el, ty ;

    ty = s2->n_type ;

    for( aux1 = s1, el = TRUE ; el && (aux1 != NULL ) ; 
                                      aux1 = aux1->n_listnext ) {
        for( aux2 = s2, found = FALSE ; !found && (aux2 != NULL) ;
                                        aux2 = aux2->n_listnext ) 
            if( (aux2->n_type != LBDA_MARK) && 
                      xequal( aux2->n_listvalue, aux1->n_listvalue ) ) {
                found = TRUE ;
                aux2->n_type = LBDA_MARK ; }
        el = found ; }

    for( marked = TRUE, aux2 = s2 ; aux2 != NULL ; aux2 = aux2->n_listnext ) {
        marked = marked && (aux2->n_type == LBDA_MARK) ;
        aux2->n_type = LIST ; }

    s2->n_type = ty ;

    return( el && marked ) ;
}


/*. xequal - internal equal function .*/
int xequal( arg1, arg2 )
struct node *arg1, *arg2 ;
{
    /* compare the arguments */
    if( xeq( arg1, arg2 ) )
        return( TRUE ) ;
    if( ((arg1 == NULL) && (arg2 != NULL)) || 
             ((arg1 != NULL) && (arg2 == NULL)) )
        return( FALSE ) ;
    if( (arg1->n_type == LIST) && (arg2->n_type == LIST) )
        return( lequal( arg1, arg2 ) ) ;
    if( ((arg1->n_type == SET) && (arg2->n_type == SET)) || 
             ((arg1->n_type == FF) && (arg2->n_type == FF)) ||
             ((arg1->n_type == REL) && (arg2->n_type == REL)) )
        return( sequal( arg1, arg2 ) ) ;
    if( (arg1->n_type == FUN) && (arg2->n_type == FUN) )
        return( xequal( arg1->n_listvalue, arg2->n_listvalue ) &&
		xequal( arg1->n_listnext, arg2->n_listnext ) ) ;
#ifdef USERTYPE
    if( (arg1->n_type >= USERDEFTYPE) && (arg1->n_type == arg2->n_type) )
        return( xequal( arg1->n_symvalue, arg2->n_symvalue ) ) ;
#endif
    return( FALSE ) ;
}


/*. equal - builtin function equal .*/
/*. equal : exp exp -> bool .*/
static struct node *equal( args )
struct node *args ;
{
    struct node arg, arg1, arg2 ;
    struct node *oldstk, *val ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &arg1, &arg2, NULL ) ;

    /* initialize */
    arg.n_ptr = args ;
    arg1.n_ptr = xlevarg( &arg.n_ptr ) ;
    arg2.n_ptr = xlevarg( &arg.n_ptr ) ;
    xllastarg( arg.n_ptr ) ;

    /* compare the arguments */
    val = (xequal( arg1.n_ptr, arg2.n_ptr )) ? true : false ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return the result */
    return( val ) ;
}


/*. lt - builtin function lt .*/
/*. lt : expint expint -> bool .*/
/*. lt : expstr expstr -> bool .*/
static struct node *lt( args ) 
struct node *args ;
{
    struct node arg, val1, val2 ;
    struct node *oldstk ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &val1, &val2, NULL ) ;

    arg.n_ptr = args ;
    val1.n_ptr = xlevarg( &arg.n_ptr ) ;
    if( (val1.n_ptr == NULL) || ((val1.n_ptr->n_type != INT) && 
                                 (val1.n_ptr->n_type != STR)) )
        xlfail( "bad argument type" ) ;
    val2.n_ptr = xlevarg( &arg.n_ptr ) ;
    if( (val2.n_ptr == NULL) || (val2.n_ptr->n_type != val1.n_ptr->n_type) )
        xlfail( "bad argument type" ) ;
    xllastarg( arg.n_ptr ) ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return pointer to the result */
    if( val1.n_ptr->n_type == INT )
        return( (val1.n_ptr->n_int < val2.n_ptr->n_int) ? true : false ) ;
    return( (strcmp( val1.n_ptr->n_str, val2.n_ptr->n_str ) < 0) ?
                                                          true : false ) ;
}
 

/*. geq - builtin function geq .*/
/*. geq : expint expint -> bool .*/
/*. geq : expstr expstr -> bool .*/
static struct node *geq( args ) 
struct node *args ;
{
    struct node arg, val1, val2 ;
    struct node *oldstk ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &val1, &val2, NULL ) ;

    arg.n_ptr = args ;
    val1.n_ptr = xlevarg( &arg.n_ptr ) ;
    if( (val1.n_ptr == NULL) || ((val1.n_ptr->n_type != INT) && 
                                 (val1.n_ptr->n_type != STR)) )
        xlfail( "bad argument type" ) ;
    val2.n_ptr = xlevarg( &arg.n_ptr ) ;
    if( (val2.n_ptr == NULL) || (val2.n_ptr->n_type != val1.n_ptr->n_type) )
        xlfail( "bad argument type" ) ;
    xllastarg( arg.n_ptr ) ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return pointer to the result */
    if( val1.n_ptr->n_type == INT )
        return( (val1.n_ptr->n_int >= val2.n_ptr->n_int) ? true : false ) ;
    return( (strcmp( val1.n_ptr->n_str, val2.n_ptr->n_str ) >= 0) ?
                                                           true : false ) ;
}

 
/*. gt - builtin function gt .*/
/*. gt : expint expint -> bool .*/
/*. gt : expstr expstr -> bool .*/
static struct node *gt( args ) 
struct node *args ;
{
    struct node arg, val1, val2 ;
    struct node *oldstk ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &val1, &val2, NULL ) ;

    arg.n_ptr = args ;
    val1.n_ptr = xlevarg( &arg.n_ptr ) ;
    if( (val1.n_ptr == NULL) || ((val1.n_ptr->n_type != INT) && 
                                 (val1.n_ptr->n_type != STR)) )
        xlfail( "bad argument type" ) ;
    val2.n_ptr = xlevarg( &arg.n_ptr ) ;
    if( (val2.n_ptr == NULL) || (val2.n_ptr->n_type != val1.n_ptr->n_type) )
        xlfail( "bad argument type" ) ;
    xllastarg( arg.n_ptr ) ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return pointer to the result */
    if( val1.n_ptr->n_type == INT )
        return( (val1.n_ptr->n_int > val2.n_ptr->n_int) ? true : false ) ;
    return( (strcmp( val1.n_ptr->n_str, val2.n_ptr->n_str ) > 0) ?
                                                           true : false ) ;
}

 
/*. leq - builtin function leq .*/
/*. leq : expint expint -> bool .*/
/*. leq : expstr expstr -> bool .*/
static struct node *leq( args ) 
struct node *args ;
{
    struct node arg, val1, val2 ;
    struct node *oldstk ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &val1, &val2, NULL ) ;

    arg.n_ptr = args ;
    val1.n_ptr = xlevarg( &arg.n_ptr ) ;
    if( (val1.n_ptr == NULL) || ((val1.n_ptr->n_type != INT) && 
                                 (val1.n_ptr->n_type != STR)) )
        xlfail( "bad argument type" ) ;
    val2.n_ptr = xlevarg( &arg.n_ptr ) ;
    if( (val2.n_ptr == NULL) || (val2.n_ptr->n_type != val1.n_ptr->n_type) )
        xlfail( "bad argument type" ) ;
    xllastarg( arg.n_ptr ) ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return pointer to the result */
    if( val1.n_ptr->n_type == INT )
        return( (val1.n_ptr->n_int <= val2.n_ptr->n_int) ? true : false ) ;
    return( (strcmp( val1.n_ptr->n_str, val2.n_ptr->n_str ) <= 0) ? 
	    true : false ) ;
}
 

/*. neq - builtin function neq .*/
/*. neq : exp exp --> bool .*/
static struct node *neq( args )
struct node *args ;
{
    return( (eq( args ) == true) ? false : true ) ;
}


/*. nequal - builtin function nequal .*/
/*. nequal : exp exp --> bool .*/
static struct node *nequal( args )
struct node *args ;
{
    return( (equal( args ) == true) ? false : true ) ;
}


/*. fif - builtin function if .*/
/*. if : expbool exp [exp] -> value .*/
static struct node *fif( args )
struct node *args ;
{
    struct node arg ;
    struct node *oldstk , *testexp, *thenexp, *elseexp, *val ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, NULL ) ;

    arg.n_ptr = args ;
    testexp = xlevarg( &arg.n_ptr ) ;
    if( (testexp != false) && (testexp != true) )
            xlfail( "bad argument type" ) ;
    thenexp = xlarg( &arg.n_ptr ) ;
#ifdef USERTYPE
    elseexp = (arg.n_ptr == NULL) ? xlenter("NIL") : xlarg( &arg.n_ptr ) ;
#else 
    elseexp = (arg.n_ptr == NULL) ? NULL : xlarg( &arg.n_ptr ) ;
#endif
    xllastarg( arg.n_ptr ) ;

    if( testexp == true )
	val = xleval( thenexp ) ;
    else
	val = xleval( elseexp ) ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return pointer to the result expression */
    return( val ) ;
}


/*. cond - builtin function cond .*/
/*. cond pair-list -> val .*/
/*.      pair = "(" expbool exp... ")" .*/
static struct node *cond( args )
struct node *args ;
{
    struct node arg, list ;
    struct node *oldstk, *val, *bool ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &list, NULL ) ;

    /* initialize */
    arg.n_ptr = args ;

    /* initialize the return value */
    val = xlenter("NIL") ;
    bool = false ;

    /* find a predicate that is true */
    while( (bool == false) && (arg.n_ptr != NULL) ) {
        /* get the next conditional */
        list.n_ptr = xlmatch( LIST, &arg.n_ptr ) ;
        /* evaluate the predicate part */
        bool = xlevarg( &list.n_ptr ) ;
        if( (bool != true) && (bool != false) )
            xlfail( "invalid boolean expression" ) ; }

    if( bool == true )
        /* evaluate each expression */
        while( list.n_ptr != NULL )
            val = xlevarg( &list.n_ptr ) ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return the value */
    return( val ) ;
}


/*. atom - builtin function atom .*/
/*. atom : exp -> bool .*/
static struct node *atom( args )
struct node *args ;
{
    struct node *arg ;

    /* get the argument */
    arg = xlevarg( &args ) ;
    xllastarg( args ) ;

    if( (arg == NULL) || (arg->n_type == SUBR) || (arg->n_type == SYM) ||
        (arg->n_type == INT) || (arg->n_type == STR) ||
#ifdef COMM
        (arg->n_type == CHAN) ||
#endif
        (arg->n_type == FPTR) )
        return( true ) ;
    return( false ) ;
}


/*. ALL - builtin function ALL .*/
/*. ALL : sym expset expbool --> bool .*/
static struct node *ALL( args )
struct node *args ;
{
    struct node arg, set, par_list ;
    struct node *oldstk, *oldenv, *sym, *exp, *res ;
    char name[STRMAX+1] ;

    oldstk = xlsave( &arg, &set, &par_list, NULL ) ;
    arg.n_ptr = args ;

    sym = xlmatch( SYM, &args ) ;
    /*    set.n_ptr = xlevmatch( SET, &args ) ;*/

    set.n_ptr = xlevarg(&args);

#ifdef USERTYPE
    while( (set.n_ptr != NULL) && (set.n_ptr->n_type >= USERDEFTYPE) )
        set.n_ptr = set.n_ptr->n_symvalue ;
#endif /* USERTYPE jj */
 if( (set.n_ptr != NULL) &&
     (set.n_ptr->n_type != SET) &&
     (set.n_ptr->n_type != REL) &&
     (set.n_ptr->n_type != FF) &&
     (set.n_ptr->n_type != LIST) )
        xlfail( "bad argument type" ) ;


    exp = xlarg( &args ) ;
    xllastarg( args ) ;

    if( *(sym->n_symname) != ' ' ) {
        par_list.n_ptr = newnode( LIST ) ;
        par_list.n_ptr->n_listvalue = sym ;
        *name = ' ' ; *(name+1) = '\0' ;
        strcat( name, sym->n_symname ) ;
        sym = xlenter( name ) ;
        subst_name( &par_list.n_ptr, &exp ) ; }

    for( res = true ; (res == true) && (set.n_ptr != NULL) ;
                                           set.n_ptr = set.n_ptr->n_listnext) {
        oldenv = xlenv ;
        xlbind( sym, set.n_ptr->n_listvalue ) ;
        xlfixbindings( oldenv ) ;
        res = xleval( exp ) ;
        xlunbind( oldenv ) ; }

    if( (res != true) && (res != false) ) xlfail( "bad argument type" ) ;

    xlstack = oldstk ;

    return( res ) ;
}


/*. EXIST - builtin function EXIST .*/
/*. EXIST : sym expset expbool --> bool .*/
static struct node *EXIST( args )
struct node *args ;
{
    struct node arg, set, par_list ;
    struct node *oldstk, *oldenv, *sym, *exp, *res ;
    char name[STRMAX+1] ;

    oldstk = xlsave( &arg, &set, &par_list, NULL ) ;
    arg.n_ptr = args ;

    sym = xlmatch( SYM, &args ) ;
    set.n_ptr = xlevarg(&args);

#ifdef USERTYPE
    while( (set.n_ptr != NULL) && (set.n_ptr->n_type >= USERDEFTYPE) )
        set.n_ptr = set.n_ptr->n_symvalue ;
#endif /* USERTYPE jj */
 if( (set.n_ptr != NULL) &&
     (set.n_ptr->n_type != SET) &&
     (set.n_ptr->n_type != REL) &&
     (set.n_ptr->n_type != FF) &&
     (set.n_ptr->n_type != LIST) )
        xlfail( "bad argument type" ) ;

/*    set.n_ptr = xlevmatch( SET, &args ) ; */
    exp = xlarg( &args ) ;
    xllastarg( args ) ;

    if( *(sym->n_symname) != ' ' ) {
        par_list.n_ptr = newnode( LIST ) ;
        par_list.n_ptr->n_listvalue = sym ;
        *name = ' ' ; *(name+1) = '\0' ;
        strcat( name, sym->n_symname ) ;
        sym = xlenter( name ) ;
        subst_name( &par_list.n_ptr, &exp ) ; }

    for( res = false ; (res == false) && (set.n_ptr != NULL) ;
                                           set.n_ptr = set.n_ptr->n_listnext) {
        oldenv = xlenv ;
        xlbind( sym, set.n_ptr->n_listvalue ) ;
        xlfixbindings( oldenv ) ;
        res = xleval( exp ) ;
        xlunbind( oldenv ) ; }

    if( (res != true) && (res != false) ) xlfail( "bad argument type" ) ;

    xlstack = oldstk ;

    return( res ) ;
}


/*. EXIST1 - builtin function EXIST1 .*/
/*. EXIST1 : sym expset expbool --> bool .*/
static struct node *EXIST1( args )
struct node *args ;
{
    struct node arg, set, par_list ;
    struct node *oldstk, *oldenv, *sym, *exp, *res ;
    int rep ;
    char name[STRMAX+1] ;

    oldstk = xlsave( &arg, &set, &par_list, NULL ) ;
    arg.n_ptr = args ;

    sym = xlmatch( SYM, &args ) ;
    set.n_ptr = xlevarg(&args);

#ifdef USERTYPE
    while( (set.n_ptr != NULL) && (set.n_ptr->n_type >= USERDEFTYPE) )
        set.n_ptr = set.n_ptr->n_symvalue ;
#endif /* USERTYPE jj */
 if( (set.n_ptr != NULL) &&
     (set.n_ptr->n_type != SET) &&
     (set.n_ptr->n_type != REL) &&
     (set.n_ptr->n_type != FF) &&
     (set.n_ptr->n_type != LIST) )
        xlfail( "bad argument type" ) ;

    /* set.n_ptr = xlevmatch( SET, &args ) ; */
    exp = xlarg( &args ) ;
    xllastarg( args ) ;

    if( *(sym->n_symname) != ' ' ) {
        par_list.n_ptr = newnode( LIST ) ;
        par_list.n_ptr->n_listvalue = sym ;
        *name = ' ' ; *(name+1) = '\0' ;
        strcat( name, sym->n_symname ) ;
        sym = xlenter( name ) ;
        subst_name( &par_list.n_ptr, &exp ) ; }

    for( rep = 0 ; (rep < 2) && (set.n_ptr != NULL) ;
                                           set.n_ptr = set.n_ptr->n_listnext) {
        oldenv = xlenv ;
        xlbind( sym, set.n_ptr->n_listvalue ) ;
        xlfixbindings( oldenv ) ;
        res = xleval( exp ) ;
        if( (res != true) && (res != false) ) xlfail( "bad argument type" ) ;
        xlunbind( oldenv ) ;
        if( res == true )
            rep++ ; }

    xlstack = oldstk ;

    return( (rep == 1) ? true : false ) ;
}


void xmboolinit() 
{
    true = xlenter( "true" ) ;
    true->n_symvalue = true ;
    false = xlenter( "false" ) ;
    false->n_symvalue = false ;

    xlsubr( "and", and ) ;
    xlsubr( "or", or ) ;
    xlsubr( "not", not ) ;
    xlsubr( "eq", eq ) ;
    xlsubr( "neq", neq ) ;
    xlsubr( "equal", equal ) ;
    xlsubr( "nequal", nequal ) ;
    xlsubr( "leq", leq ) ;
    xlsubr( "lt", lt ) ;
    xlsubr( "geq", geq ) ;
    xlsubr( "gt", gt ) ;
    xlsubr( "ALL", ALL ) ;
    xlsubr( "EXIST", EXIST ) ;
    xlsubr( "EXIST1", EXIST1 ) ;
    xlsubr( "atom", atom ) ;
    xlsubr( "if", fif ) ;
    xlsubr( "cond", cond ) ;
}
