/* xmdef.c - shell and state manipulation routines */

#include <stdio.h>
#include "xmetoo.h"


/* external variables */
extern struct node *xlenv ;
extern struct node *xlstack ;
extern struct node *oblist ;
extern struct node *oblimit ;
extern struct node lbda_env ;

/* external procedures */
extern char *malloc() ;
extern char *strcpy() ;
extern char *strcat() ;
extern int  strcmp() ;
extern int  xequal() ;
extern void xlbind() ;
extern void xlunbind() ;
extern void xlfixbindings() ;


/* global variables */
struct node *ops ;
struct node *obs ;

/*. reserv_symb - test if n is a reserved symbol .*/
int reserv_symb( n )
char *n ;
{
    struct node *p ;
    int found ;

    for( p = oblimit, found = FALSE ; (p != NULL) && !found ;
                                                 p = p->n_listnext )
        found = (strcmp( n, p->n_listvalue->n_symname ) == 0) ;
    return( found ) ;
}

#ifdef USERTYPE

/*. m2rem - remove a symbol from symbol and operation tables .*/
void m2rem( sname )
char *sname ;
{
    struct node *sptr, *prev ;
    int found ;
    
    /* check for symbol in the operation table */
    for( found = FALSE, prev = NULL, sptr = ops->n_symvalue ;
         !found && (sptr != NULL) ; prev = sptr, sptr = sptr->n_listnext )
	if( strcmp( sptr->n_listvalue->n_symname, sname ) == 0 ) {
            found = TRUE ;
            if( prev == NULL )
                ops->n_symvalue = sptr->n_listnext ;
            else
                prev->n_listnext = sptr->n_listnext ; }
    /* if not found check for symbol in the symbol table */
    for( prev = NULL, sptr = obs->n_symvalue ;
         !found && (sptr != NULL) ; prev = sptr, sptr = sptr->n_listnext )
	if( strcmp( sptr->n_listvalue->n_symname, sname ) == 0 ) {
            found = TRUE ;
            if( prev == NULL )
                obs->n_symvalue = sptr->n_listnext ;
            else
                prev->n_listnext = sptr->n_listnext ; }
}

#endif


/*. m2enter - enter a symbol into the METOO  "l1" symbol/operation table .*/
/*.           removing it, if necessary, from "l2" operation/symbol table .*/
static void m2enter( sname, l1, l2 )
char *sname ;
struct node *l1, *l2 ;
{
    struct node *sptr, *prev ;
    int found ;
    
    /* check for nil */
    if( strcmp( sname, "nil" ) == 0 )
	return ;

    /* check for symbol in the other table */
    for( found = FALSE, prev = NULL, sptr = l2->n_symvalue ;
         !found && (sptr != NULL) ; prev = sptr, sptr = sptr->n_listnext )
	if( strcmp( sptr->n_listvalue->n_symname, sname ) == 0 ) {
            found = TRUE ;
            if( prev == NULL )
                l2->n_symvalue = sptr->n_listnext ;
            else
                prev->n_listnext = sptr->n_listnext ; }

    /* check for symbol already in table */
    for( sptr = l1->n_symvalue ; sptr != NULL ; sptr = sptr->n_listnext )
	if( strcmp( sptr->n_listvalue->n_symname, sname ) == 0 )
	    return ;

    /* enter a new symbol and link it into the symbol/operation list */
    sptr = newnode( LIST ) ;
    sptr->n_listnext = l1->n_symvalue ;
    l1->n_symvalue = sptr ;
    sptr->n_listvalue = xlenter( sname ) ;
}


/*. subst_name - change formal parameter names .*/
void subst_name( par_list, exp ) 
struct node **par_list, **exp ; 
{
    struct node *p, *q, *old_par, *new, *val, *sym ;
    int found, sw ;
    char name[STRMAX+1] ;
   
    if( *exp == NULL ) return ;

    switch( (*exp)->n_type ) {
        
    case SYM : 
        for( p = *par_list, found = FALSE ; (p != NULL) && !found ; 
                                                       p = p->n_listnext )
            if( strcmp( (*exp)->n_symname, p->n_listvalue->n_symname ) == 0 ) {
                found = TRUE ;
                strcpy( name, " " ) ;
                strcat( name, (*exp)->n_symname ) ;
                *exp = xlenter( name ) ; }
        break ;

    case LIST :
        if( ((p = (*exp)->n_listvalue) == NULL) || (p->n_type != SYM) )
            for( p = *exp ; p != NULL ; p = p->n_listnext )
                subst_name( par_list, &p->n_listvalue ) ;
        else {
            if( strcmp( p->n_symname, "let" ) == 0 ) 
                sw = 1 ;
            else if( strcmp( p->n_symname, "lambda" ) == 0 )
                sw = 2 ;
            else if( (strcmp( p->n_symname, "ALL" ) == 0) ||
                     (strcmp( p->n_symname, "EXIST" ) == 0) ||
                     (strcmp( p->n_symname, "EXIST1" ) == 0) )
                sw = 3 ;
            else if( strcmp( p->n_symname, "ff" ) == 0 )
                sw = 4 ;
            else if( strcmp( p->n_symname, "foreach" ) == 0 )
                sw = 5 ;
            else if( (strcmp( p->n_symname, "set" ) == 0) ||
                     (strcmp( p->n_symname, "seq" ) == 0) ||
                     (strcmp( p->n_symname, "ff1" ) == 0) )
                sw = 6 ;
            else
                sw = 0 ;

            switch( sw ) {
            case 0 :
                for( p = *exp ; p != NULL ; p = p->n_listnext )
                    subst_name( par_list, &p->n_listvalue ) ;
                break ;
            case 1 :
                if( (*exp)->n_listnext == NULL ) xlfail( "invalid argument" ) ;
                old_par = *par_list ;
		for( p = (*exp)->n_listnext->n_listvalue ; p != NULL ;
                                                      p = p->n_listnext ) {
                    if( (p->n_listvalue == NULL) || 
                        (p->n_listvalue->n_type != LIST) || 
                        ((q = p->n_listvalue->n_listvalue) == NULL) ||
                        (q->n_type != SYM) ||
                        (p->n_listvalue->n_listnext == NULL) ||
                        (p->n_listvalue->n_listnext->n_listnext != NULL) )
                        xlfail( "invalid argument" ) ;

                    new = newnode( LIST ) ;
                    new->n_listnext = *par_list ;
                    *par_list = new ;
                    new->n_listvalue = q ;

                    *name = ' ' ; *(name+1) = '\0' ;
                    strcat( name, q->n_symname ) ;    
                    p->n_listvalue->n_listvalue = xlenter( name ) ;
                    subst_name( par_list, 
                                &(p->n_listvalue->n_listnext->n_listvalue) ) ; }
                    
                p = (*exp)->n_listnext->n_listnext ; 
                while( p != NULL ) {
                   subst_name( par_list, &(p->n_listvalue) ) ;
                   p = p->n_listnext ; }
                *par_list = old_par ;
                break ;
            case 2 :
                if( ((*exp)->n_listnext == NULL) ||
                    (((*exp)->n_listnext->n_listvalue != NULL) &&
                    ((*exp)->n_listnext->n_listvalue->n_type != LIST)) ||
                    ((*exp)->n_listnext->n_listnext == NULL) ||
                    ((*exp)->n_listnext->n_listnext->n_listnext != NULL) )
                    xlfail( "invalid argument" ) ;
                old_par = *par_list ;
		for( p = (*exp)->n_listnext->n_listvalue ; p != NULL ;
                                                      p = p->n_listnext ) {
                    if( ((q = p->n_listvalue) == NULL) || (q->n_type != SYM) )
                        xlfail( "invalid argument" ) ;

                    new = newnode( LIST ) ;
                    new->n_listnext = *par_list ;
                    *par_list = new ;
                    new->n_listvalue = q ;
 
                    *name = ' ' ; *(name+1) = '\0' ;
                    strcat( name, q->n_symname ) ;    
                    p->n_listvalue = xlenter( name ) ; }
                p = (*exp)->n_listnext->n_listnext ; 
                subst_name( par_list, &(p->n_listvalue) ) ;
                *par_list = old_par ;
                break ;
            case 3 :
                if( ((*exp)->n_listnext == NULL) ||
                    ((*exp)->n_listnext->n_listvalue == NULL) ||
                    ((*exp)->n_listnext->n_listvalue->n_type != SYM) ||
                    ((p = (*exp)->n_listnext->n_listnext) == NULL) ||
                    (p->n_listnext == NULL) ||
                    (p->n_listnext->n_listnext != NULL) )
                    xlfail( "invalid argument" ) ;
                
                old_par = *par_list ;
                subst_name( par_list, &p->n_listvalue ) ;
                new = newnode( LIST ) ;
                new->n_listnext = *par_list ;
                *par_list = new ;
                new->n_listvalue = (*exp)->n_listnext->n_listvalue ;
 
                *name = ' ' ; *(name+1) = '\0' ;
                strcat( name, new->n_listvalue->n_symname ) ;    
                (*exp)->n_listnext->n_listvalue = xlenter( name ) ;

                subst_name( par_list, &(p->n_listnext->n_listvalue) ) ;
                *par_list = old_par ;
                break ;
            case 4 :
                if( ((*exp)->n_listnext == NULL) ||
                    ((*exp)->n_listnext->n_listvalue == NULL) ||
                    ((*exp)->n_listnext->n_listvalue->n_type != SYM) ||
                    ((p = (*exp)->n_listnext->n_listnext) == NULL) ||
                    (p->n_listnext == NULL) ||
                    ((p->n_listnext->n_listnext != NULL) &&
                     (p->n_listnext->n_listnext->n_listnext != NULL)) )
                    xlfail( "invalid argument" ) ;
                
                old_par = *par_list ;
                subst_name( par_list, &p->n_listvalue ) ;
                new = newnode( LIST ) ;
                new->n_listnext = *par_list ;
                *par_list = new ;
                new->n_listvalue = (*exp)->n_listnext->n_listvalue ;
 
                *name = ' ' ; *(name+1) = '\0' ;
                strcat( name, new->n_listvalue->n_symname ) ;    
                (*exp)->n_listnext->n_listvalue = xlenter( name ) ;

                subst_name( par_list, &(p->n_listnext->n_listvalue) ) ;
                if( p->n_listnext->n_listnext != NULL )
                    subst_name( par_list,
                                &(p->n_listnext->n_listnext->n_listvalue) ) ;
                *par_list = old_par ;
                break ;
            case 5 :
                if( ((*exp)->n_listnext == NULL) ||
                    ((*exp)->n_listnext->n_listvalue == NULL) ||
                    ((*exp)->n_listnext->n_listvalue->n_type != SYM) ||
                    ((p = (*exp)->n_listnext->n_listnext) == NULL) )
                    xlfail( "invalid argument" ) ;
                
                old_par = *par_list ;
                subst_name( par_list, &p->n_listvalue ) ;
                new = newnode( LIST ) ;
                new->n_listnext = *par_list ;
                *par_list = new ;
                new->n_listvalue = (*exp)->n_listnext->n_listvalue ;
 
                *name = ' ' ; *(name+1) = '\0' ;
                strcat( name, new->n_listvalue->n_symname ) ;    
                (*exp)->n_listnext->n_listvalue = xlenter( name ) ;

                for( p = p->n_listnext ; p != NULL ; p = p->n_listnext )
                    subst_name( par_list, &(p->n_listvalue) ) ;
                *par_list = old_par ;
                break ;
            case 6 :
                if( ((*exp)->n_listnext == NULL) ||
                    ((*exp)->n_listnext->n_listnext == NULL) )
                    xlfail( "invalid argument" ) ;
                old_par = *par_list ;
                p = (*exp)->n_listnext->n_listnext ;
                do {
                    q = p ; 
                    val = xlmatch( LIST, &p ) ;
                    sym = xlmatch( SYM, &val ) ;
                    if( strcmp( sym->n_symname, "from" ) != 0 )
                        xlfail( "invalid argument" ) ; 
                    sym = xlmatch( SYM, &val ) ;

                    if( val == NULL )
                        xlfail( "too few arguments" ) ;
                    subst_name( par_list, &val->n_listvalue ) ; 
   
                    new = newnode( LIST ) ;
                    new->n_listnext = *par_list ;
                    *par_list = new ;
                    new->n_listvalue = sym ;

                    *name = ' ' ; *(name+1) = '\0' ;
                    strcat( name, sym->n_symname ) ;
                    q->n_listvalue->n_listnext->n_listvalue = xlenter( name ) ;

                    if( val->n_listnext != NULL ) {
                        if( val->n_listnext->n_listnext != NULL )
                            xlfail( "too many arguments" ) ;
                        subst_name( par_list, &val->n_listnext->n_listvalue );}
                } while( p != NULL ) ;
                subst_name( par_list, &(*exp)->n_listnext->n_listvalue ) ;
                *par_list = old_par ;
                break ; } }
        break ; }
}


/*. lambda - builtin function lambda .*/
/*. lambda : symb-list nexp -> function .*/

/*. makecopy - copy an exp created by "xlread" .*/
struct node *makecopy( exp )
struct node *exp ;
{
    struct node rese ;
    struct node *oldstk, *new, *lrese ;

    if( exp == NULL ) return( NULL ) ;

    /* create a new stack frame */
    oldstk = xlsave( &rese, NULL ) ;

    if( exp->n_type == LIST )
        for( lrese = NULL ; exp != NULL ;
                            exp = exp->n_listnext, lrese = new ) { 
            new = newnode( LIST ) ;
            if( lrese == NULL ) 
                rese.n_ptr = new ;
            else
                lrese->n_listnext = new ;
            new->n_listvalue = makecopy( exp->n_listvalue ) ; }
    else
       rese.n_ptr = exp ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    return( rese.n_ptr ) ;
}


/*. lbda_subst - substitute unmarked symbols by their values .*/
static void lbda_subst( exp ) 
struct node **exp ; 
{
    struct node *p, *q, *ls ;
    int found, sw ;
   
    if( *exp == NULL ) return ;

    switch( (*exp)->n_type ) {
        
    case SYM : 
        for( p = lbda_env.n_ptr, found = FALSE ; 
         (p->n_listvalue->n_type != LBDA_FRAME) && !found ; p = p->n_listnext )
            if( (strcmp( (*exp)->n_symname, p->n_listvalue->n_symname ) == 0) ){
                found = TRUE ;
                if( p->n_listvalue->n_type == LBDA_MARK )
                    (*exp) = (*exp)->n_symvalue ; }
        break ;

    case LIST :
        if( ((p = (*exp)->n_listvalue) == NULL) || (p->n_type != SYM) )
            for( p = *exp ; p != NULL ; p = p->n_listnext )
                lbda_subst( &p->n_listvalue ) ;
        else {
            if( strcmp( p->n_symname, "let" ) == 0 ) 
                sw = 1 ;
            else if( strcmp( p->n_symname, "lambda" ) == 0 )
                sw = 2 ;
            else if( (strcmp( p->n_symname, "ALL" ) == 0) ||
                     (strcmp( p->n_symname, "EXIST" ) == 0) ||
                     (strcmp( p->n_symname, "EXIST1" ) == 0) )
                sw = 3 ;
            else if( strcmp( p->n_symname, "ff" ) == 0 )
                sw = 4 ;
            else if( strcmp( p->n_symname, "foreach" ) == 0 )
                sw = 5 ;
            else if( (strcmp( p->n_symname, "set" ) == 0) ||
                     (strcmp( p->n_symname, "seq" ) == 0) ||
                     (strcmp( p->n_symname, "ff1" ) == 0) )
                sw = 6 ;
            else
                sw = 0 ;

            switch( sw ) {
            case 0 :
                for( p = *exp ; p != NULL ; p = p->n_listnext )
                    lbda_subst( &p->n_listvalue ) ;
                break ;
            case 1 :
		for( p = (*exp)->n_listnext->n_listvalue ; p != NULL ;
                                                      p = p->n_listnext ) {
                    q = p->n_listvalue->n_listvalue ;

                    /* mark the "let" symbol */
                    for( ls = lbda_env.n_ptr, found = FALSE ;
                         !found && (ls->n_listvalue->n_type != LBDA_FRAME) ;
                         ls = ls->n_listnext )
                        if( strcmp( q->n_symname, 
                                    ls->n_listvalue->n_symname ) == 0 ) {
                            found = TRUE ;
                            (ls->n_listvalue->n_type)-- ; } 
                    lbda_subst( &(p->n_listvalue->n_listnext->n_listvalue)) ;}

               p = (*exp)->n_listnext->n_listnext ; 
               while( p != NULL ) {
                   lbda_subst( &(p->n_listvalue) ) ;
                   p = p->n_listnext ; }

               /* unmark the  "let" symbols */
               for( p = (*exp)->n_listnext->n_listvalue ; p != NULL ;
                                                      p = p->n_listnext)
                   for( ls = lbda_env.n_ptr, found = FALSE ;
                        !found && (ls->n_listvalue->n_type != LBDA_FRAME) ;
                        ls = ls->n_listnext )
                       if( strcmp( p->n_listvalue->n_listvalue->n_symname,
                                   ls->n_listvalue->n_symname ) == 0 ) {
                           found = TRUE ;
                           (ls->n_listvalue->n_type)++ ; }
               break ;
           case 2 :
               for( p = (*exp)->n_listnext->n_listvalue ; p != NULL ;
                                                      p = p->n_listnext ) {
                   q = p->n_listvalue ;

                   /* mark the "lambda" symbol */
                   for( ls = lbda_env.n_ptr, found = FALSE ;
                       !found && (ls->n_listvalue->n_type != LBDA_FRAME) ;
                                                ls = ls->n_listnext )
                       if( strcmp( q->n_symname, 
                                   ls->n_listvalue->n_symname ) == 0 ) {
                           found = TRUE ;
                           (ls->n_listvalue->n_type)-- ; } }

               p = (*exp)->n_listnext->n_listnext ; 

               lbda_subst( &(p->n_listvalue) ) ;

               /* unmark the  "lambda" symbols */
               for( p = (*exp)->n_listnext->n_listvalue ; p != NULL ;
                                                  p = p->n_listnext)
                   for( ls = lbda_env.n_ptr, found = FALSE ;
                        !found && (ls->n_listvalue->n_type != LBDA_FRAME) ;
                                                      ls = ls->n_listnext )
                       if( strcmp( p->n_listvalue->n_symname,
                                     ls->n_listvalue->n_symname ) == 0 ) {
                           found = TRUE ;
                           (ls->n_listvalue->n_type)++ ; }
               break ;
          case 3 :
               p = (*exp)->n_listnext ;
               lbda_subst( &(p->n_listnext->n_listvalue) ) ;

               for( ls = lbda_env.n_ptr, found = FALSE ;
                       !found && (ls->n_listvalue->n_type != LBDA_FRAME) ;
                                                ls = ls->n_listnext )
                   if( strcmp( p->n_listvalue->n_symname ,
                               ls->n_listvalue->n_symname ) == 0 ) {
                       found = TRUE ;
                       (ls->n_listvalue->n_type)-- ; }

               lbda_subst( &p->n_listnext->n_listnext->n_listvalue ) ;

               for( ls = lbda_env.n_ptr, found = FALSE ;
                    !found && (ls->n_listvalue->n_type != LBDA_FRAME) ;
                                                  ls = ls->n_listnext )
                   if( strcmp( p->n_listvalue->n_symname,
                              ls->n_listvalue->n_symname ) == 0 ) {
                       found = TRUE ;
                       (ls->n_listvalue->n_type)++ ; }
               break ;
          case 4 :
               p = (*exp)->n_listnext ;
               lbda_subst( &(p->n_listnext->n_listvalue) ) ;

               for( ls = lbda_env.n_ptr, found = FALSE ;
                       !found && (ls->n_listvalue->n_type != LBDA_FRAME) ;
                                                ls = ls->n_listnext )
                   if( strcmp( p->n_listvalue->n_symname ,
                               ls->n_listvalue->n_symname ) == 0 ) {
                       found = TRUE ;
                       (ls->n_listvalue->n_type)-- ; }

               lbda_subst( &p->n_listnext->n_listnext->n_listvalue ) ;

               if( p->n_listnext->n_listnext->n_listnext != NULL )
               lbda_subst( &p->n_listnext->n_listnext->n_listnext->n_listvalue);

               for( ls = lbda_env.n_ptr, found = FALSE ;
                    !found && (ls->n_listvalue->n_type != LBDA_FRAME) ;
                                                  ls = ls->n_listnext )
                   if( strcmp( p->n_listvalue->n_symname,
                              ls->n_listvalue->n_symname ) == 0 ) {
                       found = TRUE ;
                       (ls->n_listvalue->n_type)++ ; }
               break ;
          case 5 :
               p = (*exp)->n_listnext ;
               lbda_subst( &(p->n_listnext->n_listvalue) ) ;

               for( ls = lbda_env.n_ptr, found = FALSE ;
                       !found && (ls->n_listvalue->n_type != LBDA_FRAME) ;
                                                ls = ls->n_listnext )
                   if( strcmp( p->n_listvalue->n_symname ,
                               ls->n_listvalue->n_symname ) == 0 ) {
                       found = TRUE ;
                       (ls->n_listvalue->n_type)-- ; }

               for( q = p->n_listnext->n_listnext ; q != NULL ;
                                                q = q->n_listnext )
                    lbda_subst( &q->n_listvalue ) ;

               for( ls = lbda_env.n_ptr, found = FALSE ;
                    !found && (ls->n_listvalue->n_type != LBDA_FRAME) ;
                                                  ls = ls->n_listnext )
                   if( strcmp( p->n_listvalue->n_symname,
                              ls->n_listvalue->n_symname ) == 0 ) {
                       found = TRUE ;
                       (ls->n_listvalue->n_type)++ ; }
               break ;
          case 6 :
               p = (*exp)->n_listnext->n_listnext ;
               do {
                   q = p->n_listvalue->n_listnext ;
                   lbda_subst( &q->n_listnext->n_listvalue ) ;
                   for( ls = lbda_env.n_ptr, found = FALSE ;
                        !found && (ls->n_listvalue->n_type != LBDA_FRAME) ;
                                                 ls = ls->n_listnext ) {
                       if( strcmp( q->n_listvalue->n_symname, 
                                   ls->n_listvalue->n_symname ) == 0 ) {
                           found = TRUE ;
                           (ls->n_listvalue->n_type)-- ; } }

                   if( q->n_listnext->n_listnext != NULL )
                       lbda_subst( &q->n_listnext->n_listnext->n_listvalue ) ;
                   p = p->n_listnext ;
               } while( p != NULL ) ;

               lbda_subst( &(*exp)->n_listnext->n_listvalue ) ;

               p = (*exp)->n_listnext->n_listnext ;
               do {
                   q = p->n_listvalue->n_listnext ;
                   for( ls = lbda_env.n_ptr, found = FALSE ;
                        !found && (ls->n_listvalue->n_type != LBDA_FRAME) ;
                                                 ls = ls->n_listnext ) {
                       if( strcmp( q->n_listvalue->n_symname, 
                                   ls->n_listvalue->n_symname ) == 0 ) {
                           found = TRUE ;
                           (ls->n_listvalue->n_type)++ ; } }
                   p = p->n_listnext ;
               } while( p != NULL ) ;
               break ; } }
        break ; }
}

static struct node *lambda( args )
struct node *args ;
{
    struct node arg, fun, lbdy, par_list ;
    struct node *oldstk, *fargs, *p, *q, *new ;
    int found ;
    char name[STRMAX+1] ;
    
    /* create a new stack frame */
    oldstk = xlsave( &arg, &fun, &lbdy, &par_list, NULL ) ;

    /* initialize */
    arg.n_ptr = args ;

    /* get the formal argument list */
    fargs = xlmatch( LIST, &args ) ;

    lbdy.n_ptr = xlarg( &args ) ;
    xllastarg( args ) ;

    for( p = fargs, found = TRUE ; found && (p != NULL) ; p = p->n_listnext ) {
        if( ((q = p->n_listvalue) == NULL) || (q->n_type != SYM) )
            xlfail( "invalid argument" ) ;

        if( *(q->n_symname) == ' ' )
            found = FALSE ;
        else {
            new = newnode( LIST ) ;
            new->n_listnext = par_list.n_ptr ;
            par_list.n_ptr = new ;
            new->n_listvalue = q ;
 
            *name = ' ' ; *(name+1) = '\0' ;
            strcat( name, q->n_symname ) ;    
            p->n_listvalue = xlenter( name ) ; } }
    
    if( found )
        subst_name( &par_list.n_ptr, &lbdy.n_ptr ) ;
    
    if( lbda_env.n_ptr != NULL ) {

        /* mark the formal arguments in "lbda_env" as non replaceable */
        for( p = fargs ; p != NULL ; p = p->n_listnext ) 
            for( q = lbda_env.n_ptr, found = FALSE ;
                 !found && (q->n_listvalue->n_type != LBDA_FRAME) ;
                                                  q = q->n_listnext ) {
                if( strcmp( p->n_listvalue->n_symname,
                            q->n_listvalue->n_symname ) == 0 ) {
                    found = TRUE ;
                    (q->n_listvalue->n_type)-- ; } }

        lbdy.n_ptr = makecopy( lbdy.n_ptr ) ;

        /* substitute in "lbdy" the unmarked symbols in "lbda_env" */
        lbda_subst( &lbdy.n_ptr ) ;
 
        /* unmark the formal arguments */
        for( p = fargs ; p != NULL ; p = p->n_listnext)
            for( q = lbda_env.n_ptr, found = FALSE ;
                 !found && (q->n_listvalue->n_type != LBDA_FRAME) ;
                                                   q = q->n_listnext )
                if( strcmp( p->n_listvalue->n_symname,
                            q->n_listvalue->n_symname ) == 0 ) {
                    found = TRUE ;
                    (q->n_listvalue->n_type)++ ; }
     }

    /* create a new function definition */
    fun.n_ptr = newnode( FUN ) ;
    fun.n_ptr->n_listvalue = fargs ;
    fun.n_ptr->n_listnext = newnode( LIST ) ;
    fun.n_ptr->n_listnext->n_listvalue = lbdy.n_ptr ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return the function definition */
    return( fun.n_ptr ) ;
}

/* defexp - builtin function def */
/* defexp : symbexp "lambda" symb-list nexp -> symb */
/*     ( side effect : changes "oblist" and "ops" values ) */
/* or                                                      */
/* defexp : symbexp exp -> symb */
/*     ( side effect : changes "oblist" and "obs" values ) */
static struct node *defexp( args )
struct node *args;
{
    struct node sym ;
    struct node *oldstk, arg, *saux, *p ;

    /* create a new stack frame */
    oldstk = xlsave( &sym, NULL ) ;

    arg.n_ptr = args ;
    sym.n_ptr = xlevmatch( SYM, &arg.n_ptr ) ;

    if( reserv_symb( sym.n_ptr->n_symname ) )
        xlfail( "may not redefine symbol" ) ;

    saux = xlarg( &arg.n_ptr ) ;
    if( (saux != NULL) && (saux->n_type == SYM) &&
        (strcmp( "lambda", saux->n_symname ) == 0) ) {
        sym.n_ptr->n_symvalue = lambda( arg.n_ptr ) ;
        m2enter( sym.n_ptr->n_symname, ops, obs ) ; }
    else {
       p = xleval( saux ) ;
       xllastarg( arg.n_ptr ) ;
       sym.n_ptr->n_symvalue = p ;
       m2enter( sym.n_ptr->n_symname, obs, ops ) ; }

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return the symbol */
    return( sym.n_ptr ) ;
}


/*. def - builtin function def .*/
/*. def : symb "lambda" symb-list nexp -> symb .*/
/*.     ( side effect : changes "oblist" and "ops" values ) .*/
/*. or                                                      .*/
/*. def : symb exp -> symb .*/
/*.     ( side effect : changes "oblist" and "obs" values ) .*/
static struct node *def( args )
struct node *args;
{
    struct node sym ;
    struct node *oldstk, *arg, *saux, *p ;

    /* create a new stack frame */
    oldstk = xlsave( &sym, NULL ) ;

    arg = args ;
    sym.n_ptr = xlmatch( SYM, &arg ) ; 
    /* if((sym.n_ptr=recmatch  (SYM, &arg ))== NULL) 
        sym.n_ptr=xlevmatch (SYM, &arg ) ; */

    if( reserv_symb( sym.n_ptr->n_symname ) )
        xlfail( "may not redefine symbol" ) ;

    saux = xlarg( &arg ) ;
    if( (saux != NULL) && (saux->n_type == SYM) &&
        (strcmp( "lambda", saux->n_symname ) == 0) ) {
        sym.n_ptr->n_symvalue = lambda( arg ) ;
        m2enter( sym.n_ptr->n_symname, ops, obs ) ; }
    else {
       p = xleval( saux ) ;
       xllastarg( arg ) ;
       sym.n_ptr->n_symvalue = p ;
       m2enter( sym.n_ptr->n_symname, obs, ops ) ; }

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return the symbol */
    return( sym.n_ptr ) ;
}


/*. let - builtin function let .*/
/*. let : pair-list exp... -> value .*/
/*.       pair = "(" symb exp ")" .*/
static struct node *let( args )
struct node *args ;
{
    struct node list, valarg, forms, par_list ;
    struct node *oldenv, *oldenv1, *oldstk, *val, *pair ;
    int sflag ;
    char name[STRMAX+1] ;
	
    /* create a new stack frame */
    oldstk = xlsave( &list, &forms, &valarg, &par_list, NULL ) ;

    list.n_ptr = xlmatch( LIST, &args ) ;
    forms.n_ptr = args ;

    sflag = FALSE ;
    oldenv = xlenv ;
    while( list.n_ptr != NULL ) {
        if( ((pair = list.n_ptr->n_listvalue) == NULL) ||
            (pair->n_listvalue == NULL) ||
            (pair->n_listvalue->n_type != SYM) ||
            (pair->n_listnext == NULL) ||
            (pair->n_listnext->n_listnext != NULL) )
            xlfail( "invalid argument" ) ;
        sflag = sflag || (*(pair->n_listvalue->n_symname) != ' ') ;
        if( sflag )
            subst_name( &par_list.n_ptr, &(pair->n_listnext->n_listvalue) ) ;
#ifdef LAZY
        valarg.n_ptr = delay( pair->n_listnext->n_listvalue ) ;
#else
        valarg.n_ptr = xleval( pair->n_listnext->n_listvalue ) ;
#endif
        if( sflag ) {
            val = newnode( LIST ) ;
            val->n_listnext = par_list.n_ptr ;
            par_list.n_ptr = val ;
            val->n_listvalue = pair->n_listvalue ;
            *name = ' ' ; *(name+1) = '\0' ;
            strcat( name, pair->n_listvalue->n_symname ) ;
            pair->n_listvalue = xlenter( name ) ; }
        oldenv1 = xlenv ;
        xlbind( pair->n_listvalue, valarg.n_ptr ) ;
        xlfixbindings( oldenv1 ) ; 
        list.n_ptr = list.n_ptr->n_listnext ; }

    val = NULL ;
    while( forms.n_ptr != NULL ) {
        if( sflag )
            subst_name( &par_list.n_ptr, &(forms.n_ptr->n_listnext) ) ;
	val = xleval( forms.n_ptr->n_listvalue ) ;
        forms.n_ptr = forms.n_ptr->n_listnext ; }

    xlunbind( oldenv ) ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return pointer to the last "exp" value */
    return( val ) ;
}


/*. PACK - builtin function PACK .*/
/*. PACK : (symb | int)-list -> symb .*/
static struct node *PACK( args )
struct node *args ;
{
    struct node arg ;
    struct node *oldstk, *s, *sym ;
    char name[STRMAX] ;
    char buf[STRMAX] ;
    int len ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, NULL ) ;

    arg.n_ptr = args ;
    arg.n_ptr = xlevmatch( LIST, &args ) ;
    xllastarg( args ) ;

    *name = '\0' ; len = 0 ;
    do {
       s = xlarg( &arg.n_ptr ) ;
       if( (s == NULL) || ((s->n_type != SYM) && (s->n_type != INT)) )
           xlfail( "bad argument type" ) ;
       if( s->n_type == INT ) {
           if( s->n_int < 0 ) xlfail( "invalid argument" ) ;
           sprintf( buf, "%d", s->n_int ) ; }
       else
           strcpy( buf, s->n_symname ) ;
       if( (len += strlen( buf )) >= STRMAX)
           xlfail( "name too long" ) ;
       strcat( name, buf ) ;
    } while( arg.n_ptr != NULL ) ;

    sym = xlenter( name ) ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return pointer to the new symbol */
    return( sym ) ;
}

 
/*. UNPACK - builtin function UNPACK .*/
/*. UNPACK: sym -> (sym | int)-list .*/
static struct node *UNPACK( args )
struct node *args ;
{
    struct node resl ;
    struct node *oldstk, *new, *lresl ;
    char name[STRMAX], sname[2] ;
    int i ;

    /* create a new stack frame */
    oldstk = xlsave( &resl, NULL ) ;
 
    strcpy( name, xlevmatch( SYM, &args )->n_symname ) ;
    xllastarg( args ) ;

    for( lresl = NULL, i = 0 ; name[i] != '\0' ; i++, lresl = new ) {
        new = newnode( LIST ) ;
        if( lresl == NULL )
            resl.n_ptr = new ;
        else
            lresl->n_listnext = new ;
        if( (name[i] >= '0') && (name[i] <= '9') ) {
            new = new->n_listvalue = newnode( INT ) ;
            new->n_int = (int) name[i] ; }
        else {
            sname[0] = name[i] ; sname[1] = '\0' ;
            new->n_listvalue = xlenter( sname ) ; } }
                  
    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return the constructed list */
    return( resl.n_ptr ) ;
}
    
   
/*. xlenter - enter a symbol into the symbol table .*/
DllExport struct node *xlenter( sname )
char *sname ;
{
    struct node *sptr ;

    /* check for nil */
    if( strcmp( sname, "nil" ) == 0 )
        return( NULL ) ;

    /* initialize oblist if it is the first time xlenter is called */
    if( oblist == NULL ) {
        oblist = newnode( SYM ) ;
        oblist->n_symname = strsave( "oblist" ) ;
        oblist->n_symvalue = newnode( LIST ) ;
	oblist->n_symvalue->n_listvalue = oblist ; }

    /* check for symbol already in table */
    for( sptr = oblist->n_symvalue ; sptr != NULL ; sptr = sptr->n_listnext )
        if( strcmp( sptr->n_listvalue->n_symname, sname ) == 0 )
            return( sptr->n_listvalue ) ;

    /* enter a new symbol and link it into the symbol list */
    sptr = newnode( LIST ) ;
    sptr->n_listnext = oblist->n_symvalue ;
    oblist->n_symvalue = sptr ;
    sptr->n_listvalue = newnode( SYM ) ;
    sptr->n_listvalue->n_symname = strsave( sname ) ;

    /* return the new symbol */
    return( sptr->n_listvalue ) ;
}


/*. xlsubr - define a builtin function .*/
DllExport void xlsubr( sname, subr )
char *sname ;
struct node *(*subr)() ;
{
    struct node *sym ;

    /* enter the symbol */
    sym = xlenter( sname ) ;

    /* initialize the value */
    sym->n_symvalue = newnode( SUBR ) ;
    sym->n_symvalue->n_subr = subr ;
}


/*. xmdefinit - initialize xmetoo state .*/
void xmdefinit()
{
    ops = xlenter( "ops" ) ;
    obs = xlenter( "obs" ) ;
    m2enter( "obs", obs, ops ) ;
    m2enter( "ops", obs, ops ) ;

    xlenter( "from" ) ;

    xlsubr( "lambda", lambda ) ;
    xlsubr( "def", def) ;
    xlsubr( "defexp", defexp) ;
    xlsubr( "let", let) ;
    xlsubr( "PACK", PACK) ;
    xlsubr( "UNPACK", UNPACK) ;
}


