/* xmetoo.h - system specifications and global declarations */
/* win95 changes: jj+fln */
#ifdef XM95
#define DllExport __declspec(dllexport)
#else
#define DllExport
#endif


#define true xm_true
#define false xm_false

/* system specific definitions */

/* FGETNAME	define if system supports 'fgetname' */
/* NNODES	number of nodes to allocate in each request */
/* TDEPTH	trace stack depth */

/* for unix compilers */
#ifdef unix
#define NNODES	2000
#define TDEPTH	1000
#endif

/* for the turbo C compiler */
#ifdef __TURBOC__
#define NNODES	200
#define TDEPTH	100
#define NULL 0L
#define MSDOS
#endif

/* default important definitions */
#ifndef NNODES
#define NNODES	200
#endif
#ifndef TDEPTH
#define TDEPTH	100
#endif

/* useful definitions */
#define TRUE	1
#define FALSE	0

/* program limits */
#define STRMAX	100	/* maximum length of a string constant */

/* node types */
#define LBDA_MARK  (-2)
#define LBDA_FRAME (-1)
#define FREE       0
#define SUBR       1
#define LIST       2
#define SYM        3
#define INT        4
#define NUM        4
#define STR        5
#define FPTR       6
#define SET        7
#define REL        8
#define FF         9
#define FUN        10
#ifdef LAZY
#define REC        11
#define LISTREC    12
#endif
#ifdef COMM
#define CHAN       13
#define NILM       255
#define SYNC       256
#define ENDCOMP    257
#endif
#define SAVE_MARK  128
#define NILV       254

#ifdef USERTYPE

struct lab_type {
    char            *label ;
    struct t_table  *type ;
    struct lab_type *next ;
} ;

struct type {
    struct t_table  *type ;
    struct type *next ;
} ;

union b {
    struct lab_type *tcomp ;    /* list of tuple label-component pairs */
    struct type *acomp ;        /* list of alternative components */
    struct t_table *type ;      /* set, list or equality */
    struct {                    /* relation or finite function */
        struct t_table *dom ;   /* domain type */
        struct t_table *ran ;   /* range type */
    } rel ;
} ;

struct t_table {
    char           *tname ;     /* type name */
    int            nnumber ;    /* type code */
    int            tnumber ;    /* definition code */
    union b        def ;        /* type definition */
    struct t_table *next ;
} ;

/* definition codes */
#define TUNDEF  0
#define TSYM    1
#define TSTR    2
#define TINT    3
#define TNUM    3
#define TFPTR   4
#ifdef COMM
#define TCHAN   5
#endif
#define TANY    6
#define TLIST   7
#define TSET    8
#define TEQ     9
#define TREL   10
#define TFF    11
#define TTUP   12
#define TALT   13
#define TNIL   14


/* maximum number of user defined types */
#define MAX_TYPE       500
#define TG         129

/* node types */
#define LABEL      258
#define TYPE       259
#define MACTYPE    260
/* types >= 1000 are reserved for user defined */

/* first user type code */
#define USERDEFTYPE    1000

#endif


/* node flags */
#define MARK	1
#define LEFT	2

/* string types */
#define DYNAMIC	0
#define STATIC	1


/* symbol structure */
struct xsym {
    char *xsy_name;		/* symbol name */
    struct node *xsy_value;	/* the current value */
};

/* subr node structure */
struct xsubr {
    struct node *(*xsu_subr)();	/* pointer to an internal routine */
};

/* list node structure */
struct xlist {
    struct node *xl_value;	/* value at this node */
    struct node *xl_next;	/* next node */
};

/* integer node structure */
struct xint {
#ifdef    REAL_NUMBER
    double xi_int;			/* integer value */
#else  /* REAL_NUMBER */
    int xi_int;			/* integer value */
#endif /* REAL_NUMBER */
};

/* string node structure */
struct xstr {
    int xst_type;		/* string type */
    char *xst_str;		/* string pointer */
};

/* file pointer node structure */
struct xfptr {
    FILE *xf_fp;		/* the file pointer */
};

#ifdef COMM
/* CHAN structure  */	
struct xchan {
    struct node  *q_name ;
    int          qid ;
};
#endif

#ifdef USERTYPE
/* TYPE structure */
struct xtype {
    struct t_table *n_tp ;
} ;
#endif

/* shorthand macros for accessing node substructures */

/* symbol node */
#define n_symname	n_info.n_xsym.xsy_name
#define n_symvalue	n_info.n_xsym.xsy_value

/* subr node */
#define n_subr		n_info.n_xsubr.xsu_subr

/* list (and binding and REC) node */
#define n_listvalue	n_info.n_xlist.xl_value
#define n_listnext	n_info.n_xlist.xl_next
#define n_bndsym	n_info.n_xlist.xl_value
#define n_bndvalue	n_info.n_xlist.xl_next
#define n_left		n_info.n_xlist.xl_value
#define n_right		n_info.n_xlist.xl_next
#define n_ptr		n_info.n_xlist.xl_value
#ifdef LAZY
#define n_recexp        n_info.n_xlist.xl_value
#define n_recenv        n_info.n_xlist.xl_next
#endif

/* integer node */
#define n_int		n_info.n_xint.xi_int

/* real node */
#define n_num		n_info.n_xint.xi_int

/* string node */
#define n_str		n_info.n_xstr.xst_str
#define n_strtype	n_info.n_xstr.xst_type

/* file pointer node */
#define n_fp		n_info.n_xfptr.xf_fp

#ifdef COMM
/* CHAN node    */
#define n_qname        n_info.n_xchan.q_name
#define n_qid          n_info.n_xchan.qid
#endif

#ifdef USERTYPE
/* TYPE node */
#define n_typedef      n_info.n_xtype.n_tp
#define n_macdef       n_info.n_xsym.xsy_value
#endif

/* node structure */
struct node {
    int  n_type ;		/* type of node */
    char n_flags ;		/* flag bits */
    union {			/* value */
	struct xsym n_xsym;	/*     symbol node */
	struct xsubr n_xsubr;	/*     subr node */
	struct xlist n_xlist;	/*     list node */
	struct xint n_xint;	/*     integer node */
	struct xstr n_xstr;	/*     string node */
	struct xfptr n_xfptr;	/*     file pointer node */
#ifdef COMM
	struct xchan n_xchan;	/*     chan node */
#endif
#ifdef USERTYPE
        struct xtype n_xtype;  /*     type node */
#endif
    } n_info ;
} ;

/* external procedure declarations */
extern struct node *xlread();		/* read an expression */
extern struct node *xleval();		/* evaluate an expression */
#ifdef LAZY
extern struct node *delay();            /* recipe an expression */
#endif
extern struct node *xlarg();		/* fetch an argument */
extern struct node *xlevarg();		/* fetch and evaluate an argument */
extern struct node *xljustevarg();	/* fetch and evaluate an argument */
#ifdef USERTYPE
extern struct node *userevarg();        /* xlevarg() using labeval() */
extern struct node *labeval();          /* xleval() ignoring
                                           prefix type labels */
#endif
extern struct node *xlmatch();		/* fetch an typed argument */
extern struct node *xlevmatch();	/* fetch and evaluate a typed arg */
extern void        xllastarg();         /* ensures there are no more args */
DllExport struct node *xlenter();		/* enter a symbol into the oblist */
DllExport void        xlsubr();		/* enter a builtin function */
extern struct node *xlsave();		/* generate a stack frame */
extern struct node *newnode();		/* allocate a new node */
extern char        *xmstralloc();		/* allocate string space */
extern char        *strsave();		/* make a safe copy of a string */
extern void free() ;                    /* frees space allocated by strsave() */
#define strfree( s ) free( s )         
extern void        xlfail();            /* error handling routine */
#include <string.h>
