/* xmetoo.c - main module */

#include <stdio.h>
#include <setjmp.h>
#include "xmetoo.h"
#include <signal.h>

/* TURBOC stack size definition */
#ifdef __TURBOC__
unsigned _stklen = 65500 ;
#endif


/* external variables */
extern struct node *xlenv ;
extern struct node *xlstack ;

int neverprompt,PP=0;

void printversion(){
#include "version"
   if(!neverprompt)
   { printf( "\nUM-XMETOO version %s ",xm_version ) ;
#ifdef LAZY
     printf( ".LAZY" ) ;
#endif
#ifdef DLL
     printf( ".DLL" ) ;
#endif
#ifdef COMM
     printf( ".COMM" ) ;
#endif
#ifdef USERTYPE
     printf( ".USERTYPE" ) ;
#endif
     printf( " (aap/pem/cjr/jj  %s)\n",xm_date ) ;
  }
}


/* external procedures */
/* extern int setjmp() ; */
extern void        xltin() ;
extern struct node *load() ;
extern struct node *xlreadfromstr( ) ;
extern void        xlprint() ;
extern void        xmdmeminit() ;
extern void        xmreadinit() ;
extern void        xmevalinit() ;
extern void        xmprintinit() ;
extern void        xmsaveinit() ;
extern void        xmdefinit() ;
#ifdef USERTYPE
extern void        xmutypeinit() ;
#endif
extern void        xmxlispinit() ;
extern void        xmboolinit() ;
#ifdef    REAL_NUMBER
extern void        xmnuminit() ;
#else  /* REAL_NUMBER */
extern void        xmintinit() ;
#endif /* REAL_NUMBER */

extern void        xmstrinit() ;
#ifdef DLL
extern void        dllinit() ;
#endif
extern void        pipeinit() ;
#ifdef COMM
extern void        xmmsginit() ;
#endif
extern void        xmfioinit() ;
extern void        xmlistinit() ;
extern void        xmsetinit() ;
extern void        xmrelinit() ;
extern void        xmffinit() ;
extern void        userinit() ;
extern void        userinit2() ;
extern void ppaux() ;

void usersigint(x)
int x;
{ xlfail("Interrupt by user");}

/* global variables */
jmp_buf xljmpbuf ;
struct node *oblist ;
struct node *oblimit ;
struct node lbda_env ;

char jjbuf[BUFSIZ];

DllExport int libxminit(argc,argv)
int argc ;
char *argv[] ;
{
    struct node expr ;

    int i, first_pass ;
    int first_file=0;
#ifdef __TURBOC__
   /*     int i ; */
#else
    
if(argc>1 && strcmp(argv[1],"-n")==0)
 { setvbuf(stdout,jjbuf,_IOLBF,BUFSIZ);
   first_file=2;
   neverprompt=TRUE; }
#endif
    oblist = NULL ;

    /* initialize the dynamic memory module (must be first) */
    xmdmeminit() ;

    xmreadinit() ;
    xmevalinit() ;
    xmprintinit() ;
    xmsaveinit() ;
    xmdefinit() ;
#ifdef USERTYPE
    xmutypeinit() ;
#endif
    xmxlispinit() ;
    xmboolinit() ;
    xmintinit() ;
    xmstrinit() ;
#ifdef DLL
    dllinit() ;
#endif
    pipeinit();
#ifdef COMM
    xmmsginit() ;
#endif
    xmfioinit() ;
    xmlistinit() ;
    xmsetinit() ;
    xmrelinit() ;
    xmffinit() ;
    userinit() ;/* user define functions JJ1994 (=nop by default) */

    /* delimitates reserved area in oblist */
    oblimit = oblist->n_symvalue ;

    /* initialize terminal input */
    xltin( FALSE ) ;

    /* set error jump address */
    first_pass = TRUE ;
    userinit2() ;/* user define functions JJ1994 (=nop by default) */
    if(setjmp( xljmpbuf )) return FALSE;
        
    signal(SIGINT,usersigint);

    if( first_pass ) {
        first_pass = FALSE ;
        /* initialize stack and evaluation environment */
        xlstack = xlenv = NULL ;
        /* read the input file(s) if specified */

        if( argc > first_file ) {
            xlsave( &expr, &lbda_env,  NULL ) ;
            expr.n_ptr = newnode( LIST ) ;
            expr.n_ptr->n_listvalue = newnode( STR ) ;
            for( i = first_file ; i < argc ; i++ ) {
                expr.n_ptr->n_listvalue->n_str = strsave( argv[i] ) ;
                load( expr.n_ptr ) ;
                strfree( expr.n_ptr->n_listvalue->n_str ) ; }
            expr.n_ptr->n_listvalue->n_str = NULL ; } }
    return 1;

}

DllExport struct node * libxmevalstr(str)
char *str;
        {
    struct node expr ;

    if(setjmp( xljmpbuf )) return ((struct node *)-1) ;

        /* free any previous expression and leftover context */
    xlstack = xlenv = NULL ;

        /* create a new stack frame */
    xlsave( &expr, &lbda_env, NULL ) ;

        /* read an expression */
    expr.n_ptr = xlreadfromstr( str ) ;

        /* evaluate the expression */
#ifdef USERTYPE
    expr.n_ptr = labeval( expr.n_ptr ) ;
#else
    expr.n_ptr = xleval( expr.n_ptr ) ;
#endif
    return(expr.n_ptr);
}

DllExport struct node * libxmeval(args)
struct node * args;
        {
    struct node expr ;

    if(setjmp( xljmpbuf )) return ((struct node *)-1) ;

        /* free any previous expression and leftover context */
     xlstack = xlenv = NULL ;

        /* create a new stack frame */
     xlsave( &expr, &lbda_env, NULL ) ;

     expr.n_ptr = args;

     /* evaluate the expression */
#ifdef USERTYPE
     expr.n_ptr = labeval( expr.n_ptr ) ;
#else
     expr.n_ptr = xleval( expr.n_ptr ) ;
#endif

     return(expr.n_ptr);
}

