/* xmeval.c - xmetoo evaluator */

#include <stdio.h>
#include <setjmp.h>
#ifdef __TURBOC__
  /* nao tem varargs.h */
#else
#include <varargs.h>
#endif
#include "xmetoo.h"


/* external variables */
jmp_buf xljmpbuf ;
extern struct node *xlenv ;
extern struct node lbda_env ;
extern int lev ;
extern int neverprompt ;

/* external functions */
/* extern void longjmp() ; */
extern void xlbind() ;
extern void xlfixbindings() ;
extern void xlunbind() ;
extern void xlprint() ;
extern void xltin() ;
extern struct node *m2read() ;
#ifdef USERTYPE
extern struct node *xmselect() ;
extern struct node *construct() ;
#endif


/* global variables */
struct node *xlstack ;
int         trace_pointer ;
char * last_error = NULL;	


/* local variables */
static struct node *trace_stack[TDEPTH] ;  /* trace stack */
static int         tr_flag ;


/* forward declarations */
extern struct node *evlist() ;
extern struct node *evfun() ;
extern struct node *xleval() ;


/*. tpush - add an entry to the trace stack .*/
static void tpush( nptr )
struct node *nptr ;
{
    if( ++trace_pointer < TDEPTH )
        trace_stack[trace_pointer] = nptr ;
}


/*. tpop - pop an entry from the trace stack .*/
static void tpop()
{
    trace_pointer-- ;
}


/*. trace_view - do a back trace .*/
static void trace_view()
{
    if( trace_pointer >= TDEPTH )
        trace_pointer = TDEPTH - 1 ;

    for( ; trace_pointer >= 0 ; trace_pointer-- ) {
        fprintf(stderr, "%d - ", trace_pointer ) ;
        xlprinterr( trace_stack[trace_pointer], TRUE,stderr) ;
        fputc('\n',stderr) ; }
}



#ifdef __TURBOC__
/* atencao: este codigo 'e nao portavel; ver versao associada ao #else */
/*. xlsave - save nodes on the stack .*/
struct node *xlsave( n )
struct node *n ;
{
    struct node **nptr, *oldstk ;

    /* save the old stack pointer */
    oldstk = xlstack ;

    /* save each node */
    for( nptr = &n ; *nptr != NULL ; nptr++ ) {
        (*nptr)->n_type = LIST ;
        (*nptr)->n_listvalue = NULL ;
        (*nptr)->n_listnext = xlstack ;
        xlstack = *nptr ; }

    /* return the old stack pointer */
    return( oldstk ) ;
}
#else
/* jj xlsave - save nodes on the stack */
struct node *xlsave( va_alist )
va_dcl
{
    va_list n;
    struct node *nptr, *oldstk ;

    /* save the old stack pointer */
    oldstk = xlstack ;

    va_start(n);

    /* save each node */
    for(;( nptr = va_arg(n, struct node *)) != NULL ; ) {
        (nptr)->n_type = LIST ;
        (nptr)->n_listvalue = NULL ;
        (nptr)->n_listnext = xlstack ;
        xlstack = nptr ; }

    va_end(n);

    /* return the old stack pointer */
    return( oldstk ) ;
}
#endif

/*. jjxlfail - error handling routine .*/
void jjxlfail( err, e )
char *err ;
struct node *e;

{
    /* print the error message */
    fprintf( stderr, "\nerror: %s\n", err ) ;
    last_error = err;

    /* unbind bound symbols */
    xlunbind( (struct node *) NULL ) ;

    /* restore input to the terminal */
    xltin( TRUE ) ;

    /* restore "load" depth counter */
    lev = 0 ;

    /* show an argument */
    fputs("0 - ",stderr) ;
    xlprinterr( e, TRUE,stderr) ;
    fputs("\n",stderr) ;

    /* do the back trace */
    trace_view() ;
    trace_pointer = -1 ;

    /* restart */
    longjmp( xljmpbuf, 1 ) ;
}
/*. xlfail - error handling routine .*/
void xlfail( err )
char *err ;
{
    /* print the error message */
    fprintf( stderr, "\nerror: %s\n", err ) ;
    last_error = err;

    /* unbind bound symbols */
    xlunbind( (struct node *) NULL ) ;

    /* restore input to the terminal */
    xltin( TRUE ) ;

    /* restore "load" depth counter */
    lev = 0 ;

    /* do the back trace */
    trace_view() ;
    trace_pointer = -1 ;

    /* restart */
    longjmp( xljmpbuf, 1 ) ;
}


/*. xlarg - get the next argument .*/
struct node *xlarg( pargs )
struct node **pargs ;
{
    struct node *arg ;

    /* make sure the argument exists */
    if( *pargs == (struct node *)NULL )
        xlfail( "too few arguments" ) ;

    /* get the argument value */
    arg = (*pargs)->n_listvalue ;

    /* move the argument pointer ahead */
    *pargs = (*pargs)->n_listnext ;

    /* return the argument */
    return( arg ) ;
}


/*. xlmatch - get an argument and match its type .*/
struct node *xlmatch( type, pargs )
int type ;
struct node **pargs ;
{
    struct node *arg ;

    /* get the argument */
    arg = xlarg( pargs ) ;

    /* check its type */
    if( (type == LIST) || (type == SET) || (type == FF) || (type == REL) ) {
        if( (arg != NULL) && (arg->n_type != type) )
            xlfail( "bad argument type" ) ; }
    else
        if( (arg == NULL) || (arg->n_type != type) )
            xlfail( "bad argument type" ) ;

    /* return the argument */
    return( arg ) ;
}


/*. xlevarg - get the next argument and evaluate it .*/
struct node *xlevarg( pargs )
struct node **pargs ;
{
    struct node val ;
    struct node *oldstk ;

    /* create a new stack frame */
    oldstk = xlsave( &val, NULL ) ;

    /* get the argument */
    val.n_ptr = xlarg( pargs ) ;

    /* evaluate the argument */
    val.n_ptr = xleval( val.n_ptr ) ;
/*    while( (val.n_ptr != NULL) && (val.n_ptr->n_type >= USERDEFTYPE ) )
      val.n_ptr= val.n_ptr->n_symvalue ; */

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return the argument */
    return( val.n_ptr ) ;
}

/* xljustevarg - get the next argument and evaluate it */
struct node *xljustevarg( pargs )
struct node **pargs ;
{
    struct node val ;
    struct node *oldstk ;

    /* create a new stack frame */
    oldstk = xlsave( &val, NULL ) ;

    /* get the argument */
    val.n_ptr = xlarg( pargs ) ;

    /* evaluate the argument */
    val.n_ptr = xleval( val.n_ptr ) ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return the argument */
    return( val.n_ptr ) ;
}



/*. xlevmatch - get an evaluated argument and match its type .*/
struct node *xlevmatch( type, pargs )
int type ;
struct node **pargs ;
{
    struct node *arg ;

    /* get the argument */
    arg = xlevarg( pargs ) ;

    /* check its type */
    if( (type == LIST) || (type == SET) || (type == FF) || (type == REL) ) {
        if( (arg != NULL) && (arg->n_type != type) )
            xlfail( "bad argument type" ) ; }
    else
        if( (arg == NULL) || (arg->n_type != type) )
            xlfail( "bad argument type" ) ;

    /* return the argument */
    return( arg ) ;
}


/*. xllastarg - make sure the remainder of the argument list is empty .*/
void xllastarg( args )
struct node *args ;
{
    if( args != NULL )
        xlfail( "too many arguments" ) ;
}


#ifdef LAZY

/*. delay - create a reciped expression .*/

static struct node *recipe, *l_env ; /* variables for recipe construction */

/* env_rec_exp - creates a recipe environment for a given exp */
static void env_rec_exp( exp )
struct node *exp ;
{
    struct node *new ;

    if( exp == NULL ) return ;

    if( exp->n_type == SYM ) {
        new = newnode( LIST ) ;
        if( l_env == NULL )
            recipe->n_recenv = new ;
        else
            l_env->n_listnext = new ;
        l_env = new ;
        new = new->n_listvalue = newnode( LIST ) ;
        new->n_listvalue = exp ;
        new = new->n_listnext = newnode( LIST ) ;
        new->n_listvalue = exp->n_symvalue ; }
    else
        if( exp->n_type == LIST )
            do {
                env_rec_exp( exp->n_listvalue ) ;
                exp = exp->n_listnext ;
            } while( exp != NULL ) ;
}

struct node *delay( exp )
struct node *exp ;
{
    struct node rec, e ;
    struct node *oldstk ;

    if( exp == NULL) return( NULL ) ;

    /* create a new stack frame */
    oldstk = xlsave( &rec, &e, NULL ) ;

    e.n_ptr = exp ;
 
    /* create a recipe */
    rec.n_ptr = recipe = newnode( REC ) ;   
    recipe->n_recexp = exp ;
    l_env = NULL ;
    env_rec_exp( exp ) ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return pointer to the recipe */
    return( recipe ) ;
}

#endif


/*. eval - builtin function eval .*/
/*. eval : exp -> val .*/
static struct node *eval( args )
struct node *args ;
{
    struct node *expr ;

    /* get the expression to evaluate */
    expr = xlevarg( &args ) ;

    /* make sure there aren't any more arguments */
    xllastarg( args ) ;

    /* return the expression evaluated */
    return( xleval( expr ) ) ;
}


/*. xleval - evaluate an xlisp expression .*/
struct node *xleval( expr )
struct node *expr;
{
    struct node exp, m2rec ;
    struct node *oldstk, *val ;
#ifdef LAZY
    struct node *m2p, *m2oldenv, *m2val ;
#endif

    /* evaluate null to itself */
    if( expr == NULL )
        return( NULL ) ;
                         
    /* create a new stack frame */
    oldstk = xlsave( &exp, &m2rec, NULL ) ;
    exp.n_ptr = expr ;

    /* check type of value */
    switch( expr->n_type ) {
    case LIST :
#ifdef LAZY
    case LISTREC :
#endif
        val = evlist( expr ) ;
        break ;
    case SYM :
        val = expr->n_symvalue ;
#ifdef LAZY
        if( (val != NULL) && (val->n_type == REC) )
            expr->n_symvalue = val = xleval( val ) ;
#endif
        break ;
    case INT :
    case STR :
    case SUBR :
    case FUN :
    case FPTR :
#ifdef COMM
    case CHAN :
#endif
    case SET :
    case REL :
    case FF :
#ifdef USERTYPE
    case LABEL:
    case TYPE :
#endif
        val = expr ;
        break ;
#ifdef LAZY
    /* repeated force evaluation of a recipe */
    case REC :              
        m2rec.n_ptr = expr ;
        do {
            m2oldenv = xlenv ;
            for( m2p = m2rec.n_ptr->n_recenv ; m2p != NULL ;
                                             m2p = m2p->n_listnext )
                xlbind( m2p->n_listvalue->n_listvalue,
                        m2p->n_listvalue->n_listnext->n_listvalue ) ;
            xlfixbindings( m2oldenv ) ;
            m2val = xleval( m2rec.n_ptr->n_recexp ) ;
            xlunbind( m2oldenv ) ;
            m2rec.n_ptr = m2val ;
        } while( (m2rec.n_ptr != NULL) && (m2rec.n_ptr->n_type == REC) ) ;
        val = m2rec.n_ptr ;
        break ;
#endif
    default :
#ifdef USERTYPE
        if( expr->n_type >= USERDEFTYPE) { val = expr; }
#endif
        xlfail( "can't evaluate expression" ) ; }

    /* restore the previous stack frame */
    xlstack = oldstk ;

    return( val ) ;
}


/*. evlist - evaluate a list .*/
struct node *evlist( nptr )
struct node *nptr ;
{
    struct node fun, args, fname, m2arg ;
    struct node *oldstk, *val ;
    int tol_fl, arg_c ;

    /* create a stack frame */
    oldstk = xlsave( &fun, &args, &fname,  &m2arg, NULL ) ;

    /* get the function and the argument list */
    fun.n_ptr = nptr->n_listvalue ;
    args.n_ptr = nptr->n_listnext ;

    /* get the function name to allow undefinition */
    if( (fun.n_ptr != NULL) &&
        (fun.n_ptr->n_type == SYM) && (fun.n_ptr->n_symvalue == NULL) ) {
        fname.n_ptr = fun.n_ptr ;
        tol_fl = TRUE ; }
    else 
        tol_fl = FALSE ;

    /* add trace entry */
    tpush( nptr ) ;

    /* print traced expression (if "tr_flag" is on) */
    if( tr_flag ) {
        printf( "\n%d - ", trace_pointer ) ;
        xlprint( nptr, TRUE ) ;
        printf( "\n" ) ; }

    /* evaluate the first expression */
    if( (fun.n_ptr = xleval(fun.n_ptr)) == NULL)
        if( tol_fl && ! neverprompt ) {
            printf( "\nundefined function %s :", fname.n_ptr->n_symname ) ;
            for( arg_c = 1; args.n_ptr != NULL ; arg_c++ ) {
                printf( " \n    argument %d of %s : ", 
                        arg_c, fname.n_ptr->n_symname ) ;    
#ifdef USERTYPE
                xlprint( userevarg( &args.n_ptr ), TRUE ) ; 
#else
                xlprint( xlevarg( &args.n_ptr ), TRUE ) ; 
#endif
            }
            printf( "\nresult for %s : ", fname.n_ptr->n_symname ) ;
            m2arg.n_ptr = newnode( LIST ) ;
            m2arg.n_ptr->n_listvalue = newnode( FPTR ) ;
            m2arg.n_ptr->n_listvalue->n_fp = stdin ;
            val = m2read( m2arg.n_ptr ) ; }
        else
            xlfail( "null function" ) ;

    /* evaluate the function */
    if( !tol_fl ) 
        switch( fun.n_ptr->n_type ) {
        case SUBR :
            val = (*fun.n_ptr->n_subr)(args.n_ptr) ;
            break ;
        case FUN :
            val = evfun( fun.n_ptr, args.n_ptr ) ;
            break ;
#ifdef USERTYPE
        case LABEL:
            val = xmselect( fun.n_ptr->n_symname, args.n_ptr ) ;
            break ;
        case TYPE:
            val = construct( fun.n_ptr->n_typedef, args.n_ptr ) ;
            break ;
#endif
        default :
            xlfail( "bad function" ) ; } 

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* remove trace entry */
    tpop() ;

    /* return the result value */
    return( val ) ;
}


/*. xlabind - bind the arguments for a function .*/
static void xlabind( fargs, aargs )
struct node *fargs, *aargs ;
{
    struct node farg, aarg, val ; 
    struct node *oldstk ;

    /* create a stack frame */
    oldstk = xlsave( &farg, &aarg, &val, NULL ) ;

    /* initialize the pointers */
    farg.n_ptr = fargs ;
    aarg.n_ptr = aargs ;

    /* evaluate and bind each argument */
    while( (farg.n_ptr != NULL) && (aarg.n_ptr != NULL) ) {
#ifdef LAZY
        /* create a recipe for each argument */
        val.n_ptr = delay( aarg.n_ptr->n_listvalue ) ;
        aarg.n_ptr = aarg.n_ptr->n_listnext ;
#else 
        /* evaluate the argument */
        val.n_ptr = xlevarg( &aarg.n_ptr ) ;
#endif

        /* bind the formal variable to the argument value */
        xlbind( farg.n_ptr->n_listvalue, val.n_ptr ) ;

        /* move the formal argument list pointer ahead */
        farg.n_ptr = farg.n_ptr->n_listnext ; }

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* make sure the correct number of arguments were supplied */
    if( farg.n_ptr != aarg.n_ptr )
        xlfail( "incorrect number of arguments to a function" ) ;
}


/*. evfun - evaluate a function .*/
struct node *evfun( fun, args )
struct node *fun, *args ;
{
    struct node cptr ;
    struct node *oldenv, *oldstk, *fargs, *val, *m2p, *m2new ;

    /* create a stack frame */
    oldstk = xlsave( &cptr, NULL ) ;

    /* get the formal argument list */
    if( ((fargs = fun->n_listvalue) != NULL) && (fargs->n_type != LIST) )
        xlfail( "bad formal argument list" ) ;

    /* bind the formal parameters */
    oldenv = xlenv ;
    xlabind( fargs, args ) ;
    xlfixbindings( oldenv ) ;

    /* create a new "lbda_env" frame */
    m2new = newnode( LIST ) ;
    m2new->n_listnext = lbda_env.n_ptr ;
    lbda_env.n_ptr = m2new ; 
    m2new->n_listvalue = newnode( LBDA_FRAME) ;

    /* PUSH formal parameters into "lbda_env" */
    for( m2p = fargs ; m2p != NULL ; m2p = m2p->n_listnext ) {
        m2new = newnode( LIST ) ;
        m2new->n_listnext = lbda_env.n_ptr ;
        lbda_env.n_ptr = m2new ;
        m2new = m2new->n_listvalue = newnode( LBDA_MARK ) ;
        m2new->n_symname = strsave( m2p->n_listvalue->n_symname ) ;
        m2new->n_symvalue = m2p->n_listvalue->n_symvalue ; }

    /* execute the code */
    for( cptr.n_ptr = fun->n_listnext ; cptr.n_ptr != NULL ; )
	val = xlevarg( &cptr.n_ptr ) ;

    /* restore the previous "lbda_env" */
    for( m2p = lbda_env.n_ptr ; m2p->n_listvalue->n_type != LBDA_FRAME ;
                                                   m2p = m2p->n_listnext )
        m2p->n_listvalue->n_type = SYM ;
    lbda_env.n_ptr = m2p->n_listnext ;
        
    /* restore the environment */
    xlunbind( oldenv ) ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return the result value */
    return( val ) ;
}


/*. quote - builtin function quote .*/
/*. quote : exp -> val .*/
static struct node *quote( args )
struct node *args ;
{
    struct node *arg ;

    arg = xlarg( &args ) ;
    xllastarg( args ) ;

    /* return the quoted expression */
    return( arg ) ;
}


/*. trace - builtin function trace .*/
/*. changes trace mode complementing "tr_flag" .*/
/* trace : --> nil */
static struct node *trace( args ) 
struct node *args ;
{
    xllastarg( args ) ;
    tr_flag = !tr_flag ;
    return( NULL ) ;
} 


/*. xmevalinit - initialize the evaluator .*/
void xmevalinit()
{
    /* initialize debugging stuff */
    trace_pointer = -1 ;

    /* builtin functions */
    xlsubr( "eval", eval ) ;
    xlsubr( "quote", quote ) ;
    xlsubr( "trace", trace ) ;
}

