/* xmff.c - finite function algebra */

#include <stdio.h>
#include "xmetoo.h"


/* external variables */
extern struct node *xlstack ;
extern struct node *xlenv ;
extern struct node *true ;
extern struct node *false ;


/* external procedures */
extern void xlbind() ;
extern void jjxlfail() ;
extern void xlunbind() ;
extern void xlfixbindings() ;
extern int xequal() ;


/*. makeff - builtin function makeff .*/
/*. makeff : pair... -> ff .*/
/*.          pair = "(" exp exp ")" .*/
static struct node *makeff( args )
struct node *args ;
{
    struct node arg, e1, e2, resf, pair ;
    struct node *oldstk, *lresf, *ract, *new ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &e1, &e2, &resf, &pair, NULL ) ;

    arg.n_ptr = args ;

    lresf = NULL ;
    while( args != NULL ) {
        pair.n_ptr = xlmatch( LIST, &args ) ;
        e1.n_ptr = xlevarg( &pair.n_ptr ) ;
        e2.n_ptr = xlevarg( &pair.n_ptr ) ;
        xllastarg( pair.n_ptr ) ;
        for( ract = resf.n_ptr ; ract != NULL ; ract = ract->n_listnext )
            if( xequal( e1.n_ptr, ract->n_listvalue->n_listvalue ) )
               { jjxlfail( "invalid argument (duplicated key in mapping)", 
                           ract->n_listvalue->n_listvalue ) ;
               }
        new = newnode( LIST ) ;
        if( lresf == NULL ) 
            resf.n_ptr = new ;
        else
            lresf->n_listnext = new ;
        lresf = new ;
        lresf->n_listvalue = new = newnode( LIST ) ;
        new->n_listvalue = e1.n_ptr ;
        new->n_listnext = newnode( LIST ) ;
        new->n_listnext->n_listvalue = e2.n_ptr ; }

    /* restore the old stack frame */
    xlstack = oldstk ;

    if( resf.n_ptr != NULL ) resf.n_ptr->n_type = FF ;

    /* return the created ff */
    return( resf.n_ptr ) ;
}


/*. dom - builtin function dom .*/
/*. dom : expff -> set .*/
/*. (implemented by builtin function "dom" with rel argument) .*/


/*. ran - builtin function ran .*/
/*. ran : expff -> set .*/
/*. (implemented by builtin function "ran" with rel argument) .*/


/*. ds - builtin function ds .*/
/*. ds : expff expset -> ff .*/
/*. (implemented by builtin function "ds" with rel argument) .*/


/*. dr - builtin function dr .*/
/*. dr : expff expset -> ff .*/
/*. (implemented by builtin function "dr" with rel argument) .*/


/*. ff - builtin function ff .*/
/*. ff : symb expset nexp [nexpbool] -> ff .*/
static struct node *ff( args ) 
struct node *args ;
{
    struct node arg, set, val, p, resf ;
    struct node *oldstk, *oldenv, *sym, *new, *lresf, *exp, *bool ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &set, &val, &p, &resf, NULL ) ;

    arg.n_ptr = args ;
    sym = xlmatch( SYM, &args ) ;
    set.n_ptr = xlevmatch( SET, &args ) ;
    exp = xlarg( &args ) ;
    p.n_ptr = (args == NULL) ? true : xlarg( &args ) ;
    xllastarg( args ) ;
    
    /* ff construction */
    oldenv = xlenv ; lresf = NULL ;
    while( set.n_ptr != NULL ) {
        xlbind( sym, set.n_ptr->n_listvalue ) ;
        xlfixbindings( oldenv ) ;
        bool = xleval( p.n_ptr ) ;
        if( (bool != true) && (bool != false) ) xlfail( "bad argument type" ) ;
        if( bool == true ) {
            val.n_ptr = xleval( exp ) ;
            new = newnode( LIST ) ;
            if( lresf == NULL ) 
                resf.n_ptr = new ;
            else
                lresf->n_listnext = new ;
            lresf = new ;
            lresf->n_listvalue = new = newnode( LIST ) ;
            new->n_listvalue = sym->n_symvalue ;
            new = new->n_listnext = newnode( LIST ) ;
            new->n_listvalue = val.n_ptr ;
            new->n_listnext = NULL ; }
        xlunbind( oldenv ) ;
        set.n_ptr = set.n_ptr->n_listnext ; }

    /* restore the previous stack frame */
    xlstack = oldstk ;

    if( resf.n_ptr != NULL ) resf.n_ptr->n_type = FF ;

    /* return the ff */
    return( resf.n_ptr ) ;
}

/*. fffilter - removes duplicate elements from argument list .*/
static struct node *fffilter( list )              /* em obras */
struct node *list ;
{
    struct node *act, *prev, *actel ;

    if( list != NULL ) list->n_type = FF ;

    for( actel = list ; actel != NULL ; actel = actel->n_listnext )
        for( prev = actel, act = actel->n_listnext ; act != NULL ;
                                            act = act->n_listnext )
            if( xequal( actel->n_listvalue->n_listvalue, 
                          act->n_listvalue->n_listvalue ) )
                prev->n_listnext = act->n_listnext ;
            else
                prev = act ;

    return( list ) ;
}

/*. ff1 - builtin function ff .*/
/*. ff1 : symb expset nexp [nexpbool] -> ff .*/
static struct node *ff1( args ) 
struct node *args ;
{
    return( fffilter( sequence( args ) ) ) ;
}

/*. ap - builtin function ap .*/
/*. ap : expff exp [exp] -> value .*/
static struct node *ap( args ) 
struct node *args ;
{
    struct node arg, f, e1, e2, list ;
    struct node *oldstk, *actf, *val ;
    int found ,n;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &f, &e1, &e2, NULL ) ;

    arg.n_ptr = args ;
/*    f.n_ptr = xlevmatch( FF, &arg.n_ptr ) ; */
    f.n_ptr = xlevarg(  &arg.n_ptr ) ;  /* jj */
    if(f.n_ptr == NULL || f.n_ptr->n_type == FF) {
    e1.n_ptr = xlevarg( &arg.n_ptr ) ;
    if( arg.n_ptr == NULL ) 
        e2.n_ptr = NULL ;
    else {
        e2.n_ptr = xlarg( &arg.n_ptr ) ;
        xllastarg( arg.n_ptr ) ; }
    
    for( actf = f.n_ptr, found = FALSE ; (actf != NULL) && !found ;
                                         actf = actf->n_listnext ) {
        val = actf->n_listvalue ;
        if( xequal( val->n_listvalue, e1.n_ptr ) ) {
            e2.n_ptr = val->n_listnext->n_listvalue ;
            found = TRUE ; } }
    if( !found ) e2.n_ptr = xleval( e2.n_ptr ) ;
    }
    else if(f.n_ptr->n_type == LIST) {
       /* get n */
      if( (n = xlevmatch( INT, &arg.n_ptr )->n_int) < 1 )
          xlfail( "invalid argument" ) ;
  
      /* make sure this is the only argument */
      xllastarg( arg.n_ptr ) ;
  
      /* find the nth element */
      for( ; n > 1 ; n--, f.n_ptr = f.n_ptr->n_listnext )
          if( f.n_ptr == NULL ) xlfail( "invalid argument" ) ;
      if( f.n_ptr == NULL ) xlfail( "invalid argument" ) ;
      e2.n_ptr = f.n_ptr->n_listvalue;

    } 
    else xlfail( "invalid argument" ) ;

    /* restore the previus stack frame */
    xlstack = oldstk ;

    /* return the result */
    return( e2.n_ptr ) ;
}


/*. plus - builtin function plus .*/
/*. plus : expff expff -> ff .*/
/*. plus : expseq expff -> seq .*/
static struct node *plus( args ) 
struct node *args ;
{
    struct node arg, f1, f2, resf ;
    struct node *oldstk, *actf, *lresf, *new, *aux, *ar ;
    struct node *ll, *ff, *pair, *lp, *act,  *ind, *val ;
    int i, c ,found ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &f1, &f2, &resf, NULL ) ;

    arg.n_ptr = args ;
    /*f1.n_ptr = xlevmatch( FF, &arg.n_ptr ) ;*/
    f1.n_ptr = xlevarg(&arg.n_ptr ) ;
  if(f1.n_ptr == NULL || f1.n_ptr->n_type == FF) {
    f2.n_ptr = xlevmatch( FF, &arg.n_ptr ) ;
    xllastarg( arg.n_ptr ) ;

    for( actf = f1.n_ptr, lresf = NULL ; actf != NULL ;
                                         actf = actf->n_listnext ) {
        aux = actf->n_listvalue ;
        new = newnode( LIST ) ;
        if( lresf == NULL )
            resf.n_ptr = new ;
        else
            lresf->n_listnext = new ;
        lresf = new ;
        new = new->n_listvalue = newnode( LIST ) ;
        new->n_listvalue = aux->n_listvalue ;
        aux = aux->n_listnext ;
        new->n_listnext = newnode( LIST ) ;
        new->n_listnext->n_listvalue = aux->n_listvalue ; }

    for( actf = f2.n_ptr ; actf != NULL ; actf = actf->n_listnext ) {
        aux = actf->n_listvalue->n_listnext ;
        for( ar = resf.n_ptr, found = FALSE ; !found && (ar != NULL) ;
                                                 ar = ar->n_listnext )
            if( xequal( ar->n_listvalue->n_listvalue,
                        actf->n_listvalue->n_listvalue ) ) {
                found = TRUE ;
                ar->n_listvalue->n_listnext->n_listvalue =
                            actf->n_listvalue->n_listnext->n_listvalue ; 
            }
        if( !found ) {
            new = newnode( LIST ) ;
            if( lresf == NULL )
                resf.n_ptr = new ;
            else
                lresf->n_listnext = new ;
            lresf = new ;
            new = new->n_listvalue = newnode( LIST ) ;
            new->n_listvalue = actf->n_listvalue->n_listvalue ;
            new->n_listnext = newnode( LIST ) ;
            new->n_listnext->n_listvalue = aux->n_listvalue ; } }

    if( resf.n_ptr != NULL ) resf.n_ptr->n_type = FF ;

  } else  if(f1.n_ptr->n_type == LIST) {

   f2.n_ptr = xlevmatch( FF, &arg.n_ptr ) ;
    xllastarg( arg.n_ptr ) ;

    /* make a copy of list "f1" */
    for( ll = NULL, act = f1.n_ptr ; act != NULL ;
                              ll = new, act = act->n_listnext ) {
        new = newnode( LIST ) ;
        new->n_listvalue = act->n_listvalue ;
        if( ll == NULL )
            resf.n_ptr = new ;
        else
            ll->n_listnext = new ; }

    /* substitute the elements of "resf" found in "f2" domain */
    for( ff = f2.n_ptr ;  ff != NULL ; ) {
        pair = xlarg( &ff ) ;
        ind = xlmatch( INT, &pair ) ;
        i = ind->n_int ;
        val = xlarg( &pair ) ;
        for( c = 1, lp = resf.n_ptr ; c != i ; c++, lp = lp->n_listnext )
            if( lp == NULL ) xlfail( "list index out of range" ) ;
        if( lp == NULL ) xlfail( "list index out of range" ) ;
        lp->n_listvalue = val ; }

  } else xlfail("invalid argument");
    /* restore the previous stack frame */
    xlstack = oldstk ;


    /* return pointer to the ff */
    return( resf.n_ptr ) ;
}


void xmffinit()
{
    xlsubr( "makeff", makeff ) ;
    xlsubr( "ff", ff ) ;
    xlsubr( "ff1", ff1 ) ;
    xlsubr( "ap", ap ) ;
    xlsubr( "plus", plus ) ;
}

