/* xmfio.c - xmetoo file i/o */

#include <stdio.h>
#include "xmetoo.h"

#ifdef XM95
/* nao percebi muito p=bem porque??? jj */
#define popen _popen
#define pclose _pclose
#endif

/* external variables */
extern struct node *xlstack ;


/* external procedures */
extern char *strcpy() ;
extern char *strcat() ;

/* local variables */
struct node *s ;

/* xlfopen - builtin function fopen */
/* fopen : expstr expstr -> fptr | nil */
static struct node *xlfopen( args )
struct node *args ;
{
    struct node arg, fname, mode ;
    struct node *oldstk, *val ;
    FILE *fp ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &fname, &mode, NULL ) ;

    /* initialize */
    arg.n_ptr = args ;

    /* get the file name */
    fname.n_ptr = xlevmatch( STR, &arg.n_ptr ) ;

    /* get the mode */
    mode.n_ptr = xlevmatch( STR, &arg.n_ptr ) ;

    /* make sure there aren't any more arguments */
    xllastarg( arg.n_ptr ) ;

    /* try to open the file */
    if( (fp = fopen( fname.n_ptr->n_str, mode.n_ptr->n_str )) != NULL ) {
        val = newnode( FPTR ) ;
        val->n_fp = fp ; }
    else
	val = NULL ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return the file pointer */
    return( val ) ;
}

/* xlpopen - builtin function popen */
/* popen : expstr expstr -> fptr | nil */
static struct node *xlpopen( args )
struct node *args ;
{
    struct node arg, fname, mode ;
    struct node *oldstk, *val ;
    FILE *fp ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &fname, &mode, NULL ) ;

    /* initialize */
    arg.n_ptr = args ;

    /* get the file name */
    fname.n_ptr = xlevmatch( STR, &arg.n_ptr ) ;

    /* get the mode */
    mode.n_ptr = xlevmatch( STR, &arg.n_ptr ) ;

    /* make sure there aren't any more arguments */
    xllastarg( arg.n_ptr ) ;

    /* try to open the file */
    if( (fp = popen( fname.n_ptr->n_str, mode.n_ptr->n_str )) != NULL ) {
        val = newnode( FPTR ) ;
        val->n_fp = fp ; }
    else
	val = NULL ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return the file pointer */
    return( val ) ;
}



/* xlfclose - builtin function fclose */
/* fclose : expfptr -> nil */
static struct node *xlfclose( args )
struct node *args ;
{
    struct node *fptr ;

    /* get file pointer */
    fptr = xlevmatch( FPTR, &args ) ;

    /* make sure there aren't any more arguments */
    xllastarg( args ) ;

    /* make sure the file exists */
    if( fptr->n_fp == NULL )
        xlfail( "file not open" ) ;

    /* close the file */
    if( fclose( fptr->n_fp ) ) xlfail( "error when closing file" ) ;
    fptr->n_fp = NULL ;

    /* return nil */
    return( NULL ) ;
}

/* xlpclose - builtin function pclose */
/* pclose : expfptr -> nil */
static struct node *xlpclose( args )
struct node *args ;
{
    struct node *fptr ;

    /* get file pointer */
    fptr = xlevmatch( FPTR, &args ) ;

    /* make sure there aren't any more arguments */
    xllastarg( args ) ;

    /* make sure the file exists */
    if( fptr->n_fp == NULL )
        xlfail( "pipe not open" ) ;

    /* close the file */
    if( pclose( fptr->n_fp ) ) xlfail( "error when closing pipe" ) ;
    fptr->n_fp = NULL ;

    /* return nil */
    return( NULL ) ;
}


/* xlgetc - builtin function getc */
/* getc : [expfptr] -> int | nil */
static struct node *xlgetc( args )
struct node *args ;
{
    struct node *val ;
    FILE *fp ;
    int ch ;

    /* get file pointer */
    if( args != NULL )
        fp = xlevmatch( FPTR, &args )->n_fp ;
    else
        fp = stdin ;

    /* make sure there aren't any more arguments */
    xllastarg( args ) ;

    /* make sure the file exists */
    if( fp == NULL )
        xlfail( "file not open" ) ;

    /* get character and check for eof */
    ch = fgetc( fp ) ;
    if( !feof( fp ) && ferror( fp ) ) xlfail( "read error" ) ; 

    if( ch != EOF ) {
        /* create return node */
        val = newnode( INT ) ;
        val->n_int = ch ; }
    else
        val = NULL ;

    /* return the character */
    return( val ) ;
}


/* xlputc - builtin function putc */
/* putc : expint [expfptr] -> int */
static struct node *xlputc( args )
struct node *args ;
{
    struct node arg, chr ;
    struct node *oldstk ;
    FILE *fp ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &chr, NULL ) ;

    /* initialize */
    arg.n_ptr = args ;

    /* get the character */
    chr.n_ptr = xlevmatch( INT, &arg.n_ptr ) ;

    /* get file pointer */
    if( arg.n_ptr != NULL )
        fp = xlevmatch( FPTR, &arg.n_ptr )->n_fp ;
    else
        fp = stdout ;

    /* make sure there aren't any more arguments */
    xllastarg( arg.n_ptr ) ;

    /* make sure the file exists */
    if( fp == NULL )
        xlfail( "file not open" ) ;

    /* put character to the file */
    fputc( chr.n_ptr->n_int, fp ) ;
    if( ferror( fp ) ) xlfail( "write error" ) ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return the character */
    return( chr.n_ptr ) ;
}


/* xlfgets - builtin function fgets */
/* fgets : [expfptr] -> str | nil */
static struct node *xlfgets( args )
struct node *args ;
{
    struct node str ;
    struct node *oldstk ;
    char buf[STRMAX+1] ;
    FILE *fp ;

    /* create a new stack frame */
    oldstk = xlsave( &str, NULL ) ;

    /* get file pointer */
    if( args != NULL )
        fp = xlevmatch( FPTR, &args )->n_fp ;
    else
        fp = stdin ;

    /* make sure there aren't any more arguments */
    xllastarg( args ) ;

    /* make sure the file exists */
    if( fp == NULL )
        xlfail( "file not open" ) ;

    /* get character and check for eof */
    if( fgets( buf, STRMAX, fp ) != NULL ) {
        /* create return node */
        str.n_ptr = newnode( STR ) ;
        str.n_ptr->n_str = strsave( buf ) ; }
    else
        str.n_ptr = NULL ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return the string */
    return( str.n_ptr ) ;
}


/* xlfputs - builtin function fputs */
/* fputs : expstr [expfptr] -> str */
static struct node *xlfputs( args )
struct node *args ;
{
    struct node arg, str ;
    struct node *oldstk ;
    FILE *fp ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &str, NULL ) ;

    /* initialize */
    arg.n_ptr = args ;

    /* get the string */
    str.n_ptr = xlevmatch( STR, &arg.n_ptr ) ;

    /* get file pointer */
    if( arg.n_ptr != NULL )
        fp = xlevmatch( FPTR, &arg.n_ptr )->n_fp ;
    else
        fp = stdout ;

    /* make sure there aren't any more arguments */
    xllastarg( arg.n_ptr ) ;

    /* make sure the file exists */
    if( fp == NULL )
        xlfail( "file not open" ) ;

    /* put string to the file */
#ifdef XM95
    if( fputs( str.n_ptr->n_str, fp ) == EOF ) xlfail( "write error" ) ;
#else
    if( fputs( str.n_ptr->n_str, fp ) == NULL ) xlfail( "write error" ) ;
#endif

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return the string */
    return( str.n_ptr ) ;
}


/* xlfseek - builtin function fseek */
/* fseek : expint [expfptr] -> int */
static struct node *xlfseek( args )
struct node *args ;
{
    struct node arg, chr ;
    struct node *oldstk ;
    FILE *fp ;
    int offset, seekset;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &chr, NULL ) ;

    /* initialize */
    arg.n_ptr = args ;

    /* get file pointer */
    fp = xlevmatch( FPTR, &arg.n_ptr )->n_fp ;

    /* get the offset */
    offset =  xlevmatch( INT, &arg.n_ptr )->n_int ;

    /* get the offset */
    seekset =  xlevmatch( INT, &arg.n_ptr )->n_int ;

    /* make sure there aren't any more arguments */
    xllastarg( arg.n_ptr ) ;

    /* make sure the file exists */
    if( fp == NULL )
        xlfail( "file not open" ) ;

    /* put character to the file */
    fseek(fp, (long) offset , seekset);
    if( ferror( fp ) ) xlfail( "seek error" ) ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return the character */
    return( NULL ) ;
}


/* xmfioinit - initialize file stuff */
void xmfioinit()
{
	/* stdin, stdout reserved FPTR nodes */
	s = xlenter( "stdin" ) ;
	s->n_symvalue = newnode( FPTR ) ;
	s->n_symvalue->n_fp = stdin ;
	s = xlenter( "stdout" ) ;
	s->n_symvalue = newnode( FPTR ) ;
	s->n_symvalue->n_fp = stdout ;
    xlsubr( "fopen", xlfopen ) ;
    xlsubr( "fclose", xlfclose ) ;
    xlsubr( "popen", xlpopen ) ;
    xlsubr( "pclose", xlpclose ) ;
    xlsubr( "getc", xlgetc ) ;
    xlsubr( "putc", xlputc ) ;
    xlsubr( "fgets", xlfgets ) ;
    xlsubr( "fputs", xlfputs ) ;
    xlsubr( "fseek", xlfseek ) ;
}

