/* xmint.c - integer algebra */

#include <stdio.h>
#include <sys/types.h>
#include <time.h>
#include <limits.h>
#include "xmetoo.h"

/* #if XM95 || LINUX */
#define rand_r rand
/* #endif            */


/* external variables */
extern struct node *xlstack ;


/* external procedures */
extern int strcmp() ;


/*. add - builtin function add .*/
/*. add : expint expint -> int .*/
static struct node *add( args ) 
struct node *args ;
{
    struct node arg, val1, val2 ;
    struct node *res, *oldstk ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &val1, &val2, NULL ) ;

    arg.n_ptr = args ;
    val1.n_ptr = xlevmatch( INT, &arg.n_ptr ) ;
    val2.n_ptr = xlevmatch( INT, &arg.n_ptr ) ;
    xllastarg( arg.n_ptr ) ;

    res = newnode( INT ) ;
    res->n_int = val1.n_ptr->n_int + val2.n_ptr->n_int ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return pointer to the result */
    return( res ) ;
}
 

/*. sub - builtin function sub .*/
/*. sub : expint expint -> int .*/
static struct node *sub( args ) 
struct node *args ;
{
    struct node arg, val1, val2 ;
    struct node *res, *oldstk ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &val1, &val2, NULL ) ;

    arg.n_ptr = args ;
    val1.n_ptr = xlevmatch( INT, &arg.n_ptr ) ;
    val2.n_ptr = xlevmatch( INT, &arg.n_ptr ) ;
    xllastarg( arg.n_ptr ) ;

    res = newnode( INT ) ;
    res->n_int = val1.n_ptr->n_int - val2.n_ptr->n_int ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return pointer to the result */
    return( res ) ;
}
 

/*. mul - builtin function mul .*/
/*. mul : expint expint -> int .*/
static struct node *mul( args ) 
struct node *args ;
{
    struct node arg, val1, val2 ;
    struct node *res, *oldstk ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &val1, &val2, NULL ) ;

    arg.n_ptr = args ;
    val1.n_ptr = xlevmatch( INT, &arg.n_ptr ) ;
    val2.n_ptr = xlevmatch( INT, &arg.n_ptr ) ;
    xllastarg( arg.n_ptr ) ;

    res = newnode( INT ) ;
    res->n_int = val1.n_ptr->n_int * val2.n_ptr->n_int ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return pointer to the result */
    return( res ) ;
}
 

/*. div - builtin function div .*/
/*. div : expint expint -> int .*/
static struct node *div( args ) 
struct node *args ;
{
    struct node arg, val1, val2 ;
    struct node *res, *oldstk ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &val1, &val2, NULL ) ;

    arg.n_ptr = args ;
    val1.n_ptr = xlevmatch( INT, &arg.n_ptr ) ;
    val2.n_ptr = xlevmatch( INT, &arg.n_ptr ) ;
    xllastarg( arg.n_ptr ) ;

    res = newnode( INT ) ;
    if(val2.n_ptr->n_int == 0 ) xlfail("division by zero");
    res->n_int = val1.n_ptr->n_int / val2.n_ptr->n_int ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return pointer to the result */
    return( res ) ;
}
 

/*. rem - builtin function rem .*/
/*. rem : expint expint -> int .*/
static struct node *rem( args ) 
struct node *args ;
{
    struct node arg, val1, val2 ;
    struct node *res, *oldstk ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &val1, &val2, NULL ) ;

    arg.n_ptr = args ;
    val1.n_ptr = xlevmatch( INT, &arg.n_ptr ) ;
    val2.n_ptr = xlevmatch( INT, &arg.n_ptr ) ;
    xllastarg( arg.n_ptr ) ;

    res = newnode( INT ) ;
    res->n_int = val1.n_ptr->n_int % val2.n_ptr->n_int ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return pointer to the result */
    return( res ) ;
}
 

/*. rand - builtin function rand .*/
/*. rand : expint -> int .*/
static struct node *xmrand( args )
struct node *args ;
{
    struct node *rval ;
    static unsigned int seed=0;

    int val ;

    /* evaluate the argument */
    val = xlevmatch( INT, &args)->n_int ;

    /* make sure there aren't any more arguments */
    xllastarg( args ) ;

    /* convert and check the value  */
    rval = newnode( INT ) ;
    if (seed == 0 ) seed = time(0);
    rval->n_int = val == 0 ? 0 : rand_r(&seed) % val;

    /* return the result value */
    return( rval ) ;
}

/*. abs - builtin function abs .*/
/*. abs : expint -> int .*/
static struct node *abs( args )
struct node *args ;
{
    struct node *rval ;
    int val ;

    /* evaluate the argument */
    val = xlevmatch( INT, &args)->n_int ;

    /* make sure there aren't any more arguments */
    xllastarg( args ) ;

    /* convert and check the value  */
    rval = newnode( INT ) ;
    rval->n_int = (val >= 0) ? val : -val ;

    /* return the result value */
    return( rval ) ;
}


/*. min - builtin function min .*/
/*. min : expint expint... -> int .*/
/*.       expstr expstr... -> str .*/
static struct node *xmin( args )
struct node *args ;
{
    struct node arg, res ;
    struct node *oldstk, *val ;
    int tp ;

    oldstk = xlsave( &arg, &res, NULL ) ;
    arg.n_ptr = args ;

    res.n_ptr = xlevarg( &arg.n_ptr ) ;
    if( (res.n_ptr == NULL) ||(((tp = res.n_ptr->n_type) != INT) &&(tp != STR)))
        xlfail( "min bad argument type" ) ;

    while( arg.n_ptr != NULL ) {
        val = xlevarg( &arg.n_ptr ) ;
        if( (val == NULL) || (tp != val->n_type) )
            xlfail( "bad argument type" ) ;
        if( tp == INT ) {
            if( val->n_int < res.n_ptr->n_int )
                res.n_ptr = val ; }
        else
            if( strcmp( val->n_str, res.n_ptr->n_str ) < 0 )
                res.n_ptr = val ; }

    xlstack = oldstk ;
    return( res.n_ptr ) ;
}


/*. max - builtin function max .*/
/*. max : expint expint... -> int .*/
/*.       expstr expstr... -> str .*/
static struct node *xmax( args )
struct node *args ;
{
    struct node arg, res ;
    struct node *oldstk, *val ;
    int tp ;

    oldstk = xlsave( &arg, &res, NULL ) ;
    arg.n_ptr = args ;

    res.n_ptr = xlevarg( &arg.n_ptr ) ;
    if( (res.n_ptr == NULL) ||(((tp = res.n_ptr->n_type) != INT) &&(tp != STR)))
        xlfail( "bad argument type" ) ;

    while( arg.n_ptr != NULL ) {
        val = xlevarg( &arg.n_ptr ) ;
        if( (val == NULL) || (tp != val->n_type) )
            xlfail( "bad argument type" ) ;
        if( tp == INT ) {
            if( val->n_int > res.n_ptr->n_int )
                res.n_ptr = val ; }
        else
            if( strcmp( val->n_str, res.n_ptr->n_str ) > 0 )
                res.n_ptr = val ; }

    xlstack = oldstk ;
    return( res.n_ptr ) ;
}


void xmintinit()
{
    struct node *maxint;
    maxint = xlenter("maxint");    
    maxint-> n_symvalue = (struct node *) intsexp(INT_MAX -1);
    xlsubr( "add", add ) ;
    xlsubr( "sub", sub ) ;
    xlsubr( "mul", mul ) ;
    xlsubr( "div", div ) ;
    xlsubr( "rem", rem ) ;
    xlsubr( "min", xmin ) ;
    xlsubr( "max", xmax ) ;
    xlsubr( "abs", abs ) ;
    xlsubr( "rand", xmrand ) ;
}

