/* xmlist.c - list algebra */

#include <stdio.h>
#include "xmetoo.h"


/* external variables */
extern struct node *xlstack ;
extern struct node *xlenv ;
extern struct node *true ;
extern struct node *false ;


/* external procedures */
extern void xlbind() ;
extern void xlunbind() ;
extern void xlfixbindings() ;
extern char *strcat() ;
extern char *strcpy() ;
extern void subst_name() ;


/* local variables */
static struct node *rac ;
static struct node *rse ;


/*. nil - metoo symbol .*/
/*. (implemented by "nil") .*/


/*. xlist - builtin function list .*/
/*. list : exp... -> list .*/
struct node *xlist( args )
struct node *args ;
{
    struct node arg, list, val ;
    struct node *oldstk, *last, *lptr ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &list, &val, NULL ) ;

    /* initialize */
    arg.n_ptr = args ;

    /* evaluate and append each argument */
    for( last = NULL ; arg.n_ptr != NULL ; last = lptr) {

        /* evaluate the next argument */
        val.n_ptr = xlevarg( &arg.n_ptr ) ;

        /* append this argument to the end of the list */
        lptr = newnode( LIST ) ;
        if( last == NULL )
            list.n_ptr = lptr ;
        else
            last->n_listnext = lptr ;
        lptr->n_listvalue = val.n_ptr ; }

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return the list */
    return( list.n_ptr ) ;
}


/*. makeseq - builtin function makeseq .*/
/*. makeseq : exp... -> list .*/
/*. (implemented by builtin function "list") .*/


/*. append - builtin function append .*/
/*. append : explist explist explist... -> list .*/
static struct node *append ( args )
struct node *args ;
{
    struct node arg, list, last, val ;
    struct node *oldstk, *lptr ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &list, &last, &val, NULL ) ;

    /* initialize */
    arg.n_ptr = args ;

    /* check syntax */
    xlarg( &args ) ;
    xlarg( &args ) ;

    /* evaluate and append each argument */
    while( arg.n_ptr != NULL ) {

        /* evaluate the next argument */
        list.n_ptr = xlevmatch( LIST, &arg.n_ptr ) ;

        /* append each element of this list to the result list */
        while( list.n_ptr != NULL ) {

            /* append this element */
            lptr = newnode( LIST ) ;
            if( last.n_ptr == NULL )
                val.n_ptr = lptr ;
            else
                last.n_ptr->n_listnext = lptr ;
            lptr->n_listvalue = list.n_ptr->n_listvalue ;

	    /* save the new last element */
            last.n_ptr = lptr ;

	    /* move to the next element */
	    list.n_ptr = list.n_ptr->n_listnext ; } }

    /* restore previous stack frame */
    xlstack = oldstk ;

    /* return the list */
    return( val.n_ptr ) ;
}


/*. cons - builtin function cons .*/
/*. cons : exp explist -> list .*/
static struct node *cons( args )
struct node *args ;
{
    struct node arg, new, exp, l ;
    struct node *oldstk ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &new, &exp, &l, NULL ) ;
    arg.n_ptr = args ;

#ifdef LAZY

    exp.n_ptr = xlarg( &arg.n_ptr ) ;
    l.n_ptr = xlarg( &arg.n_ptr ) ;
    xllastarg( arg.n_ptr ) ;
    
    new.n_ptr = newnode( LISTREC ) ;
    new.n_ptr->n_listvalue = delay( exp.n_ptr ) ;
    new.n_ptr->n_listnext = delay( l.n_ptr ) ;

#else

    exp.n_ptr = xlevarg( &arg.n_ptr ) ;
    l.n_ptr = xlevmatch( LIST, &arg.n_ptr ) ;
    xllastarg( arg.n_ptr ) ;

    new.n_ptr = newnode( LIST ) ;
    new.n_ptr->n_listvalue = exp.n_ptr ;
    new.n_ptr->n_listnext = l.n_ptr ;

#endif

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return pointer to the result */
    return( new.n_ptr ) ;
}


/*. hd - builtin function hd .*/
/*. hd : explist -> value .*/
static struct node *hd( args )
struct node *args ;
{
    struct node *arg ;

#ifdef LAZY

    arg = xlarg( &args ) ;
    xllastarg( args ) ;

    if( arg == NULL ) xlfail( "invalid argument" ) ;
    if( arg->n_type == REC ) xlfail( "cannot evaluate list's head" ) ;
    if( arg->n_type == LISTREC ) return( xleval( arg->n_listvalue ) ) ;
    arg = xleval( arg ) ;

    if( arg == NULL) xlfail( "invalid argument" ) ;
    if( arg->n_type == REC ) xlfail( "cannot evaluate list's head" ) ;
    if( arg->n_type == LISTREC ) return( xleval( arg->n_listvalue ) ) ;
    if( arg->n_type != LIST ) xlfail( "bad argument type" ) ;

#else

/* jj e jgr para permitir ir buscar os componentes duma funcao */
      arg = xlevarg(&args);
      if( (arg != NULL) && (arg->n_type != LIST) && (arg->n_type != FUN))
        xlfail( "bad argument type" ) ;

/*    arg = xlevmatch( LIST, &args ) ; */
    xllastarg( args ) ;

    if( arg == NULL) xlfail( "invalid argument" ) ;

#endif

    /* return pointer to the head */
    return( arg->n_listvalue ) ;
}
    

/*. tl - builtin function tl .*/
/*. tl : explist -> list .*/
static struct node *tl( args )
struct node *args ;
{
    struct node *arg ;

#ifdef LAZY

    arg = xlarg( &args ) ;
    xllastarg( args ) ;

    if( arg == NULL ) xlfail( "invalid argument" ) ;
    if( arg->n_type == REC ) xlfail( "cannot evaluate list's tail" ) ;
    if( arg->n_type == LISTREC ) return( arg->n_listnext ) ;
    arg = xleval( arg ) ;

    if( arg == NULL ) xlfail( "invalid argument" ) ;
    if( arg->n_type == REC ) xlfail( "cannot evaluate list's tail" ) ;
    if( arg->n_type == LISTREC ) return( arg->n_listnext ) ;
    if( arg->n_type != LIST ) xlfail( "bad argument type" ) ;

#else

/* jj e jgr para permitir ir buscar os componentes duma funcao */
      arg = xlevarg(&args);
      if( (arg != NULL) && (arg->n_type != LIST) && (arg->n_type != FUN))
        xlfail( "bad argument type" ) ;

/*    arg = xlevmatch( LIST, &args ) ; */
    xllastarg( args ) ;

    if( arg == NULL ) xlfail( "invalid argument" ) ;

#endif

    /* return pointer to the tail */
    return( arg->n_listnext ) ;
}


/*. length - builtin function length .*/
/*. length : explist -> int .*/
static struct node *length( args )
struct node *args ;
{
    struct node *list, *val ;
    int n ;

    /* get the list */
    list = xlevmatch( LIST, &args ) ;

    /* make sure this is the only argument */
    xllastarg( args ) ;

    /* find the length */
    for( n = 0 ; list != NULL ; n++ )
	list = list->n_listnext ;

    /* create the value node */
    val = newnode( INT ) ;
    val->n_int = n ;

    /* return the length */
    return( val ) ;
}


/*. elems - builtin function elems .*/
/*. elems : explist -> set .*/
static struct node *elems( args )
struct node *args ;
{
    struct node *setfilter() ;
    struct node l, resl ;
    struct node *oldstk, *lresl, *act, *new ;

    /* create a new stack frame */
    oldstk = xlsave( &l, &resl, NULL ) ;

    l.n_ptr = xlevmatch( LIST, &args ) ;
    xllastarg( args ) ;

    /* make a copy of list "l" */
    for( lresl = NULL, act = l.n_ptr ; act != NULL ;
                                act = act->n_listnext, lresl = new ) {
        new = newnode( LIST ) ;
        new->n_listvalue = act->n_listvalue ;
        if( lresl == NULL )
            resl.n_ptr = new ;
        else
            lresl->n_listnext = new ; }

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return pointer to the result */
    return( setfilter( resl.n_ptr ) ) ;
}

 
/*. inds - builtin function inds .*/
/*. inds : explist -> int-list .*/
static struct node *inds( args )
struct node *args ;
{
    struct node al, il ;
    struct node *oldstk, *ail, *new ;
    int i ;
          
    /* create new stack frame */
    oldstk = xlsave( &al, &il, NULL ) ;

    al.n_ptr = xlevmatch( LIST, &args ) ;
    xllastarg( args ) ;

    ail = NULL ;
    for( i = 1 ; al.n_ptr != NULL ; i++, al.n_ptr = al.n_ptr->n_listnext ) {
        new = newnode( LIST ) ;
        if( ail == NULL )
            il.n_ptr = new ;
        else
            ail->n_listnext = new ;
        ail = new ;
        new->n_listvalue = newnode( INT ) ;
        new->n_listvalue->n_int = i ; }

     /* restore the previous stack frame */
     xlstack = oldstk ;

     /* return pointer to the "index list" */   
     return( il.n_ptr ) ;
}


/*. nth - builtin function nth .*/
/*. nth : expint explist -> val .*/
struct node *nth( args )
struct node *args ;
{
    struct node arg, list ;
    struct node *oldstk ;
    int n;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &list, NULL ) ;

    /* initialize */
    arg.n_ptr = args ;

    /* get n */
    if( (n = xlevmatch( INT, &arg.n_ptr )->n_int) < 1 )
        xlfail( "invalid argument" ) ;

    /* get the list */
    list.n_ptr = xlevmatch( LIST, &arg.n_ptr ) ;

    /* make sure this is the only argument */
    xllastarg( arg.n_ptr ) ;

    /* find the nth element */
    for( ; n > 1 ; n--, list.n_ptr = list.n_ptr->n_listnext )
        if( list.n_ptr == NULL )
            xlfail( "invalid argument" ) ;
    if( list.n_ptr == NULL ) xlfail( "invalid argument" ) ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return the list nth list element */
    return( list.n_ptr->n_listvalue ) ;
}


/*. el - builtin function el .*/
/*. el : expint explist -> value .*/
/*. (implemented by builtin function "nth") .*/


/*. plusq - builtin function plusq .*/
/*. plusq : explist expff -> list .*/
static struct node *plusq( args ) 
struct node *args ;
{
    struct node arg, arg1, arg2, resl ;
    struct node *oldstk, *ll, *ff, *pair, *lp, *act, *new, *ind, *val ;
    int i, c ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &arg1, &arg2, &resl, NULL ) ;

    arg.n_ptr = args ;
    arg1.n_ptr = xlevmatch( LIST, &args ) ;
    arg2.n_ptr = xlevmatch( FF, &args ) ;
    xllastarg( args ) ;

    /* make a copy of list "arg1" */
    for( ll = NULL, act = arg1.n_ptr ; act != NULL ;
                              ll = new, act = act->n_listnext ) {
        new = newnode( LIST ) ;
        new->n_listvalue = act->n_listvalue ;
        if( ll == NULL )
            resl.n_ptr = new ;
        else
            ll->n_listnext = new ; }

    /* substitute the elements of "resl" found in "arg2" domain */
    for( ff = arg2.n_ptr ;  ff != NULL ; ) {
        pair = xlarg( &ff ) ;
        ind = xlmatch( INT, &pair ) ;
        i = ind->n_int ;
        val = xlarg( &pair ) ;
        for( c = 1, lp = resl.n_ptr ; c != i ; c++, lp = lp->n_listnext )
            if( lp == NULL ) xlfail( "list index out of range" ) ;
        if( lp == NULL ) xlfail( "list index out of range" ) ;
        lp->n_listvalue = val ; }

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return pointer to the result list */
    return( resl.n_ptr ) ;
}


/*. CONC - builtin function CONC .*/ 
/*. CONC : explist -> list .*/
static struct node *CONC( args ) 
struct node *args ;
{
    struct node list, resl ;
    struct node *oldstk, *new, *lresl, *act ;

    /* create a new stack frame */
    oldstk = xlsave( &list, &resl, NULL ) ;

    /* initialize */
    list.n_ptr = xlevmatch( LIST, &args ) ;
    xllastarg( args ) ;

    for( lresl = NULL ; list.n_ptr != NULL ; 
                               list.n_ptr = list.n_ptr->n_listnext) {
        if( (list.n_ptr->n_listvalue != NULL) && 
            (list.n_ptr->n_listvalue->n_type != LIST) )
            xlfail( "invalid argument" ) ;
        for( act = list.n_ptr->n_listvalue ; act != NULL ;
                                  lresl = new, act = act->n_listnext ) {
            /* append this element */
	    new = newnode( LIST ) ;
	    new->n_listvalue = act->n_listvalue ;
	    if( lresl == NULL )
	        resl.n_ptr = new ;
            else
	        lresl->n_listnext = new ; } }

    /* restore previous stack frame */
    xlstack = oldstk ;

    /* return the list */
    return( resl.n_ptr ) ;
}


/*. seq - builtin function seq .*/
/*. seq : nexp fromexp... -> list .*/
/*.       fromexp = "(from" symb (explist | expset) [nexpbool] ")" .*/

struct tab {
            struct node *sym ;     /* pointer to "oblist" symbol */
            struct node *env ;     /* save previous "xlenv" */ 
            struct node *seqtab ;  /* value of seq expression */
            struct node *nextel ;  /* pointer to next seq element */
            struct node *p ;       /* pointer to predicate */
            struct tab *next ;     /* next "from" frame */
            struct tab *prev ;     /* previous "from" frame */
           } ;


/*. enter symbol s in "from" table .*/
static struct tab *tabenter( s, t )
struct node *s ;
struct tab **t ;
{
    struct tab *act, *ant, *new ;
    
    ant = NULL ; act = *t ;
    while( act != NULL ) {
        if( s == act->sym ) return( NULL ) ;
        ant = act ; act = act->next ; }
    new = (struct tab *) xmstralloc( sizeof( struct tab ) - 1 ) ;
    new->sym = s ;
    new->next = NULL ;
    new->prev = ant ;
    if( ant == NULL ) 
        *t = new ;
    else
        ant->next = new ;
    return( new ) ;
}

struct node *sequence( args )
struct node *args ;
{
    struct node arg, lseq, resl, val, par_list ;
    struct node *bool, *exp, *oldstk, *sym, *sym1, *new, *lresl, *lparl ; 
    struct tab *ptab, *level, *act ;
    char name[STRMAX+1] ;
    
    /* initialize table */
    ptab = NULL ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &lseq, &resl, &val, &par_list, NULL ) ;
    
    arg.n_ptr = args ;

    /* get the expression */
    exp = xlarg( &args ) ;

    lparl = NULL ;
    /* check syntax and fill "from" table with dummy variables */
    do {
        val.n_ptr = xlmatch( LIST, &args ) ;
        sym = xlmatch( SYM, &val.n_ptr ) ;
        if( strcmp( sym->n_symname, "from" ) != 0 )
            xlfail( "invalid argument" ) ; 
        sym = xlmatch( SYM, &val.n_ptr ) ;

        if( *(sym->n_symname) != ' ' ) {
            *name = ' ' ; *(name+1) = '\0' ;
            strcat( name, sym->n_symname ) ;
            sym1 = sym ;
            sym = xlenter( name ) ;  }
        else
            sym1 = NULL ;

        if( (act = tabenter( sym, &ptab )) == NULL )
            xlfail( "invalid argument" ) ;

        if( sym1 != NULL ) {
            if( val.n_ptr == NULL )
                 xlfail( "too few arguments" ) ;
            subst_name( &par_list.n_ptr, &(val.n_ptr->n_listvalue) ) ; 
            new = newnode( LIST ) ;
            if( lparl == NULL )
                par_list.n_ptr = new ;
            else
                lparl->n_listnext= new ;
            lparl = new ;
            new->n_listvalue = sym1 ; }

        new = newnode( LIST ) ;
        new->n_listnext = lseq.n_ptr ;
        lseq.n_ptr = new ;
        act->seqtab = xlevarg( &val.n_ptr ) ;  
        if( (act->seqtab != NULL) && (act->seqtab->n_type != LIST) &&
            (act->seqtab->n_type != REL) && (act->seqtab->n_type != SET) )
            xlfail( "bad argument type" ) ;
        act->nextel = act->seqtab ;
        lseq.n_ptr->n_listvalue = act->seqtab ;

        act->p = (val.n_ptr == NULL) ? true : xlarg( &val.n_ptr ) ;
        xllastarg( val.n_ptr ) ;

        if( sym1 != NULL ) 
            subst_name( &par_list.n_ptr, &act->p ) ;

    } while( args != NULL ) ;

    if( sym1 != NULL )
        subst_name( &par_list.n_ptr, &exp ) ;

    /* evaluate "from" expressions */
    ptab->env = xlenv ;
    level = ptab ;
    lresl = NULL ;
    while( level != NULL ) {
        if( level->nextel == NULL ) {
            level->nextel = level->seqtab ;
            level = level->prev ;
            if( level != NULL ) xlunbind( level->env ) ; }
        else {
            level->env = xlenv ;
            xlbind( level->sym, level->nextel->n_listvalue ) ;
            xlfixbindings( level->env ) ;
            level->nextel = level->nextel->n_listnext ;
            bool = xleval( level->p ) ;
            if( (bool != true) && (bool != false) )
                xlfail( "bad argument type" ) ;
            if( bool == true )
                if( level->next == NULL ) {
                    val.n_ptr = xleval( exp ) ;
                    new = newnode( LIST ) ;
                    new->n_listvalue = val.n_ptr ;
                    new->n_listnext = NULL ;
                    if( lresl == NULL ) 
                        resl.n_ptr = new ;
                    else
                        lresl->n_listnext = new ;
                    lresl = new ;
                    xlunbind( level->env ) ; }
                else
                    level = level->next ;
            else
                xlunbind( level->env ) ; } }
    xlunbind( ptab->env ) ;
                           
    /* free "from" table memory */
    for( act = ptab ; act != NULL ; ) {
        ptab = act ; act = act->next ; strfree( (char*)ptab ) ;  }

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return pointer to the created seq */
   return( resl.n_ptr ) ;
}


/*. reverse_xm - reverses the argument list/set .*/
static struct node *reverse_xm( ls )
struct node *ls ;
{
    struct node resls ;
    struct node *oldstk, *new ;

    /* create a new stack frame */
    oldstk = xlsave( &resls, NULL ) ;

    while( ls != NULL) {
	new = newnode( LIST ) ;
	new->n_listvalue = ls->n_listvalue ;
        new->n_listnext = resls.n_ptr ;
        resls.n_ptr = new ;
	ls = ls->n_listnext ; }

    /* restore previous stack frame */
    xlstack = oldstk ;

    /* return the list */
    return( resls.n_ptr ) ;
}


/*. reverse - builtin function reverse .*/
/*. reverse - explist -> list .*/
static struct node *reverse( args )
struct node *args ;
{
    struct node l ;
    struct node *oldstk ;

    oldstk = xlsave( &l, NULL ) ;
    l.n_ptr = xlevmatch( LIST, &args ) ;
    xllastarg( args ) ;

    l.n_ptr = reverse_xm( l.n_ptr ) ;
    xlstack = oldstk ;

    return( l.n_ptr ) ;
}


/*. reduce - builtin function reduce with list argument .*/
/*. reduce : exp symb explist -> value .*/
static struct node *reduce( args )
struct node *args ;
{
    struct node arg, exp, s, acc ;
    struct node *oldstk, *fun, *new, *oldenv, *s1 ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &exp, &s, &acc, NULL ) ;

    arg.n_ptr = args ;
    acc.n_ptr = xlevarg( &args ) ;   
    fun = xlmatch( SYM, &args ) ;
    s.n_ptr = xlevarg( &args ) ;
    xllastarg( args ) ;

    if( (s.n_ptr != NULL) && (s.n_ptr->n_type != LIST) &&
        (s.n_ptr->n_type != REL) && (s.n_ptr->n_type != SET) )
        xlfail( "bad argument type" ) ;
/*    if( s.n_ptr == NULL ) xlfail( "invalid argument" ) ; */
    if( s.n_ptr != NULL ) {

       s.n_ptr = reverse_xm( s.n_ptr ) ;

       exp.n_ptr = newnode( LIST ) ;
       exp.n_ptr->n_listvalue = fun ;
       exp.n_ptr->n_listnext = new = newnode( LIST ) ;
       new->n_listvalue = rse ;
       new = new->n_listnext = newnode( LIST ) ;
       new->n_listvalue = rac ;

       oldenv = xlenv ;
       for( s1 = s.n_ptr ; s1 != NULL ; s1 = s1->n_listnext ) {
           xlbind( rse, s1->n_listvalue ) ;
           xlbind( rac, acc.n_ptr ) ;
           xlfixbindings( oldenv ) ;
           acc.n_ptr = xleval( exp.n_ptr ) ;
           xlunbind( oldenv ) ; }
     }

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return pointer to the result */
    return( acc.n_ptr ) ;
}


/*. pair - builtin function pair .*/
/*. pair : exp exp -> list .*/
static struct node *pair( args )
struct node *args ;
{
    struct node a1, a2, arg, l ;
    struct node *oldstk ;

    /* create a new stack frame */
    oldstk = xlsave( &l, &arg, &a1, &a2, NULL ) ;

    arg.n_ptr = args ;
    a1.n_ptr = xlevarg( &arg.n_ptr ) ;
    a2.n_ptr = xlevarg( &arg.n_ptr ) ;
    xllastarg( arg.n_ptr ) ;

    l.n_ptr = newnode( LIST ) ;
    l.n_ptr->n_listvalue = a1.n_ptr ;
    l.n_ptr->n_listnext = newnode( LIST ) ;
    l.n_ptr->n_listnext->n_listvalue = a2.n_ptr ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return pointer to the created pair */
    return( l.n_ptr ) ;
}


/*. mu2select - select the "n"th element from the tuple pointed by "args" .*/
static struct node *mu2select( n, args )
int n ;
struct node *args ;
{
    struct node *t ;

    t = userevarg( &args ) ;
    xllastarg( args ) ;
    if(t == NULL ) xlfail ("wrong type");
    if(t->n_type != LIST) xlfail ("wrong type");

    for( ; (n > 1) && (t != NULL) ; t = t->n_listnext, n-- ) ;
    if( t == NULL ) xlfail( "invalid argument" ) ;
    return( t->n_listvalue ) ;
}

/*. m2select - select the "n"th element from the tuple pointed by "args" .*/
static struct node *m2select( n, args )
int n ;
struct node *args ;
{
    struct node *t ;

    t = xlevmatch( LIST, &args ) ;
    xllastarg( args ) ;

    for( ; (n > 1) && (t != NULL) ; t = t->n_listnext, n-- ) ;
    if( t == NULL ) xlfail( "invalid argument" ) ;
    return( t->n_listvalue ) ;
}

static struct node *p1( struct node*args )
{ return( mu2select( 1, args ) ) ; }
 
static struct node *p2( struct node*args )
{ return( mu2select( 2, args ) ) ; }
 
static struct node *p3( struct node*args )
{ return( mu2select( 3, args ) ) ; }
 
static struct node *p4( struct node*args )
{ return( mu2select( 4, args ) ) ; }
 
static struct node *p5( struct node*args )
{ return( mu2select( 5, args ) ) ; }
 
static struct node *p6( struct node*args )
{ return( mu2select( 6, args ) ) ; }
 
static struct node *p7( struct node*args )
{ return( mu2select( 7, args ) ) ; }
 
/*. first - builtin function first .*/
/*. first : explist -> value .*/
static struct node *first( args )
struct node *args ;
{
    return( m2select( 1, args ) ) ;
}


/*. second - builtin function second .*/
/*. second : explist -> value .*/
static struct node *second( args )
struct node *args ;
{
    return( m2select( 2, args ) ) ;
}


/*. third - builtin function third .*/
/*. third : explist -> value .*/
static struct node *third( args )
struct node *args ;
{
    return( m2select( 3, args ) ) ;
}


/*. fourth - builtin function fourth .*/
/*. fourth : explist -> value .*/
static struct node *fourth( args )
struct node *args ;
{
    return( m2select( 4, args ) ) ;
}


/*. fifth - builtin function fifth .*/
/*. fifth : explist -> value .*/
static struct node *fifth( args )
struct node *args ;
{
    return( m2select( 5, args ) ) ;
}


/*. null - builtin function null .*/
/*. null : exp -> bool .*/
static struct node *null( args )
struct node *args ;
{
    struct node *arg ;

    /* get the argument */
    arg = xlevarg( &args ) ;
    xllastarg( args ) ;

    return( (arg == NULL ) ? true : false ) ;
}


void xmlistinit() 
{
    rac = xlenter( " " ) ;
    rse = xlenter( "  " ) ;

    xlsubr( "list", xlist ) ;
    xlsubr( "makeseq", xlist ) ;
    xlsubr( "null" , null ) ;
    xlsubr( "length", length ) ;
    xlsubr( "append", append ) ;
    xlsubr( "reverse", reverse ) ;
    xlsubr( "cons", cons ) ;
    xlsubr( "hd", hd ) ;
    xlsubr( "head", hd ) ;
    xlsubr( "tl", tl ) ;
    xlsubr( "tail", tl ) ;
    xlsubr( "elems", elems ) ;
    xlsubr( "inds", inds ) ;
    xlsubr( "nth", nth ) ;
    xlsubr( "el", nth ) ;
    xlsubr( "plusq", plusq ) ;
    xlsubr( "CONC", CONC ) ;
    xlsubr( "seq", sequence ) ;
    xlsubr( "reduce", reduce ) ;
    xlsubr( "pair", pair ) ;
    xlsubr( "first", first ) ;
    xlsubr( "second", second ) ;
    xlsubr( "third", third ) ;
    xlsubr( "fourth", fourth ) ;
    xlsubr( "fifth", fifth ) ;
    xlsubr( "p1", p1 ) ;
    xlsubr( "p2", p2 ) ;
    xlsubr( "p3", p3 ) ;
    xlsubr( "p4", p4 ) ;
    xlsubr( "p5", p5 ) ;
    xlsubr( "p6", p6 ) ;
    xlsubr( "p7", p7 ) ;
}

