/* xmmsg.c - builtin communication functions */

#ifdef COMM

#include <stdio.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/errno.h>
#include <signal.h>
#include "xmetoo.h"
#include "xmsession.h"
/*#include "xmc.h"*/
typedef struct node * SEXP;
#define XM_ERROR ((SEXP) -1)
extern SEXP arg_n();
extern SEXP arg_n_opt();



/* external variables */
extern struct node *xlstack ;
extern int errno ;

/* external procedures */
extern int msgctl() ;
extern int msgget() ;
extern int msgsnd() ;
extern int msgrcv() ;
extern char *strcpy() ;


#define MSGERROR     (-1)
#define ACCESS       0666
#define CreateMask   (IPC_CREAT | ACCESS)
#define SndMask      (0) /*(~IPC_NOWAIT)*/
#define RcvMask      (0) /*(~(IPC_NOWAIT | MSG_NOERROR))*/


/* local variables */
static struct _msgbuf {
                        long mtype ;
                        char mtext[STRMAX] ;
} qbuf ;
#define MSGLEN  (sizeof(qbuf))


/*. channel - builtin function channel .*/
/*. channel : expstr -> chan | nil .*/

/*. key_t - converts queue's name to an integer .*/
static key_t getkey( p )
char *p ;
{
    key_t  key ;
    int i ;

    key = (key_t) 0 ;
    for( i = 0 ; (i < sizeof( key_t )) && p[i] ; i++ ) {
        key = (key << 8) | p[i] ; }
    return( key ) ;
}

struct node *channel( args )
struct node *args ;
{
    struct msqid_ds buf ;
    struct node arg, str ;
    struct node *oldstk, *new ,*arg1,*arg2;
    key_t key ;
    int   st_len, qid ;
    extern char* getenv();
    char * xmsession = getenv("XMSESSION");
    short session = (xmsession == NULL ? getuid() : atoi(xmsession));

    /* create a new stack frame */
    oldstk = xlsave( &arg,  NULL ) ;

    arg.n_ptr = args ;

    arg1 = arg_n(1,args,STR);

    st_len = strlen( arg1->n_str ) ;

    if( (st_len < 1) || (st_len > sizeof( key_t )) )
        xlfail( "invalid argument1" ) ;
    
    /*
    key = getkey( str.n_ptr->n_str ) ;
    */

    arg2 = arg_n_opt(2,args,INT);
    if(arg2 != XM_ERROR) {session = arg2 -> n_int;}

    if(!xmsession_running()){
	if(fork() == 0) {
                signal(SIGINT,SIG_IGN );
		D_start_daemon();
		exit(0);
	}
	else sleep(2);
    }
    if((qid = join_channel(session, arg1->n_str)) == -1) new = NULL;
    else {
        if( msgctl( qid, IPC_STAT, &buf ) == MSGERROR )
            xlfail( "channel initialization error" ) ;
        new = newnode( CHAN ) ;
        new->n_qname = arg1 ; new->n_qid = qid ; }
   
    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return pointer to the result */
    return( new ) ;
}


/*. rmchannel - builtin function rmchannel .*/
/*. rmchannel : chan ->  nil .*/
struct node *rmchannel( args )
struct node *args ;
{
    struct msqid_ds buf ;
    struct node *arg ;
    int qid ;


    arg = xlevmatch( CHAN, &args ) ;
    xllastarg( args ) ;

        if( msgctl( arg->n_qid, IPC_RMID, &buf ) == MSGERROR )
            xlfail( "channel removal error" ) ;

    return( NULL ) ;
}


/*. send - builtin function send .*/
/*. send : expchan exp -> value .*/

#define send_mac\
        if( msgsnd( q->n_qid, (struct msgbuf*)&qbuf, MSGLEN, SndMask )\
                     == MSGERROR )\
                                   if(errno == EINVAL)\
                                              xlfail( "channel removed" );\
                                   else xlfail( "communication error" )

/*. send_unit - send S-expression "sexp" through queue "q" .*/
static void send_unit( q, sexp )
struct node *q, *sexp ;
{
#ifdef LAZY
    struct node atom ;
    struct node *oldstk ;
#endif
    struct node *p ;

    if( sexp == NULL ) {
        qbuf.mtype = (long)NILM ;
        send_mac ;
        return ; }

    switch( sexp->n_type ) { 
    case INT : 
        qbuf.mtype = (long)INT ; 
        *((int *)(qbuf.mtext)) = sexp->n_int ;
        send_mac ;
        break ;

    case STR :
        qbuf.mtype = (long)STR ;
        strcpy( qbuf.mtext, sexp->n_str ) ;
        send_mac ;
        break ;

    case SYM :
        qbuf.mtype = (long)SYM ;
        strcpy( qbuf.mtext, sexp->n_symname ) ;
        send_mac ;
        break ;

    case LIST :
    case SET :
    case FUN :
        qbuf.mtype = (long) sexp->n_type ;
        send_mac ;
        for( p = sexp ; p != NULL ; p = p->n_listnext )
            send_unit( q, p->n_listvalue ) ;
        qbuf.mtype = (long) ENDCOMP ;
        send_mac ;
        break ;

    case REL :
    case FF :
        qbuf.mtype = (long) sexp->n_type ;
        send_mac ;
        for( p = sexp ; p != NULL ; p = p->n_listnext ) {
            send_unit( q, p->n_listvalue->n_listvalue ) ;
            send_unit( q, p->n_listvalue->n_listnext->n_listvalue ); }
        qbuf.mtype = (long) ENDCOMP ;
        send_mac ;
        break ;

#ifdef LAZY
    case REC :
        send_unit( q, xleval( sexp ) ) ;
        break ;

    case LISTREC :
        /* create a new stack frame */ 
        oldstk = xlsave( &atom, NULL ) ;
        atom.n_ptr = sexp ;

        qbuf.mtype = (long) LIST ;
        send_mac ;
        do {
            send_unit( q, xleval( atom.n_ptr->n_listvalue ) ) ;
            atom.n_ptr = xleval( atom.n_ptr->n_listnext ) ;
            if( (atom.n_ptr != NULL) && (atom.n_ptr->n_type != LISTREC) )
                 xlfail( "bad argument type" ) ;
        } while( atom.n_ptr != NULL) ;
        qbuf.mtype = (long) ENDCOMP ;
        send_mac ;

        /* restore the previous stack frame */
        xlstack = oldstk ;
        break ;
#endif

    default :
#ifdef USERTYPE
        if( sexp->n_type >= USERDEFTYPE ) {
            qbuf.mtype = (long) sexp->n_type ;
            send_mac ;
            send_unit( q, sexp->n_symvalue ) ; }
        else
#endif
        xlfail( "bad argument type" ) ; }
}

static int val_type( n )
struct node *n ;
{
    if( n == NULL ) return( TRUE ) ;
    switch( n->n_type ) {
    case INT :
    case STR :
    case SYM :
    case LIST :
    case SET :
    case REL :
    case FF :
    case FUN :
#ifdef LAZY
    case REC :
    case LISTREC :
#endif
        return( TRUE ) ;
    default :
#ifdef USERTYPE
        if( n->n_type >= USERDEFTYPE )
            return( TRUE ) ;
        else
#endif
        return( FALSE ) ; }
}

static struct node *send( args )
struct node *args ;
{
    struct node arg, queue, msg ;
    struct node *oldstk ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &queue, &msg, NULL ) ;

    arg.n_ptr = args ;

    queue.n_ptr = xlevmatch( CHAN, &arg.n_ptr ) ;
    msg.n_ptr = xlevarg( &arg.n_ptr ) ;
    xllastarg( arg.n_ptr ) ;

    if( !val_type( msg.n_ptr ) ) xlfail( "bad argument type" ) ;
    
    send_unit( queue.n_ptr, msg.n_ptr ) ;

    /*restore the previous stack frame */
    xlstack = oldstk ;

    /* return the sent expression */
    return( msg.n_ptr ) ;
}


/*. receive - builtin function receive .*/
/*. receive : expchan -> value .*/

#define receive_mac\
        if( msgrcv( q->n_qid, (struct msgbuf*)&qbuf, MSGLEN, (long) 0,\
                     RcvMask ) == MSGERROR)\
                                   if(errno == EINVAL)\
                                              xlfail( "channel removed" );\
                                   else xlfail( "communication error" )


/*. rec_unit - receives an object "at" from "q" .*/
static void rec_unit( q, p_at )
struct node *q, **p_at ;
{
    struct node atom, temp ;
    struct node *oldstk, *new, *last ;
    int c_type ;

    /* create a new stack frame */
    oldstk = xlsave( &atom, &temp, NULL ) ;

    receive_mac ;
    switch( (int) qbuf.mtype ) {
       
    case NILM :
        atom.n_ptr = NULL ;
        break ;

    case ENDCOMP :
        atom.n_ptr = newnode( (int) qbuf.mtype ) ;
        break ;

    case INT :
        atom.n_ptr = newnode( INT ) ;
        atom.n_ptr->n_int = *((int *)(qbuf.mtext)) ;
        break ;

    case STR :
        atom.n_ptr = newnode( STR ) ;
        atom.n_ptr->n_str = strsave( qbuf.mtext ) ;
        break ;

    case SYM :
        atom.n_ptr = xlenter( qbuf.mtext ) ;
        break ;

    case LIST :
    case SET :
    case FUN :
        c_type = (int) qbuf.mtype ;
        rec_unit( q, &temp.n_ptr ) ;
        for( last = NULL ; (temp.n_ptr == NULL) ||
                          (temp.n_ptr->n_type != ENDCOMP) ; last = new ) {
            new = newnode( LIST ) ;
            new->n_listvalue = temp.n_ptr ;
            if( last == NULL )
                atom.n_ptr = new ;
            else
                last->n_listnext = new ;
            rec_unit( q, &temp.n_ptr ) ; }
        atom.n_ptr->n_type = c_type ;
        break ;

    case FF :
    case REL :
        c_type = (int) qbuf.mtype ;
        rec_unit( q, &temp.n_ptr ) ;
        for( last = NULL ; (temp.n_ptr == NULL) ||
                           (temp.n_ptr->n_type != ENDCOMP) ; ) {
            new = newnode( LIST ) ;
            if( last == NULL )
                atom.n_ptr = new ;
            else
                last->n_listnext = new ;
            last = new ;
            new = new->n_listvalue = newnode( LIST ) ;
            new->n_listvalue = temp.n_ptr ;
            rec_unit( q, &temp.n_ptr ) ;
            new = new->n_listnext = newnode( LIST ) ;
            new->n_listvalue = temp.n_ptr ;
            rec_unit( q, &temp.n_ptr ) ; }
        atom.n_ptr->n_type = c_type ;
        break ;

    default :
#ifdef USERTYPE
        if( (int) qbuf.mtype >= USERDEFTYPE ) {
            atom.n_ptr = newnode( (int) qbuf.mtype ) ;
            rec_unit( q, &atom.n_ptr->n_symvalue ) ; }
        else
#endif
        xlfail( "corrupted channel" ) ; }

    /* restore the previous stack frame */
    xlstack = oldstk ;

    *p_at = atom.n_ptr ;
}
 
static struct node *receive( args )
struct node *args ;
{
    struct node arg, queue ;
    struct node *oldstk, *buf ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &queue, NULL ) ;

    arg.n_ptr = args ;

    queue.n_ptr = xlevmatch( CHAN, &arg.n_ptr ) ;
    xllastarg( arg.n_ptr ) ;


    rec_unit( queue.n_ptr, &buf) ;

    /*restore the previous stack frame */
    xlstack = oldstk;

    /* return the received value */
    return( buf ) ;
}


/*. xmmsginit - xmmsg initialization procedure .*/
void xmmsginit()
{
    xlsubr( "channel", channel ) ; 
    xlsubr( "rmchannel", rmchannel ) ;
    xlsubr( "send", send ) ;
    xlsubr( "receive", receive ) ;
}
#endif

