/* xmpipe.c - xmetoo file pipes */
/* p2open  : expstr -> (fptr fptr) | nil */
/* p2close : [expfptr2] -> str | nil */
/* pgets   : [expfptr2] -> str | nil */
/* pread   : [expfptr2] -> SEXP | nil */
/* pputs   : expstr exppptr ->  */
/*---------------------------------------------------------------------
   see the use of "SEXP labeval(SEXP arg)"
                   SEXP arg_n(int n,SEXP args ,int tipo)
          SEXP|XM_ERROR arg_n_opt(int n,SEXP args ,int tipo)
---------------------------------------------------------------------*/
#ifdef PIPE
#define MAXPIPE 10
#define STRMAX 100
#include <stdio.h>
#include <unistd.h>
#include "xmc.h"

static FILE *fdl_[MAXPIPE * 2];
static int fdp_=0;
#define setfd(i,f0,f1) {i=fdp_++; fdl_[i*2] = f0; fdl_[i*2+1]= f1 ; }
#define getfd(i,f0,f1) {if(i < fdp_) {f0=fdl_[i*2] ; f1=fdl_[i*2+1];} else{xlfail("invalid pipe id");}}


int	p2open(char* comando, FILE* f[2]) {

	int	p1[2], p2[2];
	if(pipe(p1) == -1)
	{ perror("pipe"); return -1; }
	if(pipe(p2) == -1)
	{ perror("pipe"); close(p1[0]); close(p1[1]); return -1; }
	switch(fork()) {
		case -1:
			close(p1[0]); close(p1[1]);
			close(p2[0]); close(p2[1]);
			return -1;
		case  0:
			dup2(p1[0], 0); dup2(p2[1], 1);
			close(p1[1]); close(p2[0]);
			execl("/bin/sh", "sh", "-c", comando, NULL);
			perror("execl");
			return -1;
		default:
			close(p1[0]); close(p2[1]);
			f[1] = fdopen(p2[0], "r");
			f[0] = fdopen(p1[1], "w");
                        setbuf(f[0],NULL);
                        setbuf(f[1],NULL);
                        fflush(f[0]);
                        fflush(f[1]);
	}
	return 0;
}
#define error(func) if (func != 0) xlfail("pipe 2 open")

/* p2open : expstr -> (fptr fptr) | nil */
static SEXP xlp2open(SEXP args )
{
   int aux;
   FILE *fd2[2];
   
   error(p2open( arg_n(1,args,STR)->n_str, fd2));
   setfd(aux,fd2[0],fd2[1]);
   return(intsexp(aux));
}

/* xlp2close : [expfptr2] -> str | nil */
static SEXP xlp2close( SEXP args )
{
    int p;
    char buf[STRMAX+1];
    FILE* fp[2];
    p = arg_n(1,args,INT)->n_int;
    getfd(p,fp[0],fp[1]);
    fclose(fp[0]);
    fclose(fp[1]);
}

/* pgets : [expfptr2] -> str | nil */
static SEXP xlpgets(SEXP args )
{
    char buf[STRMAX+1];
    int p;
    FILE* fp[2];
    p = arg_n(1,args,INT)->n_int;    
    getfd(p,fp[0],fp[1]);
    return ( (fgets( buf, STRMAX, fp[1] ) != NULL )? 
              strsexp(buf):
              (SEXP) NULL);
}

/* pread : [expfptr2] -> SEXP | nil */
static SEXP xlpread( SEXP args )
{
    char buf[STRMAX+1];
    SEXP r;
    int p;
    FILE* fp[2];
    p = arg_n(1,args,INT)->n_int;    
    getfd(p,fp[0],fp[1]);
    return ( (r =labeval(f1("readf",fpsexp(fp[1])))) != XM_ERROR )? 
              r : (SEXP) NULL ;
}

/* pputs : expstr exppptr ->  */
static SEXP xlpputs(SEXP args )
{
    char *buf;
    int p;
    FILE* fp[2];
    p = arg_n(1,args,INT)->n_int;    
    getfd(p,fp[0],fp[1]);
    buf = arg_n(2,args,STR)->n_str;
    fprintf( fp[0], "%s\n",buf); 
    fflush(fp[0]);
    return((SEXP) NULL);
}
#endif

void pipeinit()
{
#ifdef PIPE
    xlsubr( "p2open", xlp2open ) ;
    xlsubr( "p2close", xlp2close ) ;
    xlsubr( "pgets", xlpgets ) ;
    xlsubr( "pputs", xlpputs ) ;
    xlsubr( "pread", xlpread ) ;
#endif
}

