/* xmprint.c - xmetoo print routines */

#include <stdio.h>
#include "xmetoo.h"

/* external variables */
extern struct node *xlstack ;


/* external procedures */
#ifdef USERTYPE
extern void print_type() ;
#endif


/*. putstring - output a string .*/
static void putstring( str, fp )
FILE * fp ;
char *str ;
{
    int ch ;

    /* output the initial quote */
    putc( '"' ,fp) ;

    /* output each character in the string */
    while( ch = *str++ )

        /* check for a control character */
        if( (ch < 040) || (ch == '\\') ) {
            putc( '\\' ,fp) ;
            switch( ch ) {
            case '\033' :
                putc( 'e' ,fp) ;
                break ;
            case '\n' :
                putc( 'n' ,fp) ;
                break ;
            case '\r' :
                putc( 'r' ,fp) ;
                break ;
            case '\t' :
                putc( 't' ,fp) ;
                break ;
            case '\\' :
                break;
            default :
                printf( "%03o", ch ) ;
                break ; } }

        else
        /* output a normal character */
            putc( ch ,fp) ;

    /* output the terminating quote */
    putc( '"' ,fp) ;
}

/*. xfprintf - fprintf function .*/


/*. xlprinterr - print an xlisp value .*/
void xlprinterr( vptr, flag ,fp)
FILE *fp;
struct node *vptr ;
int flag ;
{
#ifdef LAZY
    struct node m2rec ;
    struct node *oldstk ;
#endif
#ifdef USERTYPE
    struct node *prval ;
#endif
    struct node *nptr, *next ;
#ifdef FGETNAME 
    char buffer[128] ;
#endif

#ifdef LAZY
    /* create a new stack frame */
    oldstk = xlsave( &m2rec, NULL ) ;
#endif

    /* print null as the empty list */
    if( vptr == NULL ) {
        fprintf(fp, "()" ) ;
        return ; }

    /* check value type */
    switch( vptr->n_type ) {
    case SUBR :
        fprintf(fp, "<Subr: #%o>", vptr ) ;
        break ;
#ifdef LAZY
    case REC :
#ifdef USERTYPE
        xlprinterr( labeval( vptr ), flag ,fp) ;
#else
        xlprinterr( xleval( vptr ), flag ,fp) ;
#endif
        break ;
    case LISTREC :
        m2rec.n_ptr = vptr ;
        putc( '(' ,fp) ;
#ifdef USERTYPE
        xlprinterr( labeval( m2rec.n_ptr->n_listvalue ), flag ,fp) ;
#else
        xlprinterr( xleval( m2rec.n_ptr->n_listvalue ), flag ,fp) ;
#endif
        m2rec.n_ptr = m2rec.n_ptr->n_listnext ;
        do {
            m2rec.n_ptr = xleval( m2rec.n_ptr ) ;
            if( (m2rec.n_ptr != NULL) && (m2rec.n_ptr->n_type != LISTREC) )
                xlfail( "bad argument type" ) ;
            if( m2rec.n_ptr != NULL ) {
                putc( ' ' ,fp) ;
#ifdef USERTYPE
                xlprinterr( labeval( m2rec.n_ptr->n_listvalue ), flag ,fp) ;
#else
                xlprinterr( xleval( m2rec.n_ptr->n_listvalue ), flag ,fp) ;
#endif
                m2rec.n_ptr = m2rec.n_ptr->n_listnext ; }
        } while( m2rec.n_ptr != NULL ) ;
        putc( ')' ,fp) ;
        break ;
#endif
    case SET :
    case REL :
    case FF :
    case FUN :
    case LIST :
        putc( '(' ,fp) ;
        for( nptr = vptr ; nptr != NULL ; nptr = next ) {
#ifdef USERTYPE
            prval = nptr->n_listvalue ;
            while( (prval != NULL) && (prval->n_type >= USERDEFTYPE) )
                prval = prval->n_symvalue ;
            xlprinterr( prval, flag ,fp) ;
#else
            xlprinterr( nptr->n_listvalue, flag ,fp) ;
#endif
            if( (next = nptr->n_listnext) != NULL )
                putc( ' ' ,fp) ; }
        putc( ')' ,fp) ;
        break ;
    case SYM :
        fprintf(fp, "%s", vptr->n_symname ) ;
        break ;
    case INT :
        fprintf(fp, "%d", vptr->n_int ) ;
        break ;
    case STR :
        if( flag )
            putstring( vptr->n_str,fp ) ;
        else
            fprintf(fp, "%s", vptr->n_str ) ;
        break ;
    case FPTR :
#ifdef FGETNAME
        fprintf(fp, "<File: %s>", fgetname( vptr->n_fp, buffer ) ) ;
#else
        fprintf(fp, "<File: #%o>", vptr ) ;
#endif
        break;
#ifdef COMM
    case CHAN :
        fprintf(fp, "<Channel: " ) ;
        xlprinterr( vptr->n_qname, flag ,fp) ;
        fprintf(fp, " %d>", vptr->n_qid ) ;
        break ;
#endif
#ifdef USERTYPE
    case TYPE :
        fprintf(fp, "<Type: " ) ;
        print_type( vptr->n_typedef, stdout ) ;
        fprintf(fp, ">" ) ;
        break ;
    case LABEL :
        fprintf(fp, "<Label>" ) ;
        break ;
    case MACTYPE:
        fprintf(fp, "<Mactype: " ) ;
        for( nptr = vptr->n_macdef ; nptr != NULL ; nptr = next ) {
            xlprinterr( nptr->n_listvalue, flag ,fp) ;
            if( (next = nptr->n_listnext) != NULL )
                putc( ' ' ,fp) ; }
        fprintf(fp, ">" ) ;
        break ;
#endif
    default:
        xlfail( "bad node type" ) ; }

#ifdef LAZY
    /*restore the previous stack frame */
    xlstack = oldstk ;
#endif
}

/*. xlprint - print an xlisp value .*/
void xlprint( vptr, flag )
struct node *vptr ;
int flag ;
{ xlprinterr(vptr,flag,stdout);
}


/*. xprint - common print function .*/
static void xprint( args, flag )
struct node *args ;
int flag ;
{
    struct node arg ;
    struct node *oldstk ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, NULL ) ;

    /* initialize */
    arg.n_ptr = args ;

    /* evaluate and print each argument */
    while( arg.n_ptr != NULL )
      {
#ifdef USERTYPE
        xlprint( userevarg( &arg.n_ptr ), flag ) ;
#else 
        xlprint( xlevarg( &arg.n_ptr ), flag ) ;
#endif
      }

    /* restore previous stack frame */
    xlstack = oldstk ;
}


/*. print - builtin function print .*/
/*. print : exp... -> nil .*/
static struct node *print( args )
struct node *args ;
{
    xprint( args, TRUE ) ;
    return( NULL ) ;
}


/*. princ - builtin function princ .*/
/*. princ : exp... -> nil .*/
static struct node *princ( args )
struct node *args ;
{
    xprint( args, FALSE ) ;
    return( NULL ) ;
}


/*. xmprintinit - initialize the print routines .*/
void xmprintinit()
{
    /* enter builtin functions */
    xlsubr( "print", print ) ;
    xlsubr( "princ", princ ) ;
}

