/* xmread.c - xmetoo expression input routines */

#include <stdio.h>
#include <ctype.h>
#include "xmetoo.h"


/* external variables */
extern struct node *xlstack ;
extern struct node *true ;
extern struct node *false ;
extern struct node *oblist ;


/* external procedures */
extern char *strcat() ;
extern char *strsave2() ;

/* local variables */
static int savech ;
static int prompt ;
int (*xlgetc)() ;
static int xlplevel ;
static struct node *EOF_mark ;
#define MAX_LEV 10
static FILE *oldfp[MAX_LEV] ;


/* global variables */
int lev ;
extern int neverprompt,PP;

static struct node *setpp( args )
struct node *args ;
{
    /* make sure there aren't any arguments */
    xllastarg( args ) ;
    PP= ! PP;
    return(NULL);
}

/* EOF detection */
#define EOFILE     (-3)

/* forward declarations */
static struct node *parse() ;
static struct node *plist() ;
static struct node *pstring() ;
static struct node *pnumber() ;
static struct node *pquote() ;
static struct node *pname() ;
static void        pcomment() ;
int         issym() ;

/*. tgetc - get a character from the terminal .*/
static int tgetc()
{
    int ch ;

    /* prompt if necessary */
    if( prompt && (!neverprompt) ) {
        if( xlplevel > 0 )
            printf( "%d> ", xlplevel ) ;
        else
            printf( "> " ) ;
        prompt = FALSE ; }
    else if( prompt ) {putchar(1);
                       fflush(stdout);
                       prompt = FALSE;}            /* jj */

    /* get the character */
    if( (ch = getc( stdin )) == '\n' )
        prompt = TRUE ;

    /* return the character */
    return( ch ) ;
}


/*. xltin - setup terminal input .*/
void xltin( flag )
int flag ;
{
    /* flush line if flag is set */
    if( flag & !prompt )
        while( tgetc() != '\n' ) ;

    /* initialize */
    prompt = TRUE ;
    xlplevel = 0 ;
    xlgetc = tgetc ;
}


/*. xlread - read an xmetoo expression .*/
struct node *xlread( fp )
FILE *fp ;
{
    struct node *expr ;

    /* initialize */
    savech = -1 ;
    xlplevel = 0 ;

    expr = parse() ;

    if( (fp != NULL) && (savech >= 0) &&
        (savech != '\n') && (ungetc( savech, fp ) == EOF) )
        xlfail( "read error" ) ;

    return( expr ) ;
}


/*. parse - parse an xmetoo expression .*/
static struct node *parse()
{
    int ch ;

    /* keep looking for a node skipping comments */
    while( TRUE )

        /* check next character for type of node */
        switch( ch = nextch() ) {
        case EOFILE :
            /* terminate at end of input if there is */
            /* no load in progress                   */
            if(lev == 0)
             { putchar('\n');
               exit(0);
             }
            else return( EOF_mark ) ;
        case '\'' :    /* a quoted expression */
            return( pquote() ) ;
        case '(' :     /* a sublist */
            return( plist() ) ;
        case ')' :     /* closing paren - shouldn't happen */
            xlfail( "extra right paren" ) ;
        case ';' :     /* a comment */
            pcomment() ;
            break ;
        case '"' :     /* a string */
            return( pstring() ) ;
        default :
            if( isdigit( ch ) )         /* a number */
                return( pnumber( 1 ) ) ;
            else
                if( issym( ch ) )       /* a name */
                    return( pname() ) ;
                else
                   xlfail( "invalid character" ); }
}


/*. pcomment - parse a comment .*/
static void pcomment()
{
    int ch ;

    savech = -1 ;
    /* skip to end of line or end of file */
    while( ((ch = getch()) != '\n') && (ch != EOFILE) ) ;
    if( ch == EOFILE ) savech = EOFILE ;
}


/*. plist - parse a list .*/
static struct node *plist()
{
    struct node val ;
    struct node *oldstk, *lastnptr, *nptr ;
    int ch ;

    /* increment the nesting level */
    xlplevel++ ;

    /* create a new stack frame */
    oldstk = xlsave( &val, NULL ) ;

    /* skip the opening paren */
    savech = -1 ;

    /* keep appending nodes until a closing paren is found */
    ch = nextch() ;
    while( ch == ';' ) {
        pcomment() ;
        ch = nextch() ; }

    for( lastnptr = NULL ; (ch > 0) && (ch != ')') ; lastnptr = nptr ) {
        /* allocate a new node and link it into the list */
        nptr = newnode( LIST ) ;
        if( lastnptr == NULL )
            val.n_ptr = nptr ;
        else
            lastnptr->n_listnext = nptr ;
        /* initialize the new node */
        if( (nptr->n_listvalue = parse()) == EOF_mark )
            xlfail( "unterminated expression" ) ;

        ch = nextch() ;
        while( ch == ';' ) {
            pcomment() ;
            ch = nextch() ; } }

    if( ch != ')' ) xlfail( "unterminated expression" ) ;

    /* skip the closing paren */
    savech = -1 ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* decrement the nesting level */
    xlplevel-- ;

    /* return successfully */
    return( val.n_ptr ) ;
}


/*. pstring - parse a string .*/
static struct node *pstring()
{
    struct node val ;
    struct node *oldstk ;
    char sbuf[STRMAX+1] ;
    int ch, i, d1, d2, d3 ;

    /* create a new stack frame */
    oldstk = xlsave( &val, NULL ) ;

    /* skip the opening quote */
    savech = -1 ;

    /* loop looking for a closing quote */
    for( i = 0 ; ((ch = getch()) != '"') && (ch != EOFILE) &&
                               (ch != '\n') && (i < STRMAX) ; i++ ) {
        if( ch == '\\' )
            switch( ch = getch() ) {
            case EOFILE :
                xlfail( "EOF encountered" ) ;
            case 'e' :
                ch = '\033' ;
                break ;
            case 'n' :
                ch = '\n' ;
                break ;
            case 'r' :
                ch = '\r' ;
                break ;
            case 't' :
                ch = '\t' ;
                break ;
            default :
                if( (ch >= '0') && (ch <= '7') ) {
                    d1 = ch - '0' ;
                    d2 = getch() - '0' ;
                    if( (d2 < 0) || (d2 > 7) )
                         xlfail( "invalid octal number" ) ;
                    d3 = getch() - '0' ;
                    if( (d3 < 0) || (d3 > 7) )
                         xlfail( "invalid octal number" ) ;
                    ch = (d1 << 6) + (d2 << 3) + d3 ; }
                break ; }
        sbuf[i] = ch ; }
   
    if( ch == EOFILE ) xlfail( "EOF encountered" ) ;
    if( ch == '\n' ) xlfail( "unterminated string" ) ;
    if( ch != '"' ) xlfail( "string too long" ) ;
    sbuf[i] = 0 ;

    /* initialize the node */
    val.n_ptr = newnode( STR ) ;
    val.n_ptr->n_str = strsave( sbuf ) ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return the new string */
    return( val.n_ptr ) ;
}


/*. pnumber - parse a number .*/
static struct node *pnumber( sign )
int sign ;
{
    struct node *val ;
    int ch, ival ;

    /* loop looking for digits */
    for( ival = 0 ; isdigit( ch = thisch() ) ; savech = -1 )
        ival = ival * 10 + ch - '0' ;

    /* make sure the number terminated correctly */
    if( issym( ch ) )
        xlfail( "badly formed number" ) ;

    /* initialize the node */
    val = newnode( INT ) ;
    val->n_int = sign * ival ;

    /* return the new number */
    return( val ) ;
}


/*. pquote - parse a quoted expression .*/
static struct node *pquote()
{
    struct node val ;
    struct node *oldstk ;

    /* create a new stack frame */
    oldstk = xlsave( &val, NULL ) ;

    /* skip the quote character */
    savech = -1 ;

    /* allocate two nodes */
    val.n_ptr = newnode( LIST ) ;
    val.n_ptr->n_listvalue = xlenter( "quote" ) ;
    val.n_ptr->n_listnext = newnode( LIST ) ;

    /* simulate a nesting level */
    xlplevel++ ;

    /* initialize the second node to point to the quoted expression */
    if( (val.n_ptr->n_listnext->n_listvalue = parse()) == EOF_mark )
        xlfail( "unterminated expression" ) ;

    xlplevel-- ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return the quoted expression */
    return( val.n_ptr ) ;
}


/*. pname - parse a symbol name .*/
static struct node *pname()
{
    char sname[STRMAX] ;
    int ch, i ;

    /* get the first character */
    ch = sname[0] = getch() ;

    /* check for signed number */
    if( (ch == '+') || (ch == '-') )
        if( isdigit( thisch() ) )
            return( pnumber( (ch == '+') ? 1 : -1 ) ) ;

    /* get symbol name */
    for( i = 1 ; (i < STRMAX - 1) && issym( thisch() ) ; i++ )
        sname[i] = getch() ;
    if( issym( thisch() ) ) xlfail( "name too long" ) ;
    sname[i] = 0 ;

    /* initialize value */
    return( xlenter( sname ) ) ;
}


/*. nextch - look at the next non-blank character .*/
static int nextch()
{
  int c;

    /* look for a non-blank character */
    while( isspace( c = thisch() ) && c != EOFILE)
        savech = -1 ;

    /* return the character */
    return( savech ) ;
}


/*. thisch - look at the current character .*/
static int thisch()
{
    /* return and save the current character */
    return( savech = getch() ) ;
}


/*. getch - get the next character .*/
static int getch()
{
    int ch ;

    /* check for a saved character */
    if( (ch = savech) >= 0 )
	savech = -1 ;
    else
	ch = (*xlgetc)() ;

    /* check for the abort character */
    if( ch == EOF )
	if( xlplevel > 0 )
	    xlfail( "input aborted" ) ;
	else
      return( EOFILE ) ;
    else
        if( ch == -2 )
	    xlfail( "string input aborted" ) ;

    /* return the character */
    return( ch ) ;
}


/*. issym - check whether a character if valid in a symbol name .*/
int issym( ch )
int ch ;
{
    if( isspace( ch ) || (ch <  ' ') || (ch == '(') || (ch == ')') ||
        (ch == ';') || (ch == '"') || (ch == '\'') || (ch == EOFILE) )
        return( FALSE ) ;
    else
        return( TRUE ) ;
}

static void nwritestr( FILE *fp, char *str )
{
    int ch ;

    /* output the initial quote */
    fputc( '"', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;

    /* output each character in the string */
    while( ch = *str++ ) {
           /* check for a control character */
            switch( ch ) {
            case '\033' :
                fputc( '\\', fp ) ; 
                fputc( 'e', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
                break ;
            case '\n' :
                fputc( '\\', fp ) ; 
                fputc( 'n', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
                break ;
            case '\r' :
                fputc( '\\', fp ) ; 
                fputc( 'r', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
                break ;
            case '\t' :
                fputc( '\\', fp ) ; 
                fputc( 't', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
                break ;
            case '\\' :
            case '"'  :
                fputc( '\\', fp ) ; 
                fputc( ch, fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
                break;
            default :
                fputc( ch, fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
                break ; } 
     }

    /* output the terminating quote */
    fputc( '"', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
/*  fputc( '\n', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ; */
}

/*. m2writestr - output a string into a file .*/
static void m2writestr( fp, str )
FILE *fp ;
char *str ;
{
    int ch ;

    /* output the initial quote */
    fputc( '"', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;

    /* output each character in the string */
    while( ch = *str++ )

        /* check for a control character */
        if( (ch < 040) || (ch == '\\') ) {
            fputc( '\\', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
            switch( ch ) {
            case '\033' :
                fputc( 'e', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
                break ;
            case '\n' :
                fputc( 'n', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
                break ;
            case '\r' :
                fputc( 'r', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
                break ;
            case '\t' :
                fputc( 't', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
                break ;
            case '\\' :
                fputc( '\\', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
                break;
            default :
                if( fprintf( fp, "%03o", ch ) != 3 ) xlfail( "write error" ) ;
                break ; } }
        else {
        /* output a normal character */
            fputc( ch, fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ; }

    /* output the terminating quote */
    fputc( '"', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
/*  fputc( '\n', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ; */
}


/*. writef - builtin function writef .*/
/*. writef : expfptr exp -> nil .*/

static void writefaux( fp, vptr )
FILE *fp ;
struct node *vptr ;
{
    struct node m2rec ;
    struct node *oldstk ;
    struct node *nptr, *next ;

    /* print null as the empty list */
    if( vptr == NULL ) {
        if( fprintf( fp, "()\n" ) != 3 ) xlfail( "write error" ) ;
        return ; }

    /* create a new stack frame */
    oldstk = xlsave( &m2rec, NULL ) ;

    /* check value type */
    switch( vptr->n_type ) {
#ifdef LAZY
    case REC :
        writefaux( fp, xleval( vptr ) ) ;
        break ;
    case LISTREC :
        m2rec.n_ptr = vptr ;
        fputc( '(', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
        writefaux( fp, xleval( m2rec.n_ptr->n_listvalue ) ) ;
        m2rec.n_ptr = m2rec.n_ptr->n_listnext ;
        do {
            m2rec.n_ptr = xleval( m2rec.n_ptr ) ;
            if( (m2rec.n_ptr != NULL) && (m2rec.n_ptr->n_type != LISTREC) )
                xlfail( "bad argument type" ) ;
            if( m2rec.n_ptr != NULL ) {
                fputc( ' ', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
                writefaux( fp, xleval( m2rec.n_ptr->n_listvalue ) ) ;
                m2rec.n_ptr = m2rec.n_ptr->n_listnext ; }
        } while( m2rec.n_ptr != NULL ) ;
        fputc( ')', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
        fputc( '\n', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
        break ;
#endif
    case REL :
    case FF :
        fputc( '(', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
        for( nptr = vptr ; nptr != NULL ; nptr = next ) {
            fputc( '(', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
            writefaux( fp, nptr->n_listvalue->n_listvalue ) ;
            writefaux( fp, nptr->n_listvalue->n_listnext->n_listvalue ) ;
            fputc( ')', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
            if( (next = nptr->n_listnext) != NULL )
                fputc( ' ', fp ) ; if( ferror( fp ) ) xlfail( "write error" );}
        fputc( ')', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
        fputc( '\n', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
        break ;
    case LIST :
    case SET :
        fputc( '(', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
        for( nptr = vptr ; nptr != NULL ; nptr = next ) {
            writefaux( fp, nptr->n_listvalue ) ;
            if( (next = nptr->n_listnext) != NULL )
                fputc( ' ', fp ) ; if( ferror( fp ) ) xlfail( "write error" );}
        fputc( ')', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
        fputc( '\n', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
        break ;
    case SYM :
        if( fprintf( fp, "%s\n", vptr->n_symname ) !=
                              strlen( vptr->n_symname ) + 1 )
            xlfail( "write error" ) ;
        break ;
    case INT :
        if( fprintf( fp, "%d\n", vptr->n_int ) <= 0 ) xlfail( "write error" ) ;
        break ;
    case STR :
        m2writestr( fp, vptr->n_str ) ;
        break ;
    case SUBR :
        xlfail( "SUBR cannot be written into a file" ) ;
        break ;
    case FPTR :
        xlfail( "FPTR cannot be written into a file" ) ;
        break;
#ifdef COMM
    case CHAN :
        xlfail( "CHAN cannot be written into a file" ) ;
        break ;
#endif
#ifdef USERTYPE
    case TYPE:
        xlfail( "TYPE cannot be written into a file" ) ;
        break ;
    case LABEL:
        xlfail( "LABEL cannot be written into a file" ) ;
        break ;
    case MACTYPE:
        xlfail( "MACTYPE cannot be written into a file" ) ;
        break ;
#endif
    case FUN :
        xlfail( "FUN cannot be written into a file" ) ;
        break ;
    default:
        xlfail( "bad node type" ) ; }

    /* restore the previous stack frame */
    xlstack = oldstk ;
}

static struct node *writef( args )
struct node *args ;
{
    struct node arg ;
    struct node *oldstk ;
    struct node *vptr ;
    FILE *fp ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, NULL ) ;

    arg.n_ptr = args ;
    if( (fp = xlevmatch( FPTR, &arg.n_ptr )->n_fp) == NULL )
        xlfail( "file not open" ) ;

#ifdef USERTYPE
    vptr = userevarg( &arg.n_ptr ) ;
#else
    vptr = xlevarg( &arg.n_ptr ) ;
#endif

    xllastarg( arg.n_ptr ) ;

    writefaux( fp, vptr ) ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    return( NULL ) ;
}


/*. m2write - builtin function m2write .*/
/*. m2write : expfptr exp -> nil .*/

static void m2writeaux( fp, vptr )
FILE *fp ;
struct node *vptr ;
{
    struct node m2rec ;
    struct node *oldstk ;
    struct node *nptr, *next ;
#ifdef USERTYPE
    extern struct t_table *name_table[] ;
    struct t_table *tdef ;
    struct lab_type *ptup ;
    struct node *tl ;
#endif

    /* print null as the empty list */
    if( vptr == NULL ) {
        if( fprintf( fp, "()\n" ) != 3 ) xlfail( "write error" ) ;
        return ; }

    /* create a new stack frame */
    oldstk = xlsave( &m2rec, NULL ) ;

    /* check value type */
    switch( vptr->n_type ) {
#ifdef LAZY
    case REC :
        m2writeaux( fp, xleval( vptr ) ) ;
        break ;
    case LISTREC :
        m2rec.n_ptr = vptr ;
        if( fprintf( fp, "(list " ) != 6 ) xlfail( "write error" ) ;
        m2writeaux( fp, xleval( m2rec.n_ptr->n_listvalue ) ) ;
        m2rec.n_ptr = m2rec.n_ptr->n_listnext ;
        do {
            m2rec.n_ptr = xleval( m2rec.n_ptr ) ;
            if( (m2rec.n_ptr != NULL) && (m2rec.n_ptr->n_type != LISTREC) )
                xlfail( "bad argument type" ) ;
            if( m2rec.n_ptr != NULL ) {
                fputc( ' ', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
                m2writeaux( fp, xleval( m2rec.n_ptr->n_listvalue ) ) ;
                m2rec.n_ptr = m2rec.n_ptr->n_listnext ; }
        } while( m2rec.n_ptr != NULL ) ;
        fputc( ')', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
        fputc( '\n', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
        break ;
#endif
    case REL :
        if( fprintf( fp, "(makerel " ) != 9 ) xlfail( "write error" ) ;
        goto label2 ;
    case FF :
        if( fprintf( fp, "(makeff " ) != 8 ) xlfail( "write error" ) ;
label2: for( nptr = vptr ; nptr != NULL ; nptr = next ) {
            fputc( '(', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
            m2writeaux( fp, nptr->n_listvalue->n_listvalue ) ;
            m2writeaux( fp, nptr->n_listvalue->n_listnext->n_listvalue ) ;
            fputc( ')', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
            if( (next = nptr->n_listnext) != NULL )
                fputc( ' ', fp ) ; if( ferror( fp ) ) xlfail( "write error" );}
        fputc( ')', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
        fputc( '\n', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
        break ;
    case SET :
        if( fprintf( fp, "(makeset " ) != 9 ) xlfail( "write error" ) ;
        goto label1 ;
    case LIST :
        if( fprintf( fp, "(list " ) != 6 ) xlfail( "write error" ) ;
label1: for( nptr = vptr ; nptr != NULL ; nptr = next ) {
            m2writeaux( fp, nptr->n_listvalue ) ;
            if( (next = nptr->n_listnext) != NULL )
                fputc( ' ', fp ) ; if( ferror( fp ) ) xlfail( "write error" );}
        fputc( ')', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
        fputc( '\n', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
        break ;
    case SYM :
        if( fprintf( fp, "'%s\n", vptr->n_symname ) !=
                              strlen( vptr->n_symname ) + 2 )
            xlfail( "write error" ) ;
        break ;
    case INT :
        if( fprintf( fp, "%d\n", vptr->n_int ) <= 0 ) xlfail( "write error" ) ;
        break ;
    case STR :
        m2writestr( fp, vptr->n_str ) ;
        break ;
    case SUBR :
        xlfail( "SUBR cannot be written into a file" ) ;
        break ;
    case FPTR :
        xlfail( "FPTR cannot be written into a file" ) ;
        break;
#ifdef COMM
    case CHAN :
        xlfail( "CHAN cannot be written into a file" ) ;
        break ;
#endif
#ifdef USERTYPE
    case TYPE:
        xlfail( "TYPE cannot be written into a file" ) ;
        break ;
    case LABEL:
        xlfail( "LABEL cannot be written into a file" ) ;
        break ;
    case MACTYPE:
        xlfail( "MACTYPE cannot be written into a file" ) ;
        break ;
#endif
    case FUN :
        xlfail( "FUN cannot be written into a file" ) ;
        break ;
    default:
#ifdef USERTYPE
        if( vptr->n_type >= USERDEFTYPE ) {
            tdef = name_table[vptr->n_type - USERDEFTYPE] ;
            if( fprintf( fp, "(%s ", tdef->tname ) != strlen( tdef->tname )+2 )
                xlfail( "write error" ) ;
            if( tdef->tnumber == TTUP ) {
                if( ((tl = vptr->n_symvalue) == NULL) || (tl->n_type != LIST) )
                    xlfail( "type mismatch" ) ;
                for( ptup = tdef->def.tcomp ; (ptup != NULL) && (tl != NULL);
                               ptup = ptup->next, tl = tl->n_listnext )
                     m2writeaux( fp, tl->n_listvalue ) ;
                if( (ptup != NULL) || (tl != NULL) )
                    xlfail( "type mismatch" ) ; }
            else
                m2writeaux( fp, vptr->n_symvalue ) ;
            fputc( ')', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
            fputc( '\n', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ; }
        else
#endif
        xlfail( "bad node type" ) ; }

    /* restore the previous stack frame */
    xlstack = oldstk ;
}

static void indenta(j,fp)
int j;
FILE *fp;
{int i=0; for(fputc('\n',fp) ;i<j;i++) fputc(' ',fp);}

static void nwriteaux( fp, vptr,indent,nl )
int indent,nl;
FILE *fp ;
struct node *vptr ;
{
    struct node m2rec ;
    struct node *oldstk ;
    struct node *nptr, *next ;
#ifdef USERTYPE
    extern struct t_table *name_table[] ;
    struct t_table *tdef ;
    struct lab_type *ptup ;
    struct node *tl ;
#endif

    /* print null as the empty list */
    if( vptr == NULL ) {
        if( fprintf( fp, " {}" ) != 3 ) xlfail( "write error" ) ;
        return ; }

    /* create a new stack frame */
    oldstk = xlsave( &m2rec, NULL ) ;

    /* check value type */
    switch( vptr->n_type ) {
#ifdef LAZY
    case REC :
        nwriteaux( fp, xleval( vptr ) ,indent,nl) ;
        break ;
    case LISTREC :
        m2rec.n_ptr = vptr ;
        if( fprintf( fp, "< " ) != 6 ) xlfail( "write error" ) ;
        nwriteaux( fp, xleval( m2rec.n_ptr->n_listvalue ) ,indent,nl) ;
        m2rec.n_ptr = m2rec.n_ptr->n_listnext ;
        do {
            m2rec.n_ptr = xleval( m2rec.n_ptr ) ;
            if( (m2rec.n_ptr != NULL) && (m2rec.n_ptr->n_type != LISTREC) )
                xlfail( "bad argument type" ) ;
            if( m2rec.n_ptr != NULL ) {
                fputc( ',', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
                nwriteaux( fp, xleval( m2rec.n_ptr->n_listvalue ) ,indent,nl) ;
                m2rec.n_ptr = m2rec.n_ptr->n_listnext ; }
        } while( m2rec.n_ptr != NULL ) ;
        fputc( '>', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
        fputc( '\n', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
        break ;
#endif
    case REL :
        if(nl==1) indenta(indent,fp);
        nptr = vptr;
        if( fprintf( fp, "[* " ) != 3 ) xlfail( "write error" ) ;
        nwriteaux( fp, nptr->n_listvalue->n_listvalue ,indent+3,0) ;
        fputc('-' , fp );
        fputc('>' , fp ) ; 
        nwriteaux( fp, nptr->n_listvalue->n_listnext->n_listvalue ,indent+6,0) ;
        next = nptr->n_listnext;
        for( nptr = next ; nptr != NULL ; nptr = next ) {
            fputc(',' , fp );
            nwriteaux( fp, nptr->n_listvalue->n_listvalue ,indent+3,1) ;
            fputc('-' , fp );
            fputc('>' , fp ) ; 
            nwriteaux( fp, nptr->n_listvalue->n_listnext->n_listvalue ,indent+6,0) ;
            next = nptr->n_listnext; }
              if( fprintf( fp, " *]" ) != 3 ) xlfail( "write error" ) ;
        break ;
    case FF :
        if(nl==1) indenta(indent,fp);
        nptr = vptr;
        if( fprintf( fp, "[ " ) != 2 ) xlfail( "write error" ) ;
        nwriteaux( fp, nptr->n_listvalue->n_listvalue ,indent+2,0) ;
        fputc('-' , fp );
        fputc('>' , fp ) ; 
        nwriteaux( fp, nptr->n_listvalue->n_listnext->n_listvalue ,indent+5,0) ;
        next = nptr->n_listnext;
        for( nptr = next ; nptr != NULL ; nptr = next ) {
            fputc(',' , fp );
            nwriteaux( fp, nptr->n_listvalue->n_listvalue ,indent+2,1) ;
            fputc('-' , fp );
            fputc('>' , fp ) ; 
            nwriteaux( fp, nptr->n_listvalue->n_listnext->n_listvalue ,indent+5,0) ;
            next = nptr->n_listnext; }
              if( fprintf( fp, " ]" ) != 2 ) xlfail( "write error" ) ;
        break ;
    case SET :
        if(nl==1) indenta(indent,fp);
        if( fprintf( fp, "{ " ) != 2 ) xlfail( "write error" ) ;
        nptr = vptr;
        nwriteaux( fp, nptr->n_listvalue ,indent+2,0) ;
        for(nptr = nptr->n_listnext; nptr != NULL ; nptr = nptr->n_listnext ) {
            fputc( ',', fp );
            nwriteaux( fp, nptr->n_listvalue ,indent+2,0) ;}
        if( fprintf( fp, "}" ) != 1 ) xlfail( "write error" ) ;
        break ;
    case LIST :
        if(nl==1) indenta(indent,fp);
        if( fprintf( fp, "< " ) != 2 ) xlfail( "write error" ) ;
        nptr = vptr;
        nwriteaux( fp, nptr->n_listvalue ,indent+2,0) ;
        for(nptr = nptr->n_listnext; nptr != NULL ; nptr = nptr->n_listnext ) {
            fputc( ',', fp );
            nwriteaux( fp, nptr->n_listvalue ,indent+2,0) ;}
        if( fprintf( fp, ">" ) != 1 ) xlfail( "write error" ) ;
        break ;
    case SYM :
        if(nl==1) indenta(indent,fp);
        if( fprintf( fp, "'%s ", vptr->n_symname ) !=
                              strlen( vptr->n_symname ) + 2 )
            xlfail( "write error" ) ;
        break ;
    case INT :
        if(nl==1) indenta(indent,fp);
        if( fprintf( fp, "%d ", vptr->n_int ) <= 0 ) xlfail( "write error" ) ;
        break ;
    case STR :
        if(nl==1) indenta(indent,fp);
        nwritestr( fp, vptr->n_str ) ;
        break ;
    case SUBR :
        xlfail( "SUBR cannot be written into a file" ) ;
        break ;
    case FPTR :
        xlfail( "FPTR cannot be written into a file" ) ;
        break;
#ifdef COMM
    case CHAN :
        xlfail( "CHAN cannot be written into a file" ) ;
        break ;
#endif
#ifdef USERTYPE
    case TYPE:
        xlfail( "TYPE cannot be written into a file" ) ;
        break ;
    case LABEL:
        xlfail( "LABEL cannot be written into a file" ) ;
        break ;
    case MACTYPE:
        xlfail( "MACTYPE cannot be written into a file" ) ;
        break ;
#endif
    case FUN :
        xlfail( "FUN cannot be written into a file" ) ;
        break ;
    default:
#ifdef USERTYPE
        /*if(nl==1) indenta(indent,fp); */
        indenta(indent,fp);
        if( vptr->n_type >= USERDEFTYPE ) {
            tdef = name_table[vptr->n_type - USERDEFTYPE] ;
            fprintf( fp, "%s( ", tdef->tname );
            if( tdef->tnumber == TTUP ) {
                if( ((tl = vptr->n_symvalue) == NULL) || (tl->n_type != LIST) )
                    xlfail( "type mismatch" ) ;
                for( ptup = tdef->def.tcomp ; (ptup != NULL) && (tl != NULL);
                               ptup = ptup->next )
                    { nwriteaux( fp, tl->n_listvalue ,indent+3,1) ;
                      if( (tl = tl->n_listnext) != NULL )
                        fputc(',',fp); if(ferror(fp)) xlfail( "write error" );}
                if( (ptup != NULL) || (tl != NULL) )
                    xlfail( "type mismatch" ) ; }
            else
                nwriteaux( fp, vptr->n_symvalue ,indent+3,0) ;
            fputc( ')', fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ; }
        else
#endif
        xlfail( "bad node type" ) ; }

    /* restore the previous stack frame */
    xlstack = oldstk ;
}


static struct node *m2write( args )
struct node *args ;
{
    struct node arg ;
    struct node *oldstk ;
    struct node *vptr ;
    FILE *fp ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, NULL ) ;

    arg.n_ptr = args ;
    if( (fp = xlevmatch( FPTR, &arg.n_ptr )->n_fp) == NULL )
        xlfail( "file not open" ) ;
    vptr = xlevarg( &arg.n_ptr ) ;
    xllastarg( arg.n_ptr ) ;

    m2writeaux( fp, vptr ) ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    return( NULL ) ;
}

static struct node *nwrite( args )
struct node *args ;
{
    struct node arg ;
    struct node *oldstk ;
    struct node *vptr ;
    FILE *fp ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, NULL ) ;

    arg.n_ptr = args ;
    if( (fp = xlevmatch( FPTR, &arg.n_ptr )->n_fp) == NULL )
        xlfail( "file not open" ) ;
    vptr = xlevarg( &arg.n_ptr ) ;
    xllastarg( arg.n_ptr ) ;

    nwriteaux( fp, vptr ,0,0) ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    return( NULL ) ;
}

void ppaux( vptr,indent,nl )
int indent,nl;
struct node *vptr ;
{
    struct node m2rec ;
    struct node *oldstk ;
    struct node *nptr, *next ;
#ifdef USERTYPE
    extern struct t_table *name_table[] ;
    struct t_table *tdef ;
    struct lab_type *ptup ;
    struct node *tl ;
#endif

    /* print null as the empty list */
    if( vptr == NULL ) {
        if( printf( " {}" ) != 3 ) xlfail( "write error" ) ;
        return ; }

    /* create a new stack frame */
    oldstk = xlsave( &m2rec, NULL ) ;

    /* check value type */
    switch( vptr->n_type ) {
#ifdef LAZY
    case REC :
        ppaux( xleval( vptr ) ,indent,nl) ;
        break ;
    case LISTREC :
        m2rec.n_ptr = vptr ;
        if( fprintf( stdout, "< " ) != 6 ) xlfail( "write error" ) ;
        ppaux( xleval( m2rec.n_ptr->n_listvalue ) ,indent,nl) ;
        m2rec.n_ptr = m2rec.n_ptr->n_listnext ;
        do {
            m2rec.n_ptr = xleval( m2rec.n_ptr ) ;
            if( (m2rec.n_ptr != NULL) && (m2rec.n_ptr->n_type != LISTREC) )
                xlfail( "bad argument type" ) ;
            if( m2rec.n_ptr != NULL ) {
                fputc( ' ', stdout ) ; if( ferror( stdout ) ) xlfail( "write error" ) ;
                ppaux( xleval( m2rec.n_ptr->n_listvalue ) ,indent,nl) ;
                m2rec.n_ptr = m2rec.n_ptr->n_listnext ; }
        } while( m2rec.n_ptr != NULL ) ;
        fputc( '>', stdout ) ; if( ferror( stdout ) ) xlfail( "write error" ) ;
        fputc( '\n', stdout ) ; if( ferror( stdout ) ) xlfail( "write error" ) ;
        break ;
#endif
    case REL :
        if(nl==1) indenta(indent,stdout);
        nptr = vptr;
        if( fprintf( stdout, "[ " ) != 2 ) xlfail( "write error" ) ;
        ppaux( nptr->n_listvalue->n_listvalue ,indent+3,0) ;
        fputc('<' , stdout );
        fputc('-' , stdout );
        fputc('>' , stdout ) ; 
        fputc(' ' , stdout ) ; 
        ppaux( nptr->n_listvalue->n_listnext->n_listvalue ,indent+6,0) ;
        next = nptr->n_listnext;
        for( nptr = next ; nptr != NULL ; nptr = next ) {
            ppaux( nptr->n_listvalue->n_listvalue ,indent+3,1) ;
            fputc('<' , stdout );
            fputc('-' , stdout );
            fputc('>' , stdout ) ; 
            fputc(' ' , stdout ) ; 
            ppaux( nptr->n_listvalue->n_listnext->n_listvalue ,indent+6,0) ;
            next = nptr->n_listnext; }
              if( fprintf( stdout, " ]" ) != 2 ) xlfail( "write error" ) ;
        break ;
    case FF :
        if(nl==1) indenta(indent,stdout);
        nptr = vptr;
        if( fprintf( stdout, "[ " ) != 2 ) xlfail( "write error" ) ;
        ppaux( nptr->n_listvalue->n_listvalue ,indent+2,0) ;
        fputc('-' , stdout );
        fputc('>' , stdout ) ; 
        fputc(' ' , stdout );
        ppaux( nptr->n_listvalue->n_listnext->n_listvalue ,indent+5,0) ;
        next = nptr->n_listnext;
        for( nptr = next ; nptr != NULL ; nptr = next ) {
            ppaux( nptr->n_listvalue->n_listvalue ,indent+2,1) ;
            fputc('-' , stdout );
            fputc('>' , stdout ) ; 
            fputc(' ' , stdout );
            ppaux( nptr->n_listvalue->n_listnext->n_listvalue ,indent+5,0) ;
            next = nptr->n_listnext; }
              if( fprintf( stdout, "]" ) != 1 ) xlfail( "write error" ) ;
        break ;
    case SET :
        if(nl==1) indenta(indent,stdout);
        if( fprintf( stdout, "{ " ) != 2 ) xlfail( "write error" ) ;
        nptr = vptr;
        ppaux( nptr->n_listvalue ,indent+2,0) ;
        for(nptr = nptr->n_listnext; nptr != NULL ; nptr = nptr->n_listnext ) {
            fputc( ' ', stdout );
            ppaux( nptr->n_listvalue ,indent+2,0) ;}
        if( fprintf( stdout, "}" ) != 1 ) xlfail( "write error" ) ;
        break ;
    case LIST :
        if(nl==1) indenta(indent,stdout);
        if( fprintf( stdout, "< " ) != 2 ) xlfail( "write error" ) ;
        nptr = vptr;
        ppaux( nptr->n_listvalue ,indent+2,0) ;
        for(nptr = nptr->n_listnext; nptr != NULL ; nptr = nptr->n_listnext ) {
            fputc( ' ', stdout );
            ppaux( nptr->n_listvalue ,indent+2,0) ;}
        if( fprintf( stdout, ">" ) != 1 ) xlfail( "write error" ) ;
        break ;
    case SYM :
        if(nl==1) indenta(indent,stdout);
        if( fprintf( stdout, "%s ", vptr->n_symname ) !=
                              strlen( vptr->n_symname ) + 1 )
            xlfail( "write error" ) ;
        break ;
    case INT :
        if(nl==1) indenta(indent,stdout);
        if( fprintf( stdout, "%d ", vptr->n_int ) <= 0 ) xlfail( "write error" ) ;
        break ;
    case STR :
        if(nl==1) indenta(indent,stdout);
        fprintf(stdout,"%s ", vptr->n_str ) ;
        break ;
    case SUBR :
    case FPTR :
#ifdef COMM
    case CHAN :
#endif
#ifdef USERTYPE
    case TYPE:
    case LABEL:
    case MACTYPE:
#endif
        xlprint( vptr,1) ;
        break;
    case FUN :
        if(nl==1) indenta(indent,stdout);
        fprintf( stdout, "lambda" );
        nptr = vptr;
        xlprint( nptr->n_listvalue,1) ;
        indenta(indent+2,stdout);
        nptr = nptr->n_listnext; 
        xlprint( nptr->n_listvalue,1) ;
        break ;
    default:
#ifdef USERTYPE
        /*if(nl==1) indenta(indent,stdout); */
        indenta(indent,stdout);
        if( vptr->n_type >= USERDEFTYPE ) {
            tdef = name_table[vptr->n_type - USERDEFTYPE] ;
            fprintf( stdout, "%s( ", tdef->tname );
            if( tdef->tnumber == TTUP ) {
                if( ((tl = vptr->n_symvalue) == NULL) || (tl->n_type != LIST) )
                    xlfail( "type mismatch" ) ;
                for( ptup = tdef->def.tcomp ; (ptup != NULL) && (tl != NULL);
                               ptup = ptup->next )
                    { ppaux( tl->n_listvalue ,indent+3,1) ;
                      if( (tl = tl->n_listnext) != NULL )
                        fputc(',',stdout); if(ferror(stdout)) xlfail( "write error" );}
                if( (ptup != NULL) || (tl != NULL) )
                    xlfail( "type mismatch" ) ; }
            else
                ppaux( vptr->n_symvalue ,indent+3,0) ;
            fputc( ')', stdout ) ; if( ferror( stdout ) ) xlfail( "write error" ) ; }
        else
#endif
        xlfail( "bad node type" ) ; }

    /* restore the previous stack frame */
    xlstack = oldstk ;
}

/*. ppn - prity-printer ( argument) .*/
static struct node *ppn( args )
struct node *args ;
{
    struct node arg ;
    struct node *oldstk ;
    struct node *vptr ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, NULL ) ;

    arg.n_ptr = args ;
    vptr = xlevarg( &arg.n_ptr ) ;
    xllastarg( arg.n_ptr ) ;

    ppaux( vptr ,0,0) ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    return( NULL ) ;
}

/*. pp - prity-printer ( argument) .*/
static struct node *pp( args )
struct node *args ;
{
    struct node arg ;
    struct node *oldstk ;
    struct node *vptr ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, NULL ) ;

    arg.n_ptr = args ;
    vptr = xlevarg( &arg.n_ptr ) ;
    xllastarg( arg.n_ptr ) ;

    ppaux( vptr ,0,0) ;
    fputc ('\n',stdout);

    /* restore the previous stack frame */
    xlstack = oldstk ;

    return( NULL ) ;
}


static FILE *fpf ;
static int (*oldgetc)() ;

/*. m2readgetc - read a character from "fpf" .*/
static int m2readgetc()
{
    int c ;

    if( (c = fgetc( fpf )) == EOF )
       xlfail( "end of file" ) ;
    return( c ) ;
}

/*. readf - builtin function readf .*/
/*. readf: expfptr -> sexp .*/

static struct node *readf( args )
struct node *args ;
{
   struct node *res ;

   fpf = xlevmatch( FPTR, &args )->n_fp ;
   xllastarg( args ) ;
   
   if( fpf == NULL )
       xlfail( "file not open" ) ;

   /* save the old input stream */
   oldgetc = xlgetc ;

   xlgetc = m2readgetc ;

   res = xlread( fpf ) ;

   /* restore the old input stream */
   xlgetc = oldgetc ;

   /* return the read expression */
   return( res ) ;
}


/*. m2read - builtin function m2read .*/
/*. m2read: expfptr -> sexp .*/

struct node *m2read( args )
struct node *args ;
{
   struct node *res ;

   fpf = xlevmatch( FPTR, &args )->n_fp ;
   xllastarg( args ) ;
   
   if( fpf == NULL )
       xlfail( "file not open" ) ;

   /* save the old input stream */
   oldgetc = xlgetc ;

   xlgetc = m2readgetc ;

   res = xlread( fpf ) ;

   /* restore the old input stream */
   xlgetc = oldgetc ;

   /* return the read expression */
   return( xleval( res ) ) ;
}


/*. load - builtin function load .*/
/*. load : expstr -> str .*/

static FILE *fp ;

/*. fgetcx - get a character from file "fp" .*/
static int fgetcx()
{
    int c ;

    c = fgetc( fp ) ;
    if( ferror( fp ) ) xlfail( "read error" ) ;
    return( c ) ;
}

struct node *load( args )
struct node *args ;
{
    struct node fname ;
    struct node *oldstk, *p ;

    /* create a new stack frame */
    oldstk = xlsave( &fname, NULL ) ;

    /* get the file name */
    fname.n_ptr = xlevmatch( STR, &args ) ;

    /* make sure there aren't any more arguments */
    xllastarg( args ) ;

    /* push the old input stream on the "oldfp" stack*/
    if( lev == MAX_LEV ) xlfail( "too many nested loads" ) ;
    oldfp[lev++] = fp ;

    if( (fp = fopen( fname.n_ptr->n_str, "r" )) == NULL )
        xlfail( "cannot open file for input" ) ;
    
    /* jj: come as linhas #!... do inicio do ficheiro */
    {char c , s[80];    
    if ((c=getc(fp))=='#') fgets(s,80,fp);      
    else ungetc(c,fp);
    }

    xlgetc = fgetcx ;

    while( (p = xlread( fp ) ) != EOF_mark )
       xleval( p );

    if( fclose( fp ) ) xlfail( "write error when closing file" ) ;

    /* pop */
    if( --lev == 0 )
        xlgetc = tgetc ;
    else
        fp = oldfp[lev] ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return the filename */
    return( fname.n_ptr ) ;
}


/*. read - builtin function read .*/
/*. read : [expstr] -> val .*/

static char *sgetptr ;

/* sgetc - get a character from a string */
static int sgetc()
{
    if( *sgetptr == 0 )
        return( -2 ) ;
    else
        return( (int)*sgetptr++ ) ;
}


struct node *xlreadfromstr( str1 ) 
char *str1;
{
    struct node *val ;
    int (*oldgetc)() ;
    char *str;

    /* save the old input stream */
    oldgetc = xlgetc ;

    str = sgetptr = strsave2( str1 ) ;
    xlgetc = sgetc ; 

    /* read an expression */
    val = xlread( (FILE *)NULL ) ;

    if( str != NULL ) strfree( str ) ;

    /* restore the old input stream */
    xlgetc = oldgetc ;

    /* return the expression read */
    return( val ) ;
}

static struct node *read( args )
struct node *args ;
{
    struct node *val ;
    int (*oldgetc)() ;
    char *str ;

    /* save the old input stream */
    oldgetc = xlgetc ;

    /* get the string or file pointer */
    if( args != NULL ) {
        val = xlevmatch( STR, &args ) ;
        str = sgetptr = strsave( strcat( val->n_str, " " ) ) ;
        xlgetc = sgetc ; }
    else
        str = NULL ;

    /* make sure there aren't any more arguments */
    xllastarg( args ) ;

    /* read an expression */
    val = xlread( (FILE *)NULL ) ;

    if( str != NULL ) strfree( str ) ;

    /* restore the old input stream */
    xlgetc = oldgetc ;

    /* return the expression read */
    return( val ) ;
}


void xmreadinit()
{
    lev = 0 ;
    EOF_mark = xlenter( "    " ) ;

    xlsubr( "load", load ) ;
    xlsubr( "read", read ) ;
    xlsubr( "readf", readf ) ;
    xlsubr( "writef", writef ) ;
    xlsubr( "m2read", m2read ) ;
    xlsubr( "m2write", m2write ) ;
    xlsubr( "nwrite", nwrite ) ;
    xlsubr( "pp", pp ) ;
    xlsubr( "ppn", ppn ) ;
    xlsubr( "setpp", setpp ) ;
}

