/* xmrel.c - binary relation algebra */

#include <stdio.h>
#include "xmetoo.h"


/* external variables */
extern struct node *xlstack ;


/* external procedures */
extern struct node *setfilter() ;
extern struct node *sequence() ;
extern int xequal() ;


/*. relfilter - removes duplicate elements from argument list .*/
struct node *relfilter( list )
struct node *list ;
{
    struct node *act, *prev, *actel ;

    if( list != NULL ) list->n_type = REL ;

    for( actel = list ; actel != NULL ; actel = actel->n_listnext )
        for( prev = actel, act = actel->n_listnext ; act != NULL ;
                                            act = act->n_listnext )
            if( xequal( actel->n_listvalue, act->n_listvalue ) )
                prev->n_listnext = act->n_listnext ;
            else
                prev = act ;

    return( list ) ;
}


/*. rel - builtin function set .*/
/*. rel : nexp fromexp... -> rel .*/
/*.       fromexp = "(from" symb expset [nexpbool] ")" .*/
static struct node *rel( args )
struct node *args ;
{
    return( relfilter( sequence( args ) ) ) ;
}

/*. makerel - builtin function makerel .*/
/*. makerel : pair... -> rel .*/
/*.           pair = "(" exp exp ")" .*/
static struct node *makerel( args )
struct node *args ;
{
    struct node arg, e1, e2, resf, pair ;
    struct node *oldstk, *lresf, *new ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &e1, &e2, &resf, &pair, NULL ) ;

    arg.n_ptr = args ;

    lresf = NULL ;
    while( args != NULL ) {
        pair.n_ptr = xlmatch( LIST, &args ) ;
        e1.n_ptr = xlevarg( &pair.n_ptr ) ;
        e2.n_ptr = xlevarg( &pair.n_ptr ) ;
        xllastarg( pair.n_ptr ) ;
        new = newnode( LIST ) ;
        if( lresf == NULL ) 
            resf.n_ptr = new ;
        else
            lresf->n_listnext = new ;
        lresf = new ;
        lresf->n_listvalue = new = newnode( LIST ) ;
        new->n_listvalue = e1.n_ptr ;
        new->n_listnext = newnode( LIST ) ;
        new->n_listnext->n_listvalue = e2.n_ptr ; }

    /* restore the old stack frame */
    xlstack = oldstk ;

    resf.n_ptr = setfilter( resf.n_ptr ) ;
    if( resf.n_ptr != NULL) resf.n_ptr->n_type = REL ;

    /* return the created rel */
    return( resf.n_ptr ) ;
}


/*. dom - builtin function dom .*/
/*. dom : exprel -> set .*/
/*.       expff  -> set .*/
static struct node *dom( args )
struct node *args ;
{
    struct node rel, ress ;
    struct node *oldstk, *lress, *actr, *new ;

    /* create a new stack frame */
    oldstk = xlsave( &rel, &ress, NULL ) ;

    rel.n_ptr = xlevarg( &args ) ;
    xllastarg( args ) ;

    if( (rel.n_ptr != NULL) && (rel.n_ptr->n_type != REL) &&
                               (rel.n_ptr->n_type != FF) )
        xlfail( "bad argument type" ) ;

    for( lress = NULL, actr = rel.n_ptr ; actr != NULL ;
                                     lress = new, actr = actr->n_listnext ) {
        new = newnode( LIST ) ;
        new->n_listvalue = actr->n_listvalue->n_listvalue ;
        if( lress == NULL )
            ress.n_ptr = new ;
        else
            lress->n_listnext = new ; }

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return pointer to the set */
    return( setfilter( ress.n_ptr ) ) ;
}


/*. ran - builtin function ran .*/
/*. ran : exprel -> set .*/
/*.       expff  -> set .*/
static struct node *ran( args )
struct node *args ;
{
    struct node rel, ress ;
    struct node *oldstk, *lress, *actr, *new ;

    /* create a new stack frame */
    oldstk = xlsave( &rel, &ress, NULL ) ;

    rel.n_ptr = xlevarg( &args ) ;
    xllastarg( args ) ;

    if( (rel.n_ptr != NULL) && (rel.n_ptr->n_type != REL) &&
                               (rel.n_ptr->n_type != FF) )
        xlfail( "bad argument type" ) ;

    for( lress = NULL, actr = rel.n_ptr ; actr != NULL ;
                                     lress = new, actr = actr->n_listnext ) {
        new = newnode( LIST ) ;
        new->n_listvalue = actr->n_listvalue->n_listnext->n_listvalue ;
        if( lress == NULL )
            ress.n_ptr = new ;
        else
            lress->n_listnext = new ; }

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return pointer to the set */
    return( setfilter( ress.n_ptr ) ) ;
}


/*. dr - builtin function dr .*/
/*. dr : exprel expset -> rel .*/
/*.      expff  expset -> ff .*/
static struct node *dr( args )
struct node *args ;
{
    struct node arg, r, s, resr ;
    struct node *oldstk, *ar, *elem, *as, *lresr, *new ;
    int found, t ;

    /*create a new stack frame */
    oldstk = xlsave( &arg, &r, &s, &resr, NULL );

    arg.n_ptr = args ;
    r.n_ptr = xlevarg( &arg.n_ptr ) ;
    s.n_ptr = xlevmatch( SET, &arg.n_ptr ) ;
    xllastarg( arg.n_ptr ) ;

    if( (r.n_ptr != NULL) && ((t = r.n_ptr->n_type) != REL) && (t != FF) )
        xlfail( "bad argument type" ) ;

    for( lresr = NULL, ar = r.n_ptr ; ar != NULL ; ar = ar->n_listnext ) {
        elem = ar->n_listvalue->n_listvalue ;
        for( as = s.n_ptr, found = FALSE ; !found && (as != NULL) ;
                                           as = as->n_listnext )
            if( xequal( elem, as->n_listvalue ) ) {
                found = TRUE ;
                new = newnode( LIST ) ;
                if( lresr == NULL)
                    resr.n_ptr = new ;
                else
                    lresr->n_listnext = new ;
                lresr = new ;
                new = new->n_listvalue = newnode( LIST ) ;
                new->n_listvalue = elem ;
                new->n_listnext = newnode( LIST ) ;
                new->n_listnext->n_listvalue = 
                                ar->n_listvalue->n_listnext->n_listvalue ; } }

    /* restore the previous stack frame */
    xlstack = oldstk ;

    if( resr.n_ptr != NULL) resr.n_ptr->n_type = t ;

    /* return pointer to the rel */
    return( resr.n_ptr ) ;
}


/*. ds - builtin function ds .*/
/*. ds : exprel expset -> rel .*/
/*.      expff  expset -> ff .*/
static struct node *ds( args )
struct node *args ;
{
    struct node arg, r, s, resr ;
    struct node *oldstk, *ar, *elem, *as, *lresr, *new ;
    int found, t ;

    /*create a new stack frame */
    oldstk = xlsave( &arg, &r, &s, &resr, NULL );

    arg.n_ptr = args ;
    r.n_ptr = xlevarg( &arg.n_ptr ) ;
    s.n_ptr = xlevmatch( SET, &arg.n_ptr ) ;
    xllastarg( arg.n_ptr ) ;

    if( (r.n_ptr != NULL) && ((t = r.n_ptr->n_type) != REL) && (t != FF) )
        xlfail( "bad argument type" ) ;

    for( lresr = NULL, ar = r.n_ptr ; ar != NULL ; ar = ar->n_listnext ) {
        elem = ar->n_listvalue->n_listvalue ;
        for( as = s.n_ptr, found = FALSE ; !found && (as != NULL) ;
                                           as = as->n_listnext )
            if( xequal( elem, as->n_listvalue ) )
                found = TRUE ;
        if( !found ) {
            new = newnode( LIST ) ;
            if( lresr == NULL)
                resr.n_ptr = new ;
            else
                lresr->n_listnext = new ;
            lresr = new ;
            new = new->n_listvalue = newnode( LIST ) ;
            new->n_listvalue = elem ;
            new->n_listnext = newnode( LIST ) ;
            new->n_listnext->n_listvalue = 
                            ar->n_listvalue->n_listnext->n_listvalue ; } }

    /* restore the previous stack frame */
    xlstack = oldstk ;

    if( resr.n_ptr != NULL) resr.n_ptr->n_type = t ;

    /* return pointer to the rel */
    return( resr.n_ptr ) ;
}


void xmrelinit()
{
    xlsubr( "makerel", makerel ) ;
    xlsubr( "dom", dom ) ;
    xlsubr( "ran", ran ) ;
    xlsubr( "dr", dr ) ;
    xlsubr( "rel", rel ) ;
    xlsubr( "ds", ds ) ;
}

