/* xmsave.c - module for saving and loading xmetoo state */

#include <stdio.h>
#include <setjmp.h>
#include "xmetoo.h"


/* external variables */
extern jmp_buf xljmpbuf ;
extern int trace_pointer ;
extern int lev ;
extern struct node *xlstack ;
extern struct node *ops ;
extern struct node *obs ;
extern struct node *oblist ;
extern struct node *oblimit ;


/* external procedures */
/* extern void longjmp() ; */
#ifdef COMM
extern struct node *channel() ;
#endif
extern void xlunbind() ;
extern void xltin() ;
#ifdef USERTYPE
extern void load_tt();
extern void save_tt();
#endif


/*. load_init  - state initialization after (re)LOAD .*/
static void load_init()
{
    printf( "\nload process completed, state reinitialized\n" ) ;

    /* unbind bound symbols */
    xlunbind( (struct node *) NULL ) ;

    /* restore input to the terminal */
    xltin( TRUE ) ;

    /* restore "load" depth counter */
    lev = 0 ;

    /* do the back trace */
    trace_pointer = -1 ;

    /* restart */
    longjmp( xljmpbuf, 1 ) ;
}


/*. SAVE - builtin function SAVE .*/
/*. SAVE : expstr -> nil .*/

/*. put_value - write S-expression "sexp" in file "fp" .*/
static void put_value( fp, sexp )
FILE *fp ;
struct node  *sexp ;
{
    struct node *p ;

    if( sexp == NULL ) {
        putw( NILV, fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
        return ; }

    switch( sexp->n_type ) {

    case INT :
        putw( INT, fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
        putw( sexp->n_int, fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
        break ;

    case STR :
        putw( STR, fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
#ifdef __TURBOC__
        if( fputs( sexp->n_str, fp ) == EOF ) xlfail( "write error" ) ;
#else
        if( fputs( sexp->n_str, fp ) == NULL ) xlfail( "write error" ) ;
#endif
        if( fputc( 0, fp ) == EOF ) xlfail( "write error" ) ;
        break ;

    case SYM :
        putw( SYM, fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
#ifdef __TURBOC__
        if( fputs( sexp->n_symname, fp ) == EOF ) xlfail( "write error" ) ;
#else
        if( fputs( sexp->n_symname, fp ) == NULL ) xlfail( "write error" ) ;
#endif
        if( fputc( '\n', fp ) == EOF ) xlfail( "write error" ) ;
        break ;

    case SUBR :
        putw( sexp->n_type, fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
        break ; 
         
    case FPTR :
        putw( FPTR, fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
        break ; 
         
#ifdef COMM
    case CHAN :
        putw( CHAN, fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
#ifdef __TURBOC__
        if( fputs( sexp->n_qname->n_str, fp ) == EOF ) xlfail( "write error" ) ;
#else
        if( fputs( sexp->n_qname->n_str, fp ) == NULL ) xlfail( "write error" ) ;
#endif
        if( fputc( 0, fp ) == EOF ) xlfail( "write error" ) ;
        break ;
#endif

#ifdef USERTYPE
    case TYPE:
        putw( TYPE, fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
        break ;

    case LABEL:
        putw( LABEL, fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
        break ;

    case MACTYPE:
        putw( MACTYPE, fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
        put_value( fp, sexp->n_macdef ) ;       
        break ;
#endif

    case LIST :
    case SET :
    case FUN :
        putw( sexp->n_type, fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
        for( p = sexp ; p != NULL ; p = p->n_listnext )
            put_value( fp, p->n_listvalue ) ;
        putw( SAVE_MARK, fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
        break ;

    case REL :
    case FF :
        putw( sexp->n_type, fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
        for( p = sexp ; p != NULL ; p = p->n_listnext ) {
            put_value( fp, p->n_listvalue->n_listvalue ) ;
            put_value( fp, p->n_listvalue->n_listnext->n_listvalue ); }
        putw( SAVE_MARK, fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
        break ;

#ifdef LAZY
    case REC :
    case LISTREC :
        putw( sexp->n_type, fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
        put_value( fp, sexp->n_listvalue ) ;
        put_value( fp, sexp->n_listnext ) ;
        break ;
#endif

    default :
#ifdef USERTYPE
        if( sexp->n_type >= USERDEFTYPE ) {
            putw( sexp->n_type, fp ) ;
            if( ferror( fp ) ) xlfail( "write error" ) ;
            put_value( fp, sexp->n_symvalue ) ; }
        else
#endif
        fprintf(stderr,"type=%d: ",sexp->n_type);
        xlfail( "bad argument type" ) ; }
}

static struct node *SAVE( args )
struct node *args ;
{
    struct node *fname, *p ;
    FILE *fp ;
    
    fname = xlevmatch( STR, &args ) ;
    xllastarg( args ) ;

#ifdef __TURBOC__
    if( (fp = fopen( fname->n_str, "wb" )) == NULL ) 
#else
    if( (fp = fopen( fname->n_str, "w" )) == NULL ) 
#endif
        xlfail( "cannot open file for output" ) ;
    
    putw( SAVE_MARK, fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
    
    for( p = oblist->n_symvalue ; p != oblimit ; p = p->n_listnext ) {
        putw( SYM, fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
#ifdef __TURBOC__
        if( fputs( p->n_listvalue->n_symname, fp ) == EOF )
#else
        if( fputs( p->n_listvalue->n_symname, fp ) == NULL ) 
#endif
            xlfail( "write error" ) ;
        if( fputc( '\n', fp ) == EOF ) xlfail( "write error" ) ;
        put_value( fp, p->n_listvalue->n_symvalue ) ; }

    p = ops ;
    putw( SYM, fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
#ifdef __TURBOC__
        if( fputs( p->n_symname, fp ) == EOF )
#else
    if( fputs( p->n_symname, fp ) == NULL ) 
#endif
        xlfail( "write error" ) ;
    if( fputc( '\n', fp ) == EOF ) xlfail( "write error" ) ;
    put_value( fp, p->n_symvalue ) ;

    p = obs ;
    putw( SYM, fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
#ifdef __TURBOC__
    if( fputs( p->n_symname, fp ) == EOF )
#else
    if( fputs( p->n_symname, fp ) == NULL ) 
#endif
        xlfail( "write error" ) ;
    if( fputc( '\n', fp ) == EOF ) xlfail( "write error" ) ;
    put_value( fp, p->n_symvalue ) ;

    putw( SAVE_MARK, fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;

#ifdef USERTYPE
    save_tt( fp ) ;
    putw( SAVE_MARK, fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
#endif

    if( fclose( fp ) ) xlfail( "write error when closing file" ) ;

    return( NULL ) ;
} 
    
/*. LOAD - builtin function LOAD .*/
/*. LOAD : expstr -> .*/

/*. get_value - read S-expression from file "fp" .*/
static struct node *get_value( fp )
FILE *fp ;
{
    struct node *oldstk, *new, *last ;
    struct node res ;
    int t, t1, n, c ;
    char buf[STRMAX] ;

    t = getw( fp ) ; if( ferror( fp ) ) xlfail( "read error" ) ;
    
    switch( t ) { 

    case NILV : 
        res.n_ptr = NULL ;
        break ;

    case INT :
        n = getw( fp ) ; if( ferror( fp ) ) xlfail( "read error" ) ;
        res.n_ptr = newnode( INT ) ;
        res.n_ptr->n_int = n ;
        break ;

    case STR :
        c = fgetc( fp ) ; if( ferror( fp ) ) xlfail( "read error" ) ;
        t1 = 0 ;
        while( c ) {
            buf[t1++] = c ;
            c = fgetc( fp ) ; if( ferror( fp ) ) xlfail( "read error" ) ; }
        buf[t1] = '\0' ;
        res.n_ptr = newnode( STR ) ;
        res.n_ptr->n_str = strsave( buf ) ;
        break ;

    case SYM :
        if( fgets( buf, STRMAX, fp ) == NULL ) xlfail( "read error" )  ;
        buf[strlen( buf )-1] = '\0' ;
        res.n_ptr = xlenter( buf ) ; 
        break ;

    case SUBR :
        res.n_ptr = newnode( SUBR ) ;
        break ; 
         
    case FPTR :
        res.n_ptr = newnode( FPTR ) ;
        break ; 
         
#ifdef COMM
    case CHAN :
        c = fgetc( fp ) ; if( ferror( fp ) ) xlfail( "read error" ) ;
        t1 = 0 ;
        while( c ) {
            buf[t1++] = c ;
            c = fgetc( fp ) ; if( ferror( fp ) ) xlfail( "read error" ) ; }
        buf[t1] = '\0' ;
        oldstk = xlsave( &res, NULL ) ;
        res.n_ptr = newnode( LIST ) ;
        res.n_ptr->n_listvalue = newnode( STR ) ;
        res.n_ptr->n_listvalue->n_str = strsave( buf ) ;
        xlstack = oldstk ;
        res.n_ptr = channel( res.n_ptr ) ;
        break ;
#endif
 
#ifdef USERTYPE
    case TYPE :
        res.n_ptr = newnode( TYPE ) ;
        break ;
       
    case LABEL :
        res.n_ptr = newnode( LABEL ) ;
        break ;

    case MACTYPE :
        res.n_ptr = newnode( MACTYPE ) ;
        res.n_ptr->n_macdef = get_value( fp ) ;
        break ;
#endif

    case LIST :
    case SET :
    case FUN :
        oldstk  = xlsave( &res, NULL ) ;
        last = NULL ;
        do {
            t1 = getw( fp ) ; if( ferror( fp ) ) xlfail( "read error" ) ;
            if( t1 != SAVE_MARK ) {
                if( fseek( fp, (long)(-sizeof( int )), 1 ) ) 
                    xlfail( "read error" ) ;
                new = newnode( LIST ) ;
                if( last == NULL )
                   res.n_ptr = new ;
                else
                   last->n_listnext = new ; 
                last = new ;
                new->n_listvalue = get_value( fp ) ; }
         }while( t1 != SAVE_MARK ) ;
         res.n_ptr->n_type = t ; 
         xlstack = oldstk ;
         break ;

    case REL :
    case FF :
        oldstk  = xlsave( &res, NULL ) ;
        last = NULL ;
        do {
            t1 = getw( fp ) ; if( ferror( fp ) ) xlfail( "read error" ) ;
            if( t1 != SAVE_MARK ) {
                if( fseek( fp, (long)(-sizeof( int )), 1 ) ) 
                    xlfail( "read error" ) ;
                new = newnode( LIST ) ;
                if( last == NULL )
                   res.n_ptr = new ;
                else
                   last->n_listnext = new ; 
                last = new ;
                new = new->n_listvalue = newnode( LIST ) ;
                new->n_listvalue = get_value( fp ) ;
                new->n_listnext = newnode( LIST ) ;
                new->n_listnext->n_listvalue = get_value( fp ) ; }
         }while( t1 != SAVE_MARK ) ;
         res.n_ptr->n_type = t ; 
         xlstack = oldstk ;
         break ;

#ifdef LAZY
    case REC :
    case LISTREC :

        oldstk = xlsave( &res, NULL ) ;
        res.n_ptr = newnode( LIST ) ;
        res.n_ptr->n_type = t ;
        res.n_ptr->n_listvalue = get_value( fp ) ;
        res.n_ptr->n_listnext = get_value( fp ) ;
        xlstack = oldstk ; 
        break ;
#endif

    default :
#ifdef USERTYPE
       if( t >= USERDEFTYPE ) {
           oldstk = xlsave( &res, NULL ) ;
           res.n_ptr = newnode( SYM ) ;
           res.n_ptr->n_type = t ;
           res.n_ptr->n_symvalue = get_value( fp ) ; 
           xlstack = oldstk ; }
       else
#endif
       xlfail( "read error" ) ; }

    return( res.n_ptr ) ;
}

static struct node *LOAD( args )
struct node *args ;
{
    struct node *fname, *p ;
    char str_buf[STRMAX] ;
    FILE *fp ;
    int buf ;
   
    fname = xlevmatch( STR, &args ) ;
    xllastarg( args ) ;

#ifdef __TURBOC__
    if( (fp = fopen( fname->n_str, "rb" )) == NULL ) 
#else
    if( (fp = fopen( fname->n_str, "r" )) == NULL ) 
#endif
        xlfail( "cannot open file for input" ) ;
    
    if( (buf = getw( fp )) != SAVE_MARK ) xlfail( "read error" ) ;
    
    buf = getw( fp ) ; if( ferror( fp ) ) xlfail( "read error" ) ;
     
    oblist->n_symvalue = oblimit ;

    while( buf != SAVE_MARK  ) {
        if( buf != SYM ) xlfail( "corrupted state file" ) ;
        if( fgets( str_buf, STRMAX, fp ) == NULL ) xlfail( "read error" ) ;
        str_buf[strlen( str_buf )-1] = '\0' ;
        p = xlenter( str_buf ) ;
        p->n_symvalue = get_value( fp ) ;
        buf = getw( fp ) ; if( ferror( fp ) ) xlfail( "read error" ) ; }
       
#ifdef USERTYPE
    load_tt( fp ) ;
#endif

    if( fclose( fp ) ) xlfail( "error when closing file" ) ;

    load_init() ; 
}


/*. xmsaveinit - enter SAVE and LOAD builtin functions .*/
void xmsaveinit()
{
    xlsubr( "SAVE", SAVE ) ;
    xlsubr( "LOAD", LOAD ) ;
}

