/*	xmsession.c - xmetoo session server & client routines
 *	vff@di.uminho.pt
 *	03/11/95
 */

/* falta "garbage collection dos canais */
/* falta limitar #sessoes e #canais */

#ifdef COMM

#include <unistd.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <errno.h>
#include <stddef.h>
#include <search.h>
#include <stdio.h>

#include "xmsession.h"

#ifndef	EZERO
#define	EZERO	0
#endif

#define	MSGMGR		(0xbabada)
#define	PERMMASK	(0666)

typedef	enum {
	RUNNING,
	NEW_SESSION, JOIN_CHANNEL, LEAVE_CHANNEL, LIST_SESSIONS, LIST_CHANNELS,
	REMOVE_SESSION, STOP_DAEMON
} operation_t;

typedef	struct {
	long		type;
	operation_t	operation;
	pid_t		pid;
	short		session;
	char		channel[MAXCHANNAME];
} request_t;

typedef	struct {
	long		type;
	int		result;
	int		count;
	union {
		short		sessions[MAXSESSIONS];
		channel_t	channels[MAXCHANNELS];
	} u;
} reply_t;

static size_t nsessions = 0;
static session_t sessions[MAXSESSIONS];
static short highersession = 0;

static int compsessions(const void * a, const void * b) {
	return *(short *)a - *(short *)b;
}

static int compchannels(const void * a, const void * b) {
	return strcmp((char *)a, (char *)b);
}

#include <signal.h>

static
void	sigalrm() {
}

int	xmsession_running(void) {
	request_t request = { 1L, RUNNING, getpid(), 0, 0 };
	reply_t reply;
	int daemon;
	errno = EZERO;
	daemon = msgget(MSGMGR, 0);
	msgsnd(daemon, (char *)&request, sizeof(request)-sizeof(long), 0);
	alarm(2);
	signal(SIGALRM, sigalrm);
	msgrcv(daemon, (char *)&reply, sizeof(reply)-sizeof(long),
		(long)getpid(), 0);
	alarm(0);
	return errno == EZERO ? 1 : 0;
}

static
int	D_xmsession_running(int mxid, request_t * request) {
	reply_t reply = { request -> pid, 0, 0 };
	msgsnd(mxid, &reply, sizeof(reply)-sizeof(long), 0);
	return 0;
}

int	new_session(void) {
	request_t request = { 1L, NEW_SESSION, getpid(), 0, 0 };
	reply_t reply;
	int daemon;
	errno = EZERO;
	daemon = msgget(MSGMGR, 0);
	msgsnd(daemon, (char *)&request, sizeof(request)-sizeof(long), 0);
	msgrcv(daemon, (char *)&reply, sizeof(reply)-sizeof(long),
		(long)getpid(), 0);
	return errno == EZERO ? reply.result : -1;
}

static
int	D_new_session(int mxid, request_t * request) {
	reply_t reply = { request -> pid, ++highersession, 0 };
	session_t session;
	memset(&session, 0, sizeof(session_t));
	session.session = nsessions + 1;
	session.nchannels = 0;
	lsearch(&session, sessions, &nsessions, sizeof(session_t), 
		compsessions);
	msgsnd(mxid, (char *)&reply, sizeof(reply)-sizeof(long), 0);
	return 0;
}

int	join_channel(short session, char channel[MAXCHANNAME]) {
	request_t request = { 1L, JOIN_CHANNEL, getpid(), session, "" };
	reply_t reply;
	int daemon;
	errno = EZERO;
	strcpy(request.channel, channel);
	daemon = msgget(MSGMGR, 0);
	msgsnd(daemon, (char *)&request, sizeof(request)-sizeof(long), 0);
	msgrcv(daemon, (char *)&reply, sizeof(reply)-sizeof(long),
		(long)getpid(), 0);
	return errno == EZERO ? reply.result : -1;
}

static
int	D_join_channel(int mxid, request_t * request) {
	reply_t reply = { request -> pid, 0, 0 };
	session_t * session;
	channel_t * channel;
	session_t newsession;
	channel_t newchannel;
	newsession.session = request -> session;
	newsession.nchannels = 0;
	memset(newsession.channels, 0, sizeof(newsession.channels));
	strcpy(newchannel.name, request -> channel);
	newchannel.mxid = 0;
	newchannel.refcount = 0;
	if(request -> session > highersession) 
		highersession = request -> session;
	if((session = (session_t *)lsearch(&newsession, sessions, 
		&nsessions, sizeof(session_t), compsessions)) != NULL
		&&
	   (channel = (channel_t *)lsearch(&newchannel,
		session -> channels, &session -> nchannels, sizeof(channel_t), 
		compchannels)) != NULL) {
			if(channel -> mxid == 0)
				channel -> mxid = msgget(IPC_PRIVATE, PERMMASK);
			channel -> refcount++;
			reply.result = channel -> mxid;
		}
	else reply.result = -1;
	msgsnd(mxid, (char *)&reply, sizeof(reply)-sizeof(long), 0);
	return 0;
}

int	leave_channel(short session, char channel[MAXCHANNAME]) {
	request_t request = { 1L, LEAVE_CHANNEL, getpid(), session, "" };
	reply_t reply;
	int daemon;
	errno = EZERO;
	strcpy(request.channel, channel);
	daemon = msgget(MSGMGR, 0);
	msgsnd(daemon, (char *)&request, sizeof(request)-sizeof(long), 0);
	msgrcv(daemon, (char *)&reply, sizeof(reply)-sizeof(long),
		(long)getpid(), 0);
	return errno == EZERO ? reply.result : -1;
}

static
int	D_leave_channel(int mxid, request_t * request) {
	reply_t reply = { request -> pid, 0, 0 };
	session_t * session;
	channel_t * channel;
	if((session = (session_t *)lfind(&request -> session, sessions, 
		&nsessions, sizeof(session_t), compsessions)) != NULL
		&&
	   (channel = (channel_t *)lfind(request -> channel, 
		session -> channels, &session -> nchannels, sizeof(channel_t), 
		compchannels)) != NULL)
		channel -> refcount--;
	else reply.result = -1;
	msgsnd(mxid, (char *)&reply, sizeof(reply)-sizeof(long), 0);
	return 0;
}

int	list_sessions(short sessions[MAXSESSIONS], int * nsessions) {
	request_t request = { 1L, LIST_SESSIONS, getpid(), 0, "" };
	reply_t reply;
	int daemon;
	errno = EZERO;
	daemon = msgget(MSGMGR, 0);
	msgsnd(daemon, (char *)&request, sizeof(request)-sizeof(long), 0);
	msgrcv(daemon, (char *)&reply, sizeof(reply)-sizeof(long),
		(long)getpid(), 0);
	memcpy(sessions, &reply.u.sessions, sizeof(reply.u.sessions));
	*nsessions = reply.count;
	return errno == EZERO ? reply.result : -1;
}

static
int	D_list_sessions(int mxid, request_t * request) {
	reply_t reply = { request -> pid, 0, 0 };
	for(reply.count = 0; reply.count != nsessions; reply.count++)
		reply.u.sessions[reply.count] = sessions[reply.count].session;
	msgsnd(mxid, (char *)&reply, sizeof(reply)-sizeof(long), 0);
	return 0;
}

int	list_channels(short session, channel_t channels[MAXCHANNELS], 
	int * nchannels)
{
	request_t request = { 1L, LIST_CHANNELS, getpid(), session, "" };
	reply_t reply;
	int daemon;
	errno = EZERO;
	daemon = msgget(MSGMGR, 0);
	msgsnd(daemon, (char *)&request, sizeof(request)-sizeof(long), 0);
	msgrcv(daemon, (char *)&reply, sizeof(reply)-sizeof(long),
		(long)getpid(), 0);
	memcpy(channels, &reply.u.channels, sizeof(reply.u.channels));
	*nchannels = reply.count;
	return errno == EZERO ? reply.result : -1;
}

static
int	D_list_channels(int mxid, request_t * request) {
	reply_t reply = { request -> pid, 0, 0 };
	session_t * session;
	if((session = (session_t *)lfind(&request -> session, sessions, 
		&nsessions, sizeof(session_t), compsessions)) != NULL) {
		reply.count = session -> nchannels;
		memcpy(reply.u.channels, session -> channels, 
			sizeof(session -> channels));
	}
	else reply.result = -1;
	msgsnd(mxid, (char *)&reply, sizeof(reply)-sizeof(long), 0);
	return 0;
}

int	remove_session(short session) {
	request_t request = { 1L, REMOVE_SESSION, getpid(), session, "" };
	reply_t reply;
	int daemon;
	errno = EZERO;
	daemon = msgget(MSGMGR, 0);
	msgsnd(daemon, (char *)&request, sizeof(request)-sizeof(long), 0);
	msgrcv(daemon, (char *)&reply, sizeof(reply)-sizeof(long),
		(long)getpid(), 0);
	return errno == EZERO ? reply.result : -1;
}

static
int	D_remove_session(int mxid, request_t * request) {
	reply_t reply = { request -> pid, 0, 0 };
	session_t * session;
	if((session = (session_t *)lfind(&request -> session, sessions, 
		&nsessions, sizeof(session_t), compsessions)) != NULL) {
		for(; session -> nchannels != 0; session -> nchannels--)
			msgctl(session -> channels[
				session -> nchannels - 1].mxid, IPC_RMID, NULL);
		nsessions--;
		memcpy(session, session + 1, 
			((sessions + nsessions) - session) * sizeof(session_t));
		memset(&(sessions[nsessions]), 0, sizeof(session_t));
	}
	else reply.result = -1;
	msgsnd(mxid, (char *)&reply, sizeof(reply)-sizeof(long), 0);
	return 0;
}

int	stop_daemon(void) {
	request_t request = { 1L, STOP_DAEMON, getpid(), 0, "" };
	reply_t reply;
	int daemon;
	errno = EZERO;
	daemon = msgget(MSGMGR, 0);
	msgsnd(daemon, (char *)&request, sizeof(request)-sizeof(long), 0);
/*	msgrcv(daemon, (char *)&reply, sizeof(reply)-sizeof(long),
		(long)getpid(), 0);*/
	msgctl(daemon, IPC_RMID, NULL);
	nsessions = 0;
	return errno == EZERO ? reply.result : -1;
}

static
int	D_stop_daemon(int mxid, request_t * request) {
	int i, j;
	for(i = 0; i != nsessions; i++)
		for(j = 0; j != sessions[i].nchannels; j++)
			msgctl(sessions[i].channels[j].mxid, IPC_RMID, NULL);
	msgctl(mxid, IPC_RMID, NULL);
}

int	D_start_daemon(void) {
	int mxid = msgget(MSGMGR, PERMMASK);
	request_t request;
/*	atexit(D_stop_daemon);*/
	if(mxid != -1) msgctl(mxid, IPC_RMID, NULL);
	mxid = msgget(MSGMGR, IPC_CREAT + PERMMASK);
	memset(sessions, 0, sizeof(sessions));
	while(msgrcv(mxid, (char *)&request, sizeof(request)-sizeof(long), 1L,
		0) != -1) {
		/*printf("daemon: %d\n", request.operation);*/
		switch(request.operation) {
			case RUNNING:
				D_xmsession_running(mxid, &request); break;
			case NEW_SESSION:
				D_new_session(mxid, &request); break;
			case JOIN_CHANNEL:
				D_join_channel(mxid, &request); break;
			case LEAVE_CHANNEL:
				D_leave_channel(mxid, &request); break;
			case LIST_SESSIONS:
				D_list_sessions(mxid, &request); break;
			case LIST_CHANNELS:
				D_list_channels(mxid, &request); break;
			case REMOVE_SESSION:
				D_remove_session(mxid, &request); break;
			case STOP_DAEMON:
				D_stop_daemon(mxid, &request); break;
			default:;
		}
	}
	fprintf(stderr, "error: xmetoo session daemon exiting...\n");
	D_stop_daemon(-1, NULL);
	return 0;
}

#ifdef	__test_xmsession
int	main(int argc, char * argv[]) {
	if(!xmsession_running()) {
		if(fork() == 0) { D_start_daemon(); exit(0); }
	}
	if(argc > 1 && !strcmp(argv[1], "--new")) {
		int result = new_session();
		if(result == -1) fprintf(stderr, "%s: error!\n", argv[0]);
		else printf("new session id = %d\n", result);
	}
	else
	if(argc > 1 && !strcmp(argv[1], "--join")) {
		int result = join_channel(atoi(argv[2]), argv[3]);
		if(result == -1) fprintf(stderr, "%s: error!\n", argv[0]);
		else printf("message queue id = %d\n", result);
	}
	else
	if(argc > 1 && !strcmp(argv[1], "--leave")) {
		int result = leave_channel(atoi(argv[2]), argv[3]);
		if(result == -1) fprintf(stderr, "%s: error!\n", argv[0]);
	}
	else
	if(argc > 1 && !strcmp(argv[1], "--remove")) {
		int result = remove_session(atoi(argv[2]));
		if(result == -1) fprintf(stderr, "%s: error!\n", argv[0]);
		else printf("session id '%s' removed\n", argv[2]);
	}
	else
	if(argc > 1 && !strcmp(argv[1], "--list")) {
		int nsessions;
		short sessions[MAXSESSIONS];
		int result = list_sessions(sessions, &nsessions);
		if(result == -1) fprintf(stderr, "%s: error!\n", argv[0]);
		else {
			int i, j;
			for(i = 0; i != nsessions; i++) {
				int nchannels;
				channel_t channels[MAXCHANNELS];
				printf("%d: ", sessions[i]);
				list_channels(sessions[i], channels, 
					&nchannels);
				for(j = 0; j != nchannels; j++) {
					printf(" (\"%s\",%d,%d)",
						channels[j].name,
						channels[j].mxid,
						channels[j].refcount);
				}
				printf("\n");
			}
		}
	}
	else
	if(argc > 1 && !strcmp(argv[1], "--stop")) {
		stop_daemon();
		printf("daemon stopped (I'm bad! ;-)\n");
	}
	else
	if(argc > 1 && !strcmp(argv[1], "--start")) {
		if(fork() == 0) D_start_daemon();
		else printf("daemon started!\n");
	}
	else
		fprintf(stderr, "usage: %s options\n"
				"options:  [--help]\n"
				"          [--start]\n"
				"          [--stop]\n"
				"          [--new]\n"
				"          [--join #session \"channel\"]\n"
				"          [--leave #session \"channel\"]\n"
				"          [--remove #session]\n"
				"          [--list]\n", argv[0]);
	return 0;
}
#endif
#endif
