/* xmset.c - set algebra */

#include <stdio.h>
#include "xmetoo.h"


/* external variables */
extern struct node *xlstack ;
extern struct node *true ;
extern struct node *false ;


/* external procedures */
extern int xequal() ;
extern struct node *sequence() ;
extern struct node *xlist() ;


/*. setfilter - removes duplicate elements from argument list .*/
struct node *setfilter( list )
struct node *list ;
{
    struct node *act, *prev, *actel ;

    if( list != NULL ) list->n_type = SET ;

    for( actel = list ; actel != NULL ; actel = actel->n_listnext )
        for( prev = actel, act = actel->n_listnext ; act != NULL ;
                                            act = act->n_listnext )
            if( xequal( actel->n_listvalue, act->n_listvalue ) )
                prev->n_listnext = act->n_listnext ;
            else
                prev = act ;

    return( list ) ;
}


/*. empty - metoo symbol .*/
/*. (implemented by "nil" .*/


/*. makeset - builtin function makeset .*/
/*. makeset : exp... -> set .*/
static struct node *makeset( args )
struct node *args ;
{
    return( setfilter( xlist( args ) ) ) ;
}


/*. unionm2 - builtin function union .*/
/*. union : expset expset -> set .*/
static struct node *unionm2( args )
struct node *args ;
{
    struct node arg, set1, set2, ress ;
    struct node *oldstk, *lress, *act, *new ;
    int t1, t2 ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &set1, &set2, &ress, NULL ) ;

    arg.n_ptr = args ;
    set1.n_ptr = xlevarg( &args ) ;
    set2.n_ptr = xlevarg( &args ) ;
    xllastarg( args ) ;

    t1 = t2 = FREE ;
    if( (set1.n_ptr != NULL) && ((t1 = set1.n_ptr->n_type) != SET) &&
                                (t1 != REL) )
        xlfail( "bad argument type" ) ;
    if( (set2.n_ptr != NULL) && ((t2 = set2.n_ptr->n_type) != SET) &&
                                (t2 != REL) )
        xlfail( "bad argument type" ) ;
    if( (t1 != FREE) && (t2 != FREE) && (t1 != t2) )
        xlfail( "invalid argument" ) ;
   
    for( act = set1.n_ptr, lress =  NULL ; act != NULL ;
                    act = act->n_listnext, lress = new ) {
	new = newnode( LIST ) ;
	new->n_listvalue = act->n_listvalue ;
	if( lress == NULL )
	    ress.n_ptr = new ;
        else
	    lress->n_listnext = new ; }
    for( act = set2.n_ptr ; act != NULL ; 
                         act = act->n_listnext, lress = new ) {
	new = newnode( LIST ) ;
	new->n_listvalue = act->n_listvalue ;
	if( lress == NULL )
	    ress.n_ptr = new ;
        else
	    lress->n_listnext = new ; }
 
    /* restore the previous stack frame */
    xlstack = oldstk ;

    ress.n_ptr = setfilter( ress.n_ptr ) ;
    if( ress.n_ptr != NULL) ress.n_ptr->n_type = (t1 != FREE) ? t1 : t2 ;

    /* return pointer to the filtered list */
    return( ress.n_ptr ) ;
}


/*. intersection - builtin function intersection .*/
/*. intersection : expset expset -> set .*/
/*.                exprel exprel -> rel .*/
static struct node *intersection( args )
struct node *args ;
{
    struct node arg, s1, s2, ress ;
    struct node *new, *act, *oldstk, *actel, *lress ;
    int found, t1, t2 ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &s1, &s2, &ress, NULL ) ; 

    arg.n_ptr = args ;
    s1.n_ptr = xlevarg( &args ) ;
    s2.n_ptr = xlevarg( &args ) ;
    xllastarg( args ) ;

    t1 = t2 = FREE ;
    if( (s1.n_ptr != NULL) && ((t1 = s1.n_ptr->n_type) != SET) &&
                                (t1 != REL) )
        xlfail( "bad argument type" ) ;
    if( (s2.n_ptr != NULL) && ((t2 = s2.n_ptr->n_type) != SET) &&
                                (t2 != REL) )
        xlfail( "bad argument type" ) ;
    if( (t1 != FREE) && (t2 != FREE) && (t1 != t2) )
        xlfail( "invalid argument" ) ;
   
    lress = NULL ;
    for( actel = s1.n_ptr ; actel != NULL ; actel = actel->n_listnext )
        for( found = FALSE, act = s2.n_ptr ; !found && (act != NULL) ;
                                                 act = act->n_listnext )
            if( xequal( actel->n_listvalue, act->n_listvalue ) ) {
                found = TRUE ;
                new = newnode( LIST ) ;
                new->n_listvalue = act->n_listvalue ;
                new->n_listnext = NULL ;
                if( lress == NULL )
                    ress.n_ptr = new ;
                else
                    lress->n_listnext = new ;
                lress = new ; }

    /* restore the previous stack frame */
    xlstack = oldstk ;

    if( ress.n_ptr != NULL) ress.n_ptr->n_type = (t1 != FREE) ? t1 : t2 ;

    /* return the pointer to the result */
    return( ress.n_ptr ) ;
}


/*. difference - builtin function difference .*/
/*. difference : expset expset -> set .*/
/*.              exprel exprel -> rel .*/
static struct node *difference( args )
struct node *args ;
{

    struct node arg, s1, s2, ress ;
    struct node *new, *act, *oldstk, *actel, *lress ;
    int found, t1, t2 ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &s1, &s2, &ress, NULL ) ; 

    arg.n_ptr = args ;
    s1.n_ptr = xlevarg( &args ) ;
    s2.n_ptr = xlevarg( &args ) ;
    xllastarg( args ) ;

    t1 = t2 = FREE ;
    if( (s1.n_ptr != NULL) && ((t1 = s1.n_ptr->n_type) != SET) &&
                                (t1 != REL) )
        xlfail( "bad argument type" ) ;
    if( (s2.n_ptr != NULL) && ((t2 = s2.n_ptr->n_type) != SET) &&
                                (t2 != REL) )
        xlfail( "bad argument type" ) ;
    if( (t1 != FREE) && (t2 != FREE) && (t1 != t2) )
        xlfail( "invalid argument" ) ;
   
    lress = NULL ;
    for( actel = s1.n_ptr ; actel != NULL ; actel = actel->n_listnext ) {
        for( found = FALSE, act = s2.n_ptr ; !found && (act != NULL) ;
                                                 act = act->n_listnext )
            if( xequal( actel->n_listvalue, act->n_listvalue ) )
                found = TRUE ;
        if( !found ) {
            new = newnode( LIST ) ;
            new->n_listvalue = actel->n_listvalue ;
            new->n_listnext = NULL ;
            if( lress == NULL )
                ress.n_ptr = new ;
            else
                lress->n_listnext = new ;
            lress = new ; } }

    /* restore the previous stack frame */
    xlstack = oldstk ;

    if( ress.n_ptr != NULL ) ress.n_ptr->n_type = (t1 != FREE) ? t1 : t2 ;
 
   /* return the pointer to the result */
    return( ress.n_ptr ) ;
}


/*. member - builtin function member .*/
/*. member : exp expset -> bool .*/
/*.          exppair exprel -> bool .*/
static struct node *member( args )
struct node *args ;
{
    struct node arg, e ;
    struct node *oldstk, *s ;
    int bool, type ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &e, NULL ) ;

    arg.n_ptr = args ;
    e.n_ptr = xlevarg( &args ) ;
    s = xlevarg( &args ) ;
    if( (s != NULL) && ((type = s->n_type) != SET) && (type != REL) )
        xlfail( "bad argument type" ) ;
    xllastarg( args ) ;

    if( (type == REL) && ((e.n_ptr == NULL) || (e.n_ptr->n_type != LIST) ||
                     (e.n_ptr->n_listnext == NULL) ||
                     (e.n_ptr->n_listnext->n_listnext != NULL)) )
         xlfail( "bad argument type" ) ;

    /* search for "e" in "s" */
    for( bool = FALSE ; (s != NULL) && !bool ; s = s->n_listnext )
        if( xequal( e.n_ptr, s->n_listvalue ) )
            bool = TRUE ; 

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return pointer to the result */
    return( bool ? true : false ) ;
}


/*. subset - builtin function subset .*/
/*. subset : expset expset -> bool .*/
/*.          exprel exprel -> bool .*/
static struct node *subset( args )
struct node *args ;
{
    struct node arg, s1, s2 ;
    struct node *act, *oldstk, *actel ;
    int memb, found, type ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &s1, &s2, NULL ) ; 

    arg.n_ptr = args ;
    s1.n_ptr = xlevarg( &args ) ;
    s2.n_ptr = xlevarg( &args ) ;
    xllastarg( args ) ;
    if( s1.n_ptr == NULL ){
       if( (s2.n_ptr != NULL)
              && (s2.n_ptr->n_type != SET) && (s2.n_ptr->n_type != REL) )
           xlfail( "bad argument type" ) ; }
    else{
        if( ((type = s1.n_ptr->n_type) != SET) && (type != REL) )
            xlfail( "bad argument type" ) ;
        if( (s2.n_ptr != NULL) && (s2.n_ptr->n_type != type) )
            xlfail( "bad argument type" ) ; }

    for( actel = s1.n_ptr, memb = TRUE ; (actel != NULL) && memb ;
                          memb = found, actel = actel->n_listnext )
        for( act = s2.n_ptr, found = FALSE ; !found && (act != NULL) ;
                                                  act = act->n_listnext )
            found = xequal( actel->n_listvalue, act->n_listvalue ) ; 

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return the pointer to the result */
    return( memb ? true : false ) ;
}


/*. card - builtin function card .*/
/*. card : expset -> int .*/
/*.        exprel -> int .*/
static struct node *card( args )
struct node *args ;
{
    struct node *val ;
    int i ;

    val = xlevarg( &args ) ;
    if( (val != NULL) && (val->n_type != SET) && (val->n_type != REL) )
        xlfail( "bad argument type" ) ;
    xllastarg( args ) ;

    for( i = 0 ; val != NULL ; i++, val = val->n_listnext ) ;

    val = newnode( INT ) ;
    val->n_int = i ;

    /* return pointer to the result */
    return( val ) ;
}


/*. isempty - builtin function isempty .*/
/*. isempty : expset -> bool .*/
/*. isempty : exprel -> bool .*/
/*. isempty : expff -> bool .*/
static struct node *isempty( args ) 
struct node *args ;
{
    struct node *s ;

    s = xlevarg( &args ) ;
    xllastarg( args ) ;

    if( (s != NULL) && (s->n_type != SET) && (s->n_type != REL) &&
                      (s->n_type != FF) )
        xlfail( "bad argument type" ) ;
 
    /* return pointer to the result */
    return( (s == NULL) ? true : false ) ;
}


/*. the - builtin function the .*/
/*. the : expset -> value .*/
/*.       exprel -> value .*/
static struct node *the( args )
struct node *args ;
{
    struct node *s ;

    s = xlevarg( &args ) ;
    if( (s != NULL) && (s->n_type != REL) && (s->n_type != SET) )
        xlfail( "bad argument type" ) ; 
    xllastarg( args ) ;

/*    if( (s == NULL) || (s->n_listnext != NULL) )
        xlfail( "invalid argument" ) ;     */

    if( (s == NULL) || (s->n_listnext != NULL) ) return xlenter("NIL");

    /* return pointer to the set element */
    return( s->n_listvalue ) ;
}


/*. choice - builtin function choice .*/
/*. choice : expset -> value .*/
/*.          exprel -> value .*/
static struct node *choice( args )
struct node *args ;
{
    struct node *s ;

    s = xlevarg( &args ) ;
    if( (s != NULL) && (s->n_type != SET) && (s->n_type != REL) )
        xlfail( "bad argument type" ) ;
    xllastarg( args ) ;

/*    if( s == NULL ) xlfail( "invalid argument" ) ; */
    if( s == NULL ) return xlenter("NIL");

    /* return pointer to the set element */
    return( s->n_listvalue ) ;
}


/*. UNION - builtin function UNION .*/
/*. UNION : expset -> set .*/
static struct node *UNION( args ) 
struct node *args ;
{
    struct node set, ress ;
    struct node *oldstk, *new, *lress, *act ;

    /* create a new stack frame */
    oldstk = xlsave( &set, &ress, NULL ) ;

    /* initialize */
    set.n_ptr = xlevmatch( SET, &args ) ;
    xllastarg( args ) ;

    for( lress = NULL ; set.n_ptr != NULL ; 
                               set.n_ptr = set.n_ptr->n_listnext) {
        if( (set.n_ptr->n_listvalue != NULL) && 
            (set.n_ptr->n_listvalue->n_type != SET) )
            xlfail( "invalid argument" ) ;
        for( act = set.n_ptr->n_listvalue ; act != NULL ;
                                  lress = new, act = act->n_listnext ) {
            /* append this element */
	    new = newnode( LIST ) ;
	    new->n_listvalue = act->n_listvalue ;
	    if( lress == NULL )
	        ress.n_ptr = new ;
            else
	        lress->n_listnext = new ; } }

    /* restore previous stack frame */
    xlstack = oldstk ;

    /* return the list */
    return( setfilter( ress.n_ptr ) ) ;
} 


/*. set - builtin function set .*/
/*. set : nexp fromexp... -> set .*/
/*.       fromexp = "(from" symb expset [nexpbool] ")" .*/
static struct node *set( args )
struct node *args ;
{
    return( setfilter( sequence( args ) ) ) ;
}


/*. reduce - builtin function reduce with set argument .*/
/*. reduce : exp symb expset -> value .*/
/*. (implemented by "reduce" with list argument) .*/


/*. inseg - builtin function inseg .*/
/*. inseg : expint -> int-set .*/
static struct node *inseg( args )
struct node *args ;
{
    struct node il ;
    struct node *oldstk, *ail, *new ;
    int i, n ;
          
    /* create new stack frame */
    oldstk = xlsave( &il, NULL ) ;

    n = xlevmatch( INT, &args )->n_int ;
    xllastarg( args ) ;

    ail = NULL ;
    for( i = 1 ; i <= n ; i++, ail = new ) {
        new = newnode( LIST ) ;
        if( ail == NULL )
            il.n_ptr = new ;
        else
            ail->n_listnext = new ;
        new->n_listvalue = newnode( INT ) ;
        new->n_listvalue->n_int = i ; }

    if( il.n_ptr != NULL ) il.n_ptr->n_type = SET ;

     /* restore the previous stack frame */
     xlstack = oldstk ;

     /* return pointer to the "index list" */   
     return( il.n_ptr ) ;
}


void xmsetinit()
{
    xlenter( "empty" ) ;

    xlsubr( "makeset", makeset ) ;
    xlsubr( "union", unionm2 ) ;
    xlsubr( "intersection", intersection ) ;
    xlsubr( "difference", difference ) ;
    xlsubr( "member", member ) ;
    xlsubr( "subset", subset ) ;
    xlsubr( "card", card ) ;
    xlsubr( "isempty", isempty ) ;
    xlsubr( "the", the ) ;
    xlsubr( "choice", choice ) ;
    xlsubr( "UNION", UNION ) ;
    xlsubr( "set", set ) ;
    xlsubr( "inseg", inseg ) ;
}

