/* xmutype.c - user type module */
#define jjputs(s,fp)  {if( fputs( s, fp ) == EOF ) xlfail( "write error" );} 
#define jjputc(c,fp)  {if( fputc( c, fp ) == EOF ) xlfail( "write error" ) ;}

#ifdef USERTYPE

#include <stdio.h>
#include "xmetoo.h"

/* external variables */
extern struct node *xlstack ;
extern struct node *true ;
extern struct node *false ;


/* external procedures */
extern int strcmp() ;
extern int reserv_symb() ;
extern char *malloc() ;
extern void free() ;
extern void m2rem() ;

/* global variables */
struct t_table *name_table[MAX_TYPE] ;

/* local variables */
static struct node *NIL ;
static struct node *ANY ;
static int name_number ;
static char *null_str ;
static struct t_table *type_table ;
static struct t_table *p_nil, *p_any, *p_int, *p_str, *p_fptr, *p_sym ;
static struct node *recmatch();
#ifdef COMM
static struct t_table *p_chan ;
#endif


/*. labeval - evaluate an expression suppressing the header labels .*/
struct node *labeval( expr )
struct node *expr ;
{
    if( (expr != NULL) && (expr->n_type < USERDEFTYPE) )
        expr = xleval( expr ) ;

    while( (expr != NULL) && (expr->n_type >= USERDEFTYPE) ) 
        expr = expr->n_symvalue ;

    return( expr ) ;
}


/*. userevarg - get the next argument and evaluate it .*/
struct node *userevarg( pargs )
struct node **pargs ;
{
    struct node val ;
    struct node *oldstk ;

    /* create a new stack frame */
    oldstk = xlsave( &val, NULL ) ;

    /* get the argument */
    val.n_ptr = xlarg( pargs ) ;

    /* evaluate the argument */
    val.n_ptr = labeval( val.n_ptr ) ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return the argument */
    return( val.n_ptr ) ;
}


/*. save_tt - saves type_table on file fp .*/

/*. save_antt - saves an anonimous type definition .*/
static void save_antt( fp, tt )
FILE *fp ;
struct t_table *tt ;
{
    if( fputc( '\n', fp ) == EOF ) xlfail( "write error" ) ;
    putw( tt->tnumber, fp ) ; if( ferror( fp ) ) xlfail( "write error" ) ;
    switch( tt->tnumber ) {
    case TSET: case TLIST:
        if( *(tt->def.type->tname) == '\0' )
            save_antt( fp, tt->def.type ) ;
        else {
            jjputs( tt->def.type->tname, fp ); 
            jjputc( '\n', fp ) ; }
        break ;
    case TREL: case TFF:
        if( *(tt->def.rel.dom->tname) == '\0' )
            save_antt( fp, tt->def.rel.dom ) ;
        else {
            jjputs( tt->def.rel.dom->tname, fp );
            jjputc( '\n', fp ) ; }
        if( *(tt->def.rel.ran->tname) == '\0' )
            save_antt( fp, tt->def.rel.ran ) ;
        else {
            jjputs( tt->def.rel.ran->tname, fp );
            jjputc( '\n', fp ) ; }
        break ; }
}

/*. save_tt1 - saves type table .*/
static void save_tt1( fp, tt )
FILE *fp ;
struct t_table *tt ;
{
    struct type *palt ;
    struct lab_type *ptup ;

    for( ; tt != p_nil; tt = tt->next ) {
            putw( tt->tnumber, fp ) ;
            if( ferror( fp ) ) xlfail( "write error" ) ;
            jjputs( tt->tname, fp ) ;
            jjputc( '\n', fp ) ;
            putw( tt->nnumber, fp ) ;
            if( ferror( fp ) ) xlfail( "write error" ) ;
            switch( tt->tnumber ) {
            case TSET: case TLIST:
                if( *(tt->def.type->tname) == '\0' )
                    save_antt( fp, tt->def.type ) ;
                else {
                    jjputs( tt->def.type->tname, fp ) ;
                    jjputc( '\n', fp ); }
                break ;
            case TREL: case TFF:
                if( *(tt->def.rel.dom->tname) == '\0' )
                    save_antt( fp, tt->def.rel.dom ) ;
                else {
                    jjputs( tt->def.rel.dom->tname, fp );
                    jjputc( '\n', fp ) ; }
                if( *(tt->def.rel.ran->tname) == '\0' )
                    save_antt( fp, tt->def.rel.ran ) ;
                else {
                    jjputs( tt->def.rel.ran->tname, fp );
                    jjputc( '\n', fp ) ; }
                break ;
            case TEQ:
                jjputs( tt->def.type->tname, fp ) ;
                jjputc( '\n', fp ) ;
                break ;
            case TALT:
                for( palt = tt->def.acomp; palt != NULL; palt = palt->next ) {
                    putw( TG, fp ) ;
                    if( ferror( fp ) ) xlfail( "write error" ) ;
                    if( *(palt->type->tname) == '\0' )
                        save_antt( fp, palt->type ) ;
                    else {
                        jjputs( palt->type->tname, fp );
                        jjputc( '\n', fp )  ; } }
                 putw( SAVE_MARK, fp ) ;
                 if( ferror( fp ) ) xlfail( "write error" ) ;
                 break ;
            case TTUP:
                for( ptup = tt->def.tcomp; ptup != NULL; ptup = ptup->next ) {
                    putw( TG, fp ) ;
                    if( ferror( fp ) ) xlfail( "write error" ) ;
                    jjputs( ptup->label, fp ) ;
                    jjputc( '\n', fp ) ;
                    if( *(ptup->type->tname) == '\0' )
                        save_antt( fp, ptup->type ) ;
                    else {
                       jjputs( ptup->type->tname, fp ) ;
                       jjputc( '\n', fp )  ; } }
                 putw( SAVE_MARK, fp ) ;
                 if( ferror( fp ) ) xlfail( "write error" ) ;
                 break ; } }
}

void save_tt( fp )
FILE *fp ;
{
    save_tt1( fp, type_table ) ;
}


/*. load_tt - loads type_table from file fp .*/

/*. insert_tt - inserts type tname in type_table .*/
static struct t_table *insert_tt ( tname )
char *tname ;
{
    struct t_table *tt ;

    for( tt = type_table; (tt != p_nil) && (strcmp( tt->tname, tname ) != 0);
                      tt = tt->next ) ;
    if( tt == p_nil ) {
        if( (tt = (struct t_table *)malloc( sizeof( struct t_table ))) == NULL )
            xlfail( "out of heap space" ) ;
        tt->next = type_table ; 
        type_table = tt ;
        tt->tname = xlenter( tname )->n_symname ;
        tt->tnumber = TUNDEF ; }
    return( tt ) ;
}

/*. load_antt - loads an anonimous type definition .*/
static struct t_table *load_antt( fp )
FILE * fp ;
{
    struct t_table *antt ;
    char straux[STRMAX] ;
    int tnumber ;

    tnumber = getw( fp ) ; if( ferror( fp ) ) xlfail( "read error" ) ;
    if( tnumber == TINT )
        antt = p_int ;
    else if( tnumber == TNIL )
        antt = p_nil ;
    else if( tnumber == TSTR )
        antt = p_str ;
    else if( tnumber == TFPTR )

        antt = p_fptr ;
#ifdef COMM
    else if( tnumber == TCHAN )
        antt = p_chan ;
#endif
    else if( tnumber == TSYM )
        antt = p_sym ;
    else {
        if( (antt = (struct t_table *)malloc( sizeof( struct t_table )) )
            == NULL ) xlfail( "out of heap space" ) ;
        antt->next = NULL ;
        antt->tname = null_str ;
        antt->tnumber = tnumber ;
        if( fgets( straux, STRMAX, fp ) == NULL ) xlfail( "read error" )  ;
        straux[strlen( straux )-1] = '\0' ;
        switch( tnumber ) {
        case TLIST: case TSET:
            antt->def.type = (*straux != '\0') ? insert_tt( straux ) :
                                                 load_antt( fp ) ;
            break ;
        case TREL: case TFF:
            antt->def.rel.dom = (*straux != '\0') ? insert_tt( straux ) :
                                                    load_antt( fp ) ;
            if( fgets( straux, STRMAX, fp ) == NULL ) xlfail( "read error" ) ;
            straux[strlen( straux )-1] = '\0' ;
            antt->def.rel.ran = (*straux != '\0') ? insert_tt( straux ) :
                                                    load_antt( fp ) ;
            break ;
        default:
            xlfail( "corrupted file" ) ; } }
    return( antt ) ;
}

void load_tt( fp ) 
FILE *fp ;
{
    struct node *t, *label ;
    struct t_table *newtt ;
    struct type *palt, *last_ac, *new_ac ;
    struct lab_type *ptup, *last_tc, *new_tc ;
    char straux[STRMAX] ;
    int mark, t1 ;

    name_number = USERDEFTYPE ;
    type_table = p_nil ;

    mark = getw( fp ) ; if( ferror( fp ) ) xlfail( "read error" ) ;
    while( mark != SAVE_MARK ) {
        if( fgets( straux, STRMAX, fp ) == NULL ) xlfail( "read error" )  ;
        straux[strlen( straux )-1] = '\0' ;
        newtt = insert_tt( straux ) ;
        newtt->tnumber = mark ;
        newtt->nnumber = getw( fp ) ; if( ferror( fp )) xlfail( "read error" );

        if( newtt->tnumber != TUNDEF ) {
            if( name_number < newtt->nnumber ) name_number = newtt->nnumber ; 
            name_table[newtt->nnumber-USERDEFTYPE] = newtt ; 
            if( ((t = xlenter( straux )->n_symvalue) != NULL) && 
                                                   (t->n_type == TYPE ) )
                t->n_typedef = newtt ; }

        switch( mark ) {
        case TSET: case TLIST:
            if( fgets( straux, STRMAX, fp ) == NULL ) xlfail( "read error" )  ;
            straux[strlen( straux )-1] = '\0' ;
            newtt->def.type = (*straux != '\0') ? insert_tt( straux ) :
                                                  load_antt( fp ) ;
            break ;
        case TREL: case TFF:
            if( fgets( straux, STRMAX, fp ) == NULL ) xlfail( "read error" )  ;
            straux[strlen( straux )-1] = '\0' ;
            newtt->def.rel.dom = (*straux != '\0') ? insert_tt( straux ) :
                                                     load_antt( fp ) ;
            if( fgets( straux, STRMAX, fp ) == NULL ) xlfail( "read error" )  ;
            straux[strlen( straux )-1] = '\0' ;
            newtt->def.rel.ran = (*straux != '\0') ? insert_tt( straux ) :
                                                     load_antt( fp ) ;
            break ;
        case TEQ:
            if( fgets( straux, STRMAX, fp ) == NULL ) xlfail( "read error" )  ;
            straux[strlen( straux )-1] = '\0' ;
            newtt->def.type = insert_tt( straux ) ;
            break ;
        case TALT:
            t1 = getw( fp ) ; if( ferror( fp ) ) xlfail( "read error" ) ;
            last_ac = palt = NULL ;
            while( t1 == TG ) {
                if( (new_ac = (struct type *)malloc( sizeof( struct type )))
                     == NULL ) xlfail( "out of heap space" ) ;
                new_ac->next = NULL ;
                if( last_ac == NULL )
                    palt = new_ac ;
                else
                    last_ac->next = new_ac ;
                last_ac = new_ac ;
                if( fgets( straux, STRMAX, fp ) == NULL )
                    xlfail( "read error" )  ;
                straux[strlen( straux )-1] = '\0' ;
                last_ac->type = (*straux != '\0' ) ? insert_tt( straux ) :
                                                     load_antt( fp ) ; 
                t1 = getw( fp ) ; if( ferror( fp ) ) xlfail( "read error" ) ; }
            newtt->def.acomp = palt ;
            break ;
        case TTUP:
            t1 = getw( fp ) ; if( ferror( fp ) ) xlfail( "read error" ) ;
            last_tc = ptup = NULL ;
            while( t1 == TG ) {
                if( (new_tc = (struct lab_type *)malloc(
                                   sizeof( struct lab_type ))) == NULL )
                    xlfail( "out of heap space" ) ;
                new_tc->next = NULL ;
                if( last_tc == NULL )
                    ptup = new_tc ;
                else
                    last_tc->next = new_tc ;
                last_tc = new_tc ;
                if( fgets( straux, STRMAX, fp ) == NULL )
                    xlfail( "read error" )  ;
                straux[strlen( straux )-1] = '\0' ;
                label = xlenter( straux ) ;
                if( (label->n_symvalue != NULL) &&
                                (label->n_symvalue->n_type == LABEL) )
                    label->n_symvalue->n_symname = label->n_symname ;
                last_tc->label = label->n_symname ;
                if( fgets( straux, STRMAX, fp ) == NULL )
                    xlfail( "read error" )  ;
                straux[strlen( straux )-1] = '\0' ;
                last_tc->type = (*straux != '\0' ) ? insert_tt( straux ) :
                                                     load_antt( fp ) ; 
                t1 = getw( fp ) ; if( ferror( fp ) ) xlfail( "read error" ) ; }
            newtt->def.tcomp = ptup ;
            break ; }
        mark = getw( fp ) ; if( ferror( fp ) ) xlfail( "read error" ) ; }
}


/*. print_type - writes tp definition on file fp .*/

/*. print_type1 - writes tp name or definition on file fp .*/
static void print_type1( tp, fp )
struct t_table *tp ;
FILE *fp ;
{
if( tp->tname == null_str )
    switch( tp->tnumber ) {
    case TNIL:
         if( fprintf( fp, "NIL" ) < 0 )
             xlfail( "write error" ) ;
         break ;
    case TANY:
         if( fprintf( fp, "ANY" ) < 0 )
             xlfail( "write error" ) ;
         break ;
    case TSYM:
         if( fprintf( fp, "SYM" ) < 0 )
             xlfail( "write error" ) ;
         break ;
    case TINT:
         if( fprintf( fp, "INT" ) < 0 )
             xlfail( "write error" ) ;
         break ;
    case TSTR:
         if( fprintf( fp, "STR" ) < 0 )
             xlfail( "write error" ) ;
         break ;
    case TFPTR:
         if( fprintf( fp, "FPTR" ) < 0 )
             xlfail( "write error" ) ;
         break ;
#ifdef COMM
    case TCHAN:
         if( fprintf( fp, "CHAN" ) < 0 )
             xlfail( "write error" ) ;
         break ;
#endif
    case TSET:
         if( fprintf( fp, "SET " ) < 0 )
             xlfail( "write error" ) ;
         print_type1( tp->def.type, fp ) ;
         break ;
    case TLIST:
         if( fprintf( fp, "LIST " ) < 0 )
             xlfail( "write error" ) ;
         print_type1( tp->def.type, fp ) ;
         break ;
    case TREL:
         if( fprintf( fp, "REL " ) < 0 )
             xlfail( "write error" ) ;
         print_type1( tp->def.rel.dom, fp ) ;
         if( fprintf( fp, " " ) < 0 )
             xlfail( "write error" ) ;
         print_type1( tp->def.rel.ran, fp ) ;
         break ;
    case TFF:
         if( fprintf( fp, "FF " ) < 0 )
             xlfail( "write error" ) ;
         print_type1( tp->def.rel.dom, fp ) ;
         if( fprintf( fp, " " ) < 0 )
             xlfail( "write error" ) ;
         print_type1( tp->def.rel.ran, fp ) ;
         break ;
    default:
         xlfail( "bad type" ) ; }
else
    if( fprintf( fp, "%s", tp->tname ) < 0 )
         xlfail( "write error" ) ;
}

void print_type( tp, fp )
struct t_table *tp ;
FILE *fp ;
{
    struct type *pa ;
    struct lab_type *pt ;

    switch( tp->tnumber ) {
    case TUNDEF:
         if( fprintf( fp, "undefined_type" ) < 0 )
             xlfail( "write error" ) ;
         break ;
    case TANY:
         if( fprintf( fp, "ANY" ) < 0 )
             xlfail( "write error" ) ;
         break ;
    case TNIL:
         if( fprintf( fp, "NIL" ) < 0 )
             xlfail( "write error" ) ;
         break ;
    case TSYM:
         if( fprintf( fp, "SYM" ) < 0 )
             xlfail( "write error" ) ;
         break ;
    case TINT:
         if( fprintf( fp, "INT" ) < 0 )
             xlfail( "write error" ) ;
         break ;
    case TSTR:
         if( fprintf( fp, "STR" ) < 0 )
             xlfail( "write error" ) ;
         break ;
    case TFPTR:
         if( fprintf( fp, "FPTR" ) < 0 )
             xlfail( "write error" ) ;
         break ;
#ifdef COMM
    case TCHAN:
         if( fprintf( fp, "CHAN" ) < 0 )
             xlfail( "write error" ) ;
         break ;
#endif
    case TSET:
         if( fprintf( fp, "SET " ) < 0 )
             xlfail( "write error" ) ;
         print_type1( tp->def.type, fp ) ;
         break ;
    case TLIST:
         if( fprintf( fp, "LIST " ) < 0 )
             xlfail( "write error" ) ;
         print_type1( tp->def.type, fp ) ;
         break ;
    case TEQ:
         if( fprintf( fp, "EQ %s", tp->def.type->tname ) < 0 )
             xlfail( "write error" ) ;
         break ;
    case TREL:
         if( fprintf( fp, "REL " ) < 0 )
             xlfail( "write error" ) ;
         print_type1( tp->def.rel.dom, fp ) ;
         if( fprintf( fp, " " ) < 0 )
             xlfail( "write error" ) ;
         print_type1( tp->def.rel.ran, fp ) ;
         break ;
    case TFF:
         if( fprintf( fp, "FF " ) < 0 )
             xlfail( "write error" ) ;
         print_type1( tp->def.rel.dom, fp ) ;
         if( fprintf( fp, " " ) < 0 )
             xlfail( "write error" ) ;
         print_type1( tp->def.rel.ran, fp ) ;
         break ;
    case TALT:
         if( fprintf( fp, "ALT" ) < 0 )
             xlfail( "write error" ) ;
         for( pa = tp->def.acomp; pa != NULL; pa = pa->next ) {
             if( fprintf( fp, " " ) < 0 )
                 xlfail( "write error" ) ;
             print_type1( pa->type, fp ) ; }
         break ;
    case TTUP:
         if( fprintf( fp, "TUP" ) < 0 )
             xlfail( "write error" ) ;
         for( pt = tp->def.tcomp; pt != NULL; pt = pt->next ) {
             if( fprintf( fp, " (%s ", pt->label ) < 0 )
                 xlfail( "write error" ) ;
             print_type1( pt->type, fp ) ;
             if( fprintf( fp, ")" ) < 0 )
                 xlfail( "write error" ) ; }
         break ;
    default:
         xlfail( "bad type" ) ; }
}


/*. deftype - builtin function deftype .*/
/*. deftype : sym sym exptype -> sym .*/

/*. recmatch - tests type of a node .*/
static struct node *recmatch( t, nla )
int t ;
struct node **nla ;
{
    struct node *res ;

    if( (*nla == NULL) || ((*nla)->n_listvalue->n_type != t) )
        return( NULL ) ;
    res = (*nla)->n_listvalue ;
    *nla = (*nla)->n_listnext ;
    return( res ) ;
}


/*. free_type - frees a t_table created by "recognize" .*/
static void free_type( tt )
struct t_table *tt ;
{
    struct t_table* taux ;

    for( taux = tt ; taux != NULL ;) {
        tt = taux ; taux = taux->next ;
        switch( tt->tnumber ) {
        case TUNDEF: case TSYM: case TINT:
        case TNIL: case TSTR: case TFPTR:
#ifdef COMM
        case TCHAN:
#endif
            free( (char *)tt ) ;
            break ;
        case TSET: case TLIST:
            free_type( tt->def.type ) ;
            free( (char *)tt ) ;
            break ;
        case TREL: case TFF:
            free_type( tt->def.rel.dom ) ;
            free_type( tt->def.rel.ran ) ;
            free( (char *)tt ) ;
            break ; } }
}

        
/*. recognize - recognizes a type expression tname in phrase .*/
static struct t_table *recognize( tname, phrase )
char *tname ;
struct node **phrase ;
{
    struct node *t1, *t2 ;
    struct t_table *tp1, *tp2, *trec ;
    int tnum ;

    if( (strcmp( tname, "TUP" ) == 0) || (strcmp( tname, "ALT" ) == 0) ||
        (strcmp( tname, "EQ" ) == 0) )
        return( NULL) ;
    if( strcmp( tname, "ANY" ) == 0 )
        tnum = TANY ;
    else if( strcmp( tname, "NIL" ) == 0 )
        tnum = TNIL ;
    else if( strcmp( tname, "STR" ) == 0 )
        tnum = TSTR ;
    else if( strcmp( tname, "INT" ) == 0 )
        tnum = TINT ;
    else if( strcmp( tname, "FPTR" ) == 0 )
        tnum = TFPTR ;
    else if( strcmp( tname, "SYM" ) == 0 )
        tnum = TSYM ;
#ifdef COMM
    else if( strcmp( tname, "CHAN" ) == 0 )
        tnum = TCHAN ;
#endif
    else if( strcmp( tname, "SET" ) == 0 )
        tnum = TSET ;
    else if( strcmp( tname, "LIST" ) == 0 )
        tnum = TLIST ;
    else if( strcmp( tname, "REL" ) == 0 )
        tnum = TREL ;
    else if( strcmp( tname, "FF" ) == 0 )
        tnum = TFF ;
    else if( reserv_symb( tname ) )
        return( NULL ) ;
    else
        tnum = TUNDEF ;

   switch( tnum ) {
#ifdef COMM
   case TCHAN:
#endif
   case TSTR: case TINT: case TFPTR: case TSYM: case TANY: case TNIL:
        if( (trec = (struct t_table *)malloc( sizeof( struct t_table ) ))
            == NULL ) xlfail( "out of heap space" ) ;
        trec->tname = null_str ;
        trec->tnumber = tnum ;
        trec->next = NULL ;
        break ;
   case TLIST: case TSET:
        if ((t1=recmatch  (SYM, phrase ))== NULL) /* jj/jgrenhas */
           t1=xlevmatch (SYM, phrase ) ;
        if( (tp1 = recognize( t1->n_symname, phrase )) == NULL )
            return( NULL ) ;
        if( (trec = (struct t_table *)malloc( sizeof( struct t_table ) ))
            == NULL ) xlfail( "out of heap space" ) ;
        trec->tname = null_str ;
        trec->tnumber = tnum ;
        trec->def.type = tp1 ;
        trec->next = NULL ;
        break ;
   case TREL: case TFF:
        if ((t1=recmatch  (SYM, phrase ))== NULL) /* jj/jgrenhas */
           t1=xlevmatch (SYM, phrase ) ;
        if( (tp1 = recognize( t1->n_symname, phrase )) == NULL )
            return( NULL ) ;
        if ((t2=recmatch  (SYM, phrase ))== NULL) /* jj/jgrenhas */
           t2=xlevmatch (SYM, phrase ) ;
        if( (tp2 = recognize( t2->n_symname, phrase )) == NULL ) {
            free_type( tp1 ) ;
            return( NULL ) ; }
        if( (trec = (struct t_table *)malloc( sizeof( struct t_table ) ))
            == NULL ) xlfail( "out of heap space" ) ;
        trec->tname = null_str ;
        trec->tnumber = tnum ;
        trec->def.rel.dom = tp1 ;
        trec->def.rel.ran = tp2 ;
        trec->next = NULL ;
        break ;
    case TUNDEF:
        if( (trec = (struct t_table *)malloc( sizeof( struct t_table ) ))
            == NULL ) xlfail( "out of heap space" ) ;
        trec->tname = tname ;
        trec->tnumber = TUNDEF ;
        trec->next = NULL ;
        break ; }
        
    return( trec ) ;
}


/*. obl_type - creates a TYPE node linking sym to tdef .*/
static void obl_type( tdef, sym )
struct t_table *tdef ;
struct node *sym ;
{
    struct node *t ;

    if( (sym->n_symvalue == NULL) || (sym->n_symvalue->n_type != TYPE) ) {
        t = newnode( TYPE ) ;
        t->n_typedef = tdef ;
        sym->n_symvalue = t ; }
}


/*. ins_tdef1 - inserts type definition t in type_table .*/
static struct t_table *ins_tdef1( t )
struct t_table *t;
{
    struct t_table *p ;
    struct lab_type *pt1, *pt2 ;
    struct type *pa1, *pa2 ;

    for( p = type_table; (p != NULL) &&
                  (strcmp( p->tname, t->tname ) != 0); p = p->next ) ;
    if( p == NULL ) {
        if( (p = (struct t_table *)malloc( sizeof( struct t_table ) ))
            == NULL ) xlfail( "out of heap space" ) ;
        p->next = type_table ;
        type_table = p ;
        p->tname = t->tname ;
        if( name_number - USERDEFTYPE == MAX_TYPE ) {
            free( (char *) t ) ;
            xlfail( "too many types defined, corrupted type table" ) ; }
        p->nnumber = name_number ;
        name_table[name_number++ - USERDEFTYPE] = p ; }
    else
        if( p->tnumber == TUNDEF ) {
            if( name_number - USERDEFTYPE == MAX_TYPE ) {
                free( (char *) t ) ;
                xlfail( "too many types defined, corrupted type table" ) ; }
            p->nnumber = name_number ;
            name_table[name_number++ - USERDEFTYPE] = p ; }
        else if( p->tnumber == TTUP )
            for( pt1 = p->def.tcomp; pt1 != NULL; ) {
                pt2 = pt1 ;
                pt1 = pt1->next ;
                free( (char *)pt2 ) ; }
        else if( p->tnumber == TALT )
            for( pa1 = p->def.acomp; pa1 != NULL; ) {
                pa2 = pa1 ;
                pa1 = pa1->next ;
                free( (char *)pa2 ) ; }
    p->tnumber = t->tnumber ; 

    switch( p->tnumber ) {
    case TEQ: case TSET: case TLIST:
        p->def.type = t->def.type ;
        break ;
    case TREL: case TFF:
        p->def.rel.dom = t->def.rel.dom ;
        p->def.rel.ran = t->def.rel.ran ;
        break ;
    case TALT:
        p->def.acomp = t->def.acomp ;
        break ;
    case TTUP:
        p->def.tcomp = t->def.tcomp ;
        break ; }
    
    return( p ) ;
}


/*. ins_tdef - merges t_table tdef with type_table .*/
static struct t_table *ins_tdef( tdef )
struct t_table *tdef ;
{
    struct t_table *p ;

    switch( tdef->tnumber ) {
    case TUNDEF:
        for( p = type_table; (p != NULL) &&
                      (strcmp( p->tname, tdef->tname ) != 0); p = p->next ) ;
        if( p == NULL ) {
            if( (p = (struct t_table *)malloc( sizeof( struct t_table ) ))
                == NULL ) xlfail( "out of heap space" ) ;
            p->next = type_table ;
            type_table = p ;
            p->tname = tdef->tname ;
            p->tnumber = TUNDEF ; }
        break ;
#ifdef COMM
    case TCHAN:
        p = p_chan ;
        break ;
#endif
    case TNIL:
        p = p_nil ;
        break ;
    case TANY:
        p = p_any ;
        break ;
    case TSYM:
        p = p_sym ;
        break ;
    case TINT:
        p = p_int ;
        break ;
    case TSTR:
        p = p_str ;
        break ;
    case TFPTR:
        p = p_fptr ;
        break ;
    case TLIST: case TSET:
        if( (p = (struct t_table *)malloc( sizeof( struct t_table ) ))
            == NULL ) xlfail( "out of heap space" ) ;
        p->next = NULL ;
        p->tname = null_str ;
        p->tnumber = tdef->tnumber ;
        p->def.type = ins_tdef( tdef->def.type ) ;
        break ;
    case TREL: case TFF:
        if( (p = (struct t_table *)malloc( sizeof( struct t_table ) ))
            == NULL ) xlfail( "out of heap space" ) ;
        p->next = NULL ;
        p->tname = null_str ;
        p->tnumber = tdef->tnumber ;
        p->def.rel.dom = ins_tdef( tdef->def.rel.dom ) ;
        p->def.rel.ran = ins_tdef( tdef->def.rel.ran ) ;
        break ;
    default:
        xlfail( "bad type" ) ; }

    return( p ) ;
}


/*. process_tuple - inserts tuple type with name tn, labels lblist and
                   components cplist .*/
static struct t_table *process_tuple( tn, lblist, cplist )
char *tn ;
struct node *lblist ;
struct t_table *cplist ;
{
    struct node lb ;
    struct node *oldstk, *label ;
    struct t_table *tp, *tup ;
    struct lab_type *last_comp, *lcomp, *newcomp ;
    struct type *last_t, *tlist, *newt, *tl ;

    oldstk = xlsave( &lb, NULL ) ;

    for( tp = cplist, last_t = NULL, tlist = NULL; tp != NULL;
                                  last_t = newt, tp = tp->next ) {
        if( (newt = (struct type *)malloc( sizeof( struct type ) ))
            == NULL ) xlfail( "out of heap space" ) ;
        newt->type = ins_tdef( tp ) ;
        newt->next = NULL ;
        if( last_t == NULL )
            tlist = newt ;
        else
            last_t->next = newt ; }
    free_type( cplist ) ;
    if( (tp = (struct t_table *)malloc( sizeof( struct t_table ) ))
        == NULL ) xlfail( "out of heap space" ) ;
    tp->tname = tn ;
    tp->tnumber = TTUP ;
    tp->next = NULL ;
    for( lb.n_ptr = lblist, last_comp = lcomp = NULL ; lb.n_ptr != NULL ;
                      lb.n_ptr = lb.n_ptr->n_listnext, last_comp = newcomp ) {
        tl = tlist ;
        tlist = tlist->next ;
        if( (newcomp = (struct lab_type *)malloc( sizeof( struct lab_type ) ))
            == NULL ) xlfail( "out of heap space" ) ;
        if( last_comp == NULL )
            lcomp = newcomp ;
        else
            last_comp->next = newcomp ;
        newcomp->label = lb.n_ptr->n_listvalue->n_symname ;
        newcomp->type = tl->type ;
        newcomp->next = NULL ;
        free( (char *) tl ) ;
        label = newnode( LABEL ) ;
        label->n_symname = lb.n_ptr->n_listvalue->n_symname ;
        lb.n_ptr->n_listvalue->n_symvalue = label ; }
    tp->def.tcomp = lcomp ;
    tup = ins_tdef1( tp ) ;
    free( (char *) tp ) ;
  
    xlstack = oldstk ;

    return( tup ) ;
}


/*. process_alt - inserts alternative type with name an and components cplist .*/
static struct t_table *process_alt( an, cplist )
char *an ;
struct t_table *cplist ;
{
    struct t_table *a, *tp, *alt ;
    struct type *last_alt, *altlist, *newalt ;

    if( (a = (struct t_table *)malloc( sizeof( struct t_table ) ))
        == NULL ) xlfail( "out of heap space" ) ;
    a->tname = an ;
    a->tnumber = TALT ;
    a->next = NULL ;
    for( tp = cplist, last_alt = altlist = NULL ; tp != NULL ;
                                     tp = tp->next, last_alt = newalt ) {
        if( (newalt = (struct type *)malloc( sizeof( struct type ) ))
            == NULL ) xlfail( "out of heap space" ) ;
        if( last_alt == NULL )
            altlist = newalt ;
        else
            last_alt->next = newalt ;
        newalt->type = ins_tdef( tp ) ;
        newalt->next = NULL ; }
    free_type( cplist ) ;
    a->def.acomp = altlist ;
    alt = ins_tdef1( a ) ;
    free( (char *) a ) ;
    return( alt );
}


/*. process_sle - inserts set/list/equality type with name slen
                 and element type definition el_def .*/
static struct t_table *process_sle( slen, sle, el_def )
char *slen ;
int sle ;
struct t_table *el_def ;
{
    struct t_table *def_sle, *sle_res ;

    if( (def_sle = (struct t_table *)malloc( sizeof( struct t_table ) ))
        == NULL ) xlfail( "out of heap space" ) ;
    def_sle->tname = slen ;
    def_sle->tnumber = sle ;
    def_sle->next = NULL ;
    def_sle->def.type = ins_tdef( el_def ) ;
    free( (char *) el_def ) ;
    sle_res = ins_tdef1( def_sle ) ;
    free( (char *) def_sle ) ;
    return( sle_res );
}


/*. process_rf - inserts rel/finite function type with name rfn
                and elements definition el_def .*/
static struct t_table *process_rf( rfn, rf, el_def )
char *rfn ;
int rf ;
struct t_table *el_def ;
{
    struct t_table *def_rf, *rf_res ;

    if( (def_rf = (struct t_table *)malloc( sizeof( struct t_table ) ))
        == NULL ) xlfail( "out of heap space" ) ;
    def_rf->tname = rfn ;
    def_rf->tnumber = rf ;
    def_rf->next = NULL ;
    def_rf->def.rel.dom = ins_tdef( el_def ) ;
    def_rf->def.rel.ran = ins_tdef( el_def->next ) ;
    free_type( el_def ) ;
    rf_res = ins_tdef1( def_rf ) ;
    free( (char *) def_rf ) ;
    return( rf_res );
}

static struct node *deftype( args )
struct node *args ;
{
    struct node arg, lb_list ;
    struct node *oldstk, *t, *t1, *t2, *type, *lb, *last_lb,
                *l, *new_lb ;
    struct t_table *tpn, *tp1, *tp2, *last_tp, *tdef ;
    int tnum, found ;

    if((t=recmatch  (SYM, &args ))== NULL) /* jj/jgrenhas */
        t=xlevmatch (SYM, &args ) ;

    if( reserv_symb( t->n_symname ) )
        xlfail( "cannot redefine symbol" ) ;
    type = xlmatch( SYM, &args ) ;

    if( strcmp( type->n_symname, "ANY" ) == 0 )
        tnum = TANY ;
    else if( strcmp( type->n_symname, "NIL" ) == 0 )
        tnum = TNIL ;
    else if( strcmp( type->n_symname, "SYM" ) == 0 )
        tnum = TSYM ;
    else if( strcmp( type->n_symname, "INT" ) == 0 )
        tnum = TINT ;
    else if( strcmp( type->n_symname, "STR" ) == 0 )
        tnum = TSTR ;
    else if( strcmp( type->n_symname, "FPTR" ) == 0 )
        tnum = TFPTR ;
#ifdef COMM
    else if( strcmp( type->n_symname, "CHAN" ) == 0 )
        tnum = TCHAN ;
#endif
    else if( strcmp( type->n_symname, "EQ" ) == 0 )
        tnum = TEQ ;
    else if( strcmp( type->n_symname, "LIST" ) == 0 )
        tnum = TLIST ;
    else if( strcmp( type->n_symname, "SET" ) == 0 )
        tnum = TSET ;
    else if( strcmp( type->n_symname, "REL" ) == 0 )
        tnum = TREL ;
    else if( strcmp( type->n_symname, "FF" ) == 0 )
        tnum = TFF ;
    else if( strcmp( type->n_symname, "ALT" ) == 0 )
        tnum = TALT ;
    else if( strcmp( type->n_symname, "TUP" ) == 0 )
        tnum = TTUP ;
    else
        xlfail( "invalid type expression" ) ;

    switch( tnum ) {
    
#ifdef COMM
    case TCHAN:
#endif
    case TSYM: case TINT: case TSTR: case TFPTR: case TANY: case TNIL:
        xllastarg( args ) ;
        if( (tp1 = (struct t_table *)malloc( sizeof( struct t_table ) ))
            == NULL ) xlfail( "out of heap space" ) ;
        tp1->tname = t->n_symname ;
        tp1->tnumber = tnum ;
        tp1->next = NULL ;
        tdef = ins_tdef1( tp1 ) ; free( (char *)tp1 ) ;
        obl_type( tdef, t ) ;
        break ;

    case TEQ:
        if((t1=recmatch  (SYM, &args ))== NULL) /* jj/jgrenhas */
            t1=xlevmatch (SYM, &args ) ;
        xllastarg( args ) ;
        if( reserv_symb( t1->n_symname ) )
            xlfail( "may not redefine symbol" ) ;
        if( (tp1 = (struct t_table *)malloc( sizeof( struct t_table ) ))
            == NULL ) xlfail( "out of heap space" ) ;
        tp1->tname = t1->n_symname ;
        tp1->tnumber = TUNDEF ;
        tp1->next = NULL ;
        tdef = process_sle( t->n_symname, TEQ, tp1 ) ;
        obl_type( tdef, t ) ;
        break ;

    case TLIST: case TSET:
        if((t1=recmatch  (SYM, &args ))== NULL) /* jj/jgrenhas */
            t1=xlevmatch (SYM, &args ) ;
        if( (tp1 = recognize( t1->n_symname, &args )) == NULL )
            xlfail( "invalid type expression" ) ;
        xllastarg( args ) ;
        tdef = process_sle( t->n_symname, tnum, tp1 ) ;       
        obl_type( tdef, t ) ;
        break ;

    case TREL: case TFF:
        if((t1=recmatch  (SYM, &args ))== NULL) /* jj/jgrenhas */
            t1=xlevmatch (SYM, &args ) ;
        if( (tp1 = recognize( t1->n_symname, &args )) == NULL )
            xlfail( "invalid type expression" ) ;
        if( (t2 = recmatch( SYM, &args )) == NULL) /* jj/jgrenhas */
            t2=xlevmatch (SYM, &args ) ;
        if( (tp2 = recognize( t2->n_symname, &args )) == NULL ) {
            free( (char *) tp1 ) ;
            xlfail( "invalid type expression" ) ; }
        xllastarg( args ) ;
        tp1->next = tp2 ;
        tdef = process_rf( t->n_symname, tnum, tp1 ) ;
        obl_type( tdef, t ) ;
        break ;

    case TALT:
        if((t1=recmatch  (SYM, &args ))== NULL) /* jj/jgrenhas */
            t1=xlevmatch (SYM, &args ) ;
        if( strcmp( t1->n_symname, "NIL" ) == 0 ) {
            if( (tp1 = (struct t_table *)malloc( sizeof( struct t_table ) ))
                == NULL ) xlfail( "out of heap space" ) ;
            tp1->tname = null_str ;
            tp1->tnumber = TNIL ;
            tp1->next = NULL ;
            if ((t1=recmatch  (SYM, &args ))== NULL) /* jj/jgrenhas */
               t1=xlevmatch (SYM, &args ) ;
	    }
        else
            tp1 = NULL ;
        if( (tpn = recognize( t1->n_symname, &args )) == NULL )
            xlfail( "invalid type expression" ) ;
        if( tp1 != NULL ) 
            tp1->next = tpn ;
        else
            tp1 = tpn ; 
        for( last_tp = tpn; args != NULL; last_tp = tp2 ) {
           if ((t1=recmatch  (SYM, &args ))== NULL) /* jj/jgrenhas */
              t1=xlevmatch (SYM, &args ) ;
            if( (tp2 = recognize( t1->n_symname, &args )) == NULL ) {
                free_type( tp1) ;
                xlfail( "invalid type expression" ) ; }
            last_tp->next = tp2 ; }
        tdef = process_alt( t->n_symname, tp1 ) ;
        obl_type( tdef, t ) ;
        break ;

    case TTUP:
        oldstk = xlsave( &arg, &lb_list, NULL ) ;
        arg.n_ptr = args ;
        l = xlmatch( LIST, &arg.n_ptr ) ;
        lb = xlmatch( SYM, &l ) ;
        if( reserv_symb( lb->n_symname ) )
            xlfail( "may not redefine symbol" ) ;
           /* identificador do tipo: */
        if( (t1 = recmatch( SYM, &l )) == NULL ) /* jj/jgrenhas */
            t1=xlevmatch (SYM, &l ) ;
        if( (tp1 = recognize( t1->n_symname, &l )) == NULL )
            xlfail( "invalid type expression" ) ;
        if( l != NULL ) {
            free_type( tp1 ) ;
            xlfail( "invalid type expression" ) ; }
        lb_list.n_ptr = newnode( LIST ) ;
        lb_list.n_ptr->n_listvalue = lb ;
        for( last_lb = lb_list.n_ptr, last_tp = tp1; arg.n_ptr != NULL;
                                     last_lb = new_lb, last_tp = tp2 ) {
            if( (l = recmatch( LIST, &arg.n_ptr )) == NULL ) {
                free_type( tp1 ) ;
                xlfail( "invalid type expression" ) ; }
		/* etiqueta: */
            if( (lb = recmatch( SYM, &l )) == NULL ) {
                free_type( tp1 ) ;
                xlfail( "invalid type expression" ) ; }
            if( reserv_symb( lb->n_symname ) )
                xlfail( "may not redefine symbol" ) ;
                /* identificador do tipo: */
            if( (t1 = recmatch( SYM, &l )) == NULL ) /* jj/jgrenhas */
                t1=xlevmatch (SYM, &l ) ;
            if( (tp2 = recognize( t1->n_symname, &l )) == NULL ) {
                free_type( tp1 ) ;
                xlfail( "invalid type expression" ) ; }
            if( l != NULL ) {
                free_type( tp1 ) ;
                xlfail( "invalid type expression" ) ; }
            for( l = lb_list.n_ptr, found = FALSE; l != NULL && !found;
                                                     l = l->n_listnext )
                 found = (l->n_listvalue == lb) ;
            if( found ) xlfail( "tuple with duplicate labels" ) ;
            new_lb = newnode( LIST ) ;
            new_lb->n_listvalue = lb ;
            last_lb->n_listnext = new_lb ;
            last_tp->next = tp2 ; }
        tdef = process_tuple( t->n_symname, lb_list.n_ptr, tp1 ) ;
        obl_type( tdef, t ) ;
        xlstack = oldstk ;
        break ; 
    }

    m2rem( t->n_symname ) ;

    return( t ) ;
}

/*. type_match - tests if expression "arg" is of "tdef" type .*/
static int type_match( arg, tdef )
struct node *arg ;
struct t_table *tdef ;
{
    struct node *pn ;
    struct type *palt ;
    int succ ;

    if( tdef->tnumber == TANY ) return( 1 ) ;

    if( (arg != NULL) && (arg->n_type >= USERDEFTYPE) ) {
        if( tdef->nnumber == arg->n_type )
            return( TRUE ) ;
        switch( tdef->tnumber ) {
        case TEQ:
        /* gets the 'real' type definition at the end of the "EQ"-chain */
            while( tdef->tnumber == TEQ )
                tdef = tdef->def.type ;
            return( type_match( arg, tdef ) ) ;
        case TALT:
            for( palt = tdef->def.acomp, succ = FALSE;
                 (!succ) && (palt != NULL); palt = palt->next )
                 if( *(palt->type->tname) != '\0' )
                     succ = type_match( arg, palt->type ) ;
            return( succ ) ; 
        default:
            return( FALSE ) ; } }
        
    switch( tdef->tnumber ) {
    case TNIL:
        return( arg == NIL ) ;
    case TINT:
        if( arg == NULL )
            return( FALSE ) ;
        return( arg->n_type == INT ) ;
    case TSTR:
        if( arg == NULL )
            return( FALSE ) ;
        return( arg->n_type == STR ) ;
    case TFPTR:
        if( arg == NULL )
            return( FALSE ) ;
        return( arg->n_type == FPTR ) ;
    case TSYM:
        if( arg == NULL )
            return( FALSE ) ;
        return( arg->n_type == SYM ) ;
#ifdef COMM
    case TCHAN:
        if( arg == NULL )
            return( FALSE ) ;
        return( arg->n_type == CHAN ) ;
#endif
    case TSET:
        if( arg == NULL )
            return( TRUE ) ;
        if( arg->n_type != SET )
            return( FALSE ) ;
        for( pn = arg; 
             (pn != NULL) && type_match( pn->n_listvalue, tdef->def.type );
             pn = pn->n_listnext) ;
        return( pn == NULL ) ;
    case TLIST:
        if( arg == NULL )
            return( TRUE ) ;
        if( arg->n_type != LIST )
            return( FALSE ) ;
        for( pn = arg; 
             (pn != NULL) && type_match( pn->n_listvalue, tdef->def.type );
             pn = pn->n_listnext) ;
        return( pn == NULL ) ;
    case TREL:
        if( arg == NULL )
            return( TRUE ) ;
        if( arg->n_type != REL )
            return( FALSE ) ;
        for( pn = arg; 
             (pn != NULL) && 
             type_match( pn->n_listvalue->n_listvalue, tdef->def.rel.dom )
 && type_match( pn->n_listvalue->n_listnext->n_listvalue, tdef->def.rel.ran );
             pn = pn->n_listnext) ;
        return( pn == NULL ) ;
    case TFF:
        if( arg == NULL )
            return( TRUE ) ;
        if( arg->n_type != FF )
            return( FALSE ) ;
        for( pn = arg; 
             (pn != NULL) && 
             type_match( pn->n_listvalue->n_listvalue, tdef->def.rel.dom )
 && type_match( pn->n_listvalue->n_listnext->n_listvalue, tdef->def.rel.ran );
             pn = pn->n_listnext) ;
        return( pn == NULL ) ;
    case TEQ:
        /* gets the 'real' type definition at the end of the "EQ"-chain */
        while( tdef->tnumber == TEQ )
            tdef = tdef->def.type ;
        return( type_match( arg, tdef ) ) ; 
    case TALT:
        for( palt = tdef->def.acomp, succ = FALSE;
             (!succ) && (palt != NULL); palt = palt->next )
             succ = type_match( arg, palt->type ) ;
        return( succ ) ; 
    default:
        return( FALSE ) ; } 
}


/*. construct - constructs a node with type "tdef" and value "args" .*/

/*. cons_atom - constructs an atomic type node .*/
static struct node *cons_atom( tdef, tvalue )
struct t_table *tdef ;
struct node *tvalue ;
{
    struct node *patom ;

    if( *(tdef->tname) == '\0' )
        return( tvalue ) ;

    patom = newnode( tdef->nnumber ) ;
    patom->n_symvalue = tvalue ;
    return( patom ) ;
}

struct node *construct( tdef, args )
struct t_table *tdef ;
struct node *args ;
{
    struct node *oldstk, *res, *last_c, *newc ;
    struct node arg, resaux ; 
    struct t_table *tdef1 ;
    struct lab_type *ptup ;
    struct type *palt ;
    int succ ;

    oldstk = xlsave( &arg, &resaux, NULL ) ;
    arg.n_ptr = args ;
 
    switch( tdef->tnumber ) {
    
    case TANY:
        arg.n_ptr = xljustevarg( &args ) ;
        xllastarg( args ) ;
        res = cons_atom( tdef, arg.n_ptr ) ;  
        break ;
	
    case TINT:
        arg.n_ptr = xljustevarg( &args ) ;
        xllastarg( args ) ;
        if( (arg.n_ptr == NULL) || (arg.n_ptr->n_type != INT) ) /* cjr */
            xlfail( "expression-type mismatch");
        res = cons_atom( tdef, arg.n_ptr ) ;  
        break ;

    case TSTR:
        arg.n_ptr = xljustevarg( &args ) ;
        xllastarg( args ) ;
        if( (arg.n_ptr == NULL) || (arg.n_ptr->n_type != STR) ) /* cjr */
            xlfail( "expression-type mismatch");
        res = cons_atom( tdef, arg.n_ptr ) ;
        break ;

    case TSYM:
        arg.n_ptr = xljustevarg( &args ) ;
        xllastarg( args ) ;
        if( (arg.n_ptr == NULL) || (arg.n_ptr->n_type != SYM) ) /* cjr */
            xlfail( "expression-type mismatch");
        res = cons_atom( tdef, arg.n_ptr ) ;  
        break ;
    case TFPTR:
        arg.n_ptr = xljustevarg( &args ) ;
        xllastarg( args ) ;
        if( (arg.n_ptr == NULL) || (arg.n_ptr->n_type != FPTR) ) /* cjr */
            xlfail( "expression-type mismatch");
        res = cons_atom( tdef, arg.n_ptr ) ;
        break ;

#ifdef COMM
    case TCHAN:
        arg.n_ptr = xljustevarg( &args ) ;
        xllastarg( args ) ;
        if( (arg.n_ptr == NULL) || (arg.n_ptr->n_type != CHAN) ) /* cjr */
            xlfail( "expression-type mismatch");
        res = cons_atom( tdef, arg.n_ptr ) ;  
        break ;
#endif

    case TEQ:
        for( tdef1 = tdef->def.type ;
             (tdef1->tnumber == TEQ) && (tdef1->nnumber != arg.n_ptr->n_type) ;
                                                     tdef1 = tdef1->def.type ) ;
        if( tdef1->tnumber == TTUP )
            if( (args != NULL) && (args->n_listnext != NULL) ) {
                res = construct( tdef1, args ) ;
                res->n_type = tdef->nnumber ; }
            else {
                arg.n_ptr = xljustevarg( &args ) ;
                if( (arg.n_ptr == NULL) ||
                    (arg.n_ptr->n_type != tdef1->nnumber) ) {
                    if( (tdef1->def.tcomp->next != NULL) || 
                            !type_match( arg.n_ptr, tdef1->def.tcomp->type) )
                        xlfail( "expression-type mismatch" ) ;
                    resaux.n_ptr = newnode( tdef->nnumber ) ;
                    resaux.n_ptr->n_symvalue = newnode( LIST ) ;
                    resaux.n_ptr->n_symvalue->n_listvalue = arg.n_ptr ;
                    res = resaux.n_ptr ; }
                else {
                    res = arg.n_ptr ; 
                    res->n_type = tdef->nnumber ; } }
        else {
            arg.n_ptr = xljustevarg( &args ) ;
            xllastarg( args ) ;
            if( (arg.n_ptr == NULL) || (arg.n_ptr->n_type < USERDEFTYPE ) ) {
                if( !type_match( arg.n_ptr, tdef ) ) 
                    xlfail( "expression-type mismatch" ) ;
                res = newnode( tdef->nnumber ) ; 
                res->n_symvalue = arg.n_ptr ; }
            else {
                for( tdef1 = tdef->def.type ;
                     (tdef1->tnumber == TEQ) &&
                     (tdef1->nnumber != arg.n_ptr->n_type) ;
                     tdef1 = tdef1->def.type ) ;
                if( tdef1->nnumber == arg.n_ptr->n_type) {
                    arg.n_ptr->n_type = tdef->nnumber ;
                    res = arg.n_ptr ; }
                else {
                    if( tdef1->tnumber != TALT )
                        xlfail( "expression-type mismatch" ) ;
                    for( succ = FALSE, palt = tdef1->def.acomp ;
                             (palt != NULL) && !succ ; palt = palt->next )
                        if( *(palt->type->tname) != '\0' )
                            succ = (palt->type->nnumber == arg.n_ptr->n_type) ;
                    if( !succ ) xlfail( "expression-type mismatch" ) ;
                    res = newnode( tdef->nnumber ) ; 
                    res->n_symvalue = arg.n_ptr ; }}}
        break ;

    case TSET: case TLIST:
        arg.n_ptr = xljustevarg( &args ) ;
        xllastarg( args ) ;
        if( !type_match( arg.n_ptr, tdef ) )
            xlfail( "expression-type mismatch" ) ;
        res = cons_atom( tdef, arg.n_ptr ) ;  
        break ;

    case TREL: case TFF:
        arg.n_ptr = xljustevarg( &args ) ;
        xllastarg( args ) ;
        if( !type_match( arg.n_ptr, tdef ) )
            xlfail( "expression-type mismatch" ) ;
        res = cons_atom( tdef, arg.n_ptr ) ;  
        break ;

    case TALT:
        arg.n_ptr = xljustevarg( &args ) ;
        xllastarg( args ) ;
        for( succ = FALSE, palt = tdef->def.acomp; 
                            !succ && (palt != NULL); palt = palt->next )
            succ = type_match( arg.n_ptr, palt->type ) ; 
        if( !succ ) xlfail( "expression-type mismatch" ) ;
            res = arg.n_ptr ;
        break ;

    case TTUP:
        last_c = resaux.n_ptr = NULL ;
        ptup = tdef->def.tcomp; 

        do {
            newc = newnode( LIST ) ;
            if( last_c == NULL ) 
                resaux.n_ptr = newc ;
            else
                last_c->n_listnext = newc ;
            last_c = newc ;
            newc->n_listvalue = xljustevarg( &arg.n_ptr ) ;
            if( !type_match( newc->n_listvalue, ptup->type ) )
                xlfail( "expression-type mismatch" ) ;
            ptup = ptup->next ; 
        } while( (ptup != NULL) && (arg.n_ptr != NULL) ) ; 

        if( (ptup != NULL) || (arg.n_ptr != NULL) )
            xlfail( "expression-type mismatch" ) ;
        res = newnode( tdef->nnumber ) ;
        res->n_symvalue = resaux.n_ptr ;
        break ; 

    default:
        xlfail( "expression-type mismatch" ) ; }

    xlstack = oldstk ;

    return( res ) ;
}


/*. xmselect - selects the component labeled "ln" of tuple "args" .*/
struct node *xmselect( ln, args )
char *ln ;
struct node *args ;
{
    struct node *tup, *p ;
    struct t_table *tupdef ;
    struct lab_type *pl ;

    tup = xljustevarg( &args ) ;
    xllastarg( args ) ;
    if(tup == NULL) xlfail ("bad  argument type") ;  /* jjcjr */
    if( tup->n_type < USERDEFTYPE ) xlfail( "bad argument type" ) ;
    tupdef = name_table[tup->n_type - USERDEFTYPE] ;
    while( tupdef->tnumber == TEQ )
         tupdef = tupdef->def.type ;
    if( tupdef->tnumber != TTUP ) xlfail( "bad argument type" ) ;

    for( pl = tupdef->def.tcomp, p = tup->n_symvalue; 
       (pl != NULL) && (p != NULL) && (strcmp( ln, pl->label ) != 0); 
        pl = pl->next, p = p->n_listnext  ) ;

    if( (pl == NULL) || (p == NULL) ) xlfail( "label-tuple mismatch" ) ;

    return( p->n_listvalue ) ;
}


/*. is - builtin function is .*/
/*. is : exptype exp -> bool .*/

/*. bind_undef - binds undefined types to the definitions in"type_table" .*/
static struct t_table *bind_undef( tt )
struct t_table *tt ;
{
    struct node *naux ;

    switch( tt->tnumber ) {
    case TUNDEF :
        naux = xlenter( tt->tname ) ;
        if( ((naux = naux->n_symvalue) == NULL) || (naux->n_type != TYPE))
            xlfail( "type expression containing undefined type" ) ;

        tt->nnumber = naux->n_typedef->nnumber ;
        tt->tnumber = naux->n_typedef->tnumber ;

        switch( tt->tnumber ) {
        case TSET: case TLIST: case TEQ:
            tt->def.type = naux->n_typedef->def.type ;
            break ;
        case TREL: case TFF:
            tt->def.rel.dom = naux->n_typedef->def.rel.dom ;
            tt->def.rel.ran = naux->n_typedef->def.rel.ran ;
            break ;
        case TALT:
            tt->def.acomp = naux->n_typedef->def.acomp ;
            break ;
        case TTUP:
            tt->def.tcomp = naux->n_typedef->def.tcomp ;
            break ; }
        break ;
    case TSET: case TLIST: case TEQ:
        tt->def.type = bind_undef( tt->def.type ) ;
        break ;
    case TREL: case TFF:
        tt->def.rel.dom = bind_undef( tt->def.rel.dom ) ;
        tt->def.rel.ran = bind_undef( tt->def.rel.ran ) ;
        break ; }
  
    return( tt ) ;
}

/*. copy_list - constructs a copy of a symbol list .*/
static struct node *copy_list( sl )
struct node *sl ;
{
    struct node *oldstk, *new, *last ;
    struct node res ;

    oldstk = xlsave( &res, NULL ) ;

    for(res.n_ptr=last=NULL; sl != NULL; sl=sl->n_listnext, last=new){
        new = newnode( LIST ) ;
        if( last == NULL )
            res.n_ptr = new ;
        else 
            last->n_listnext = new ;
        new->n_listvalue = sl->n_listvalue ; }

    xlstack = oldstk ;

    return( res.n_ptr ) ;
}

static struct node *is( args )
struct node *args ;
{
    struct node *type, *exp ;
    struct t_table *tt ;

    /* type = xlmatch( SYM, &args ) ; */
    if((type=recmatch  (SYM, &args ))== NULL) /* jj/jgrenhas */
        type=xlevmatch (SYM, &args ) ;
    if( (type->n_symvalue != NULL) && (type->n_symvalue->n_type == MACTYPE) ) {
        type = copy_list( type->n_symvalue->n_macdef ) ;
        tt = recognize( type->n_listvalue->n_symname, &type->n_listnext ) ; 
        if( tt == NULL ) xlfail( "bad type expression" ) ;
        tt = bind_undef( tt ) ; }
    else
        if( strcmp( type->n_symname, "NIL" ) == 0 )
            tt = p_nil ;
        else if( strcmp( type->n_symname, "ANY" ) == 0 )
            tt = p_any ;
        else {
            tt = recognize( type->n_symname, &args ) ;
            if( tt == NULL ) xlfail( "bad type expression" ) ;
            tt = bind_undef( tt ) ; }
    exp = xljustevarg( &args ) ;
    xllastarg( args ) ;
    return( type_match( exp, tt ) ? true : false ) ;
}
    

/*. mactype - builtin function mactype .*/
/*. mactype : sym exptype -> mac .*/

/*. process_mac - tests if "arg" is a list of symbols .*/
static struct node *process_mac( arg )
struct node *arg ;
{
    struct node *p ;

    if( arg == NULL ) xlfail( "bad macro definition" ) ;
    for( p = arg; p != NULL; xlmatch( SYM, &p ) );

    return( arg ) ;
}

static struct node *mactype( args )
struct node *args ;
{
    struct node arg ;
    struct node *oldstk, *sym ;

    oldstk = xlsave( &arg, NULL ) ;
    arg.n_ptr = args ;
    sym = xlmatch( SYM, &arg.n_ptr ) ;
    if( reserv_symb( sym->n_symname ) )
        xlfail( "may not redefine symbol" ) ;
    sym->n_symvalue = newnode( MACTYPE ) ;
    sym->n_symvalue->n_macdef = process_mac( arg.n_ptr ) ;
    xlstack = oldstk ;

    return( sym ) ;
}


/*. ignlab - builtin function ignlab .*/
/*. ignlab : sexp -> sexp .*/
static struct node *ignlab( args )
struct node *args ;
{
    struct node *expr ;

    expr = xlarg( &args ) ;
    xllastarg( args ) ;

    return( labeval( expr ) ) ;
}


/*. gettype - builtin function gettype .*/
/*. gettype : SYM -> SYM-list .*/


extern struct node *oblimit ;


/*. gettipo - recuperacao das definicoes de tipos .*/

/*. getsym - retorna apontador para um simbolo .*/
struct node *getsym( n )
char *n ;
{
    struct node *p, *ant ;
    int found ;

    for( ant=p=oblimit, found=FALSE ; (p != NULL) && !found ; ant=p,
                                                 p = p->n_listnext )
        found = (strcmp( n, p->n_listvalue->n_symname ) == 0) ;
    if (!found) 
      xlfail("Symbol not found!");
    return( ant->n_listvalue ) ;
}


static void gettipo2(ptr,tp)
struct node **ptr;
struct t_table *tp;
{ struct t_table *ptipo;

  if (!strcmp(tp->tname,""))
    switch (tp->tnumber)
     { case TSYM : (*ptr)->n_listvalue=getsym("SYM");
                   break;
       case TANY : (*ptr)->n_listvalue=getsym("ANY");
                   break;
       case TNIL : (*ptr)->n_listvalue=getsym("NIL");
                   break;
       case TINT : (*ptr)->n_listvalue=getsym("INT");
                   break;
       case TSTR : (*ptr)->n_listvalue=getsym("STR");
                   break;
       case TSET : (*ptr)->n_listvalue=getsym("SET"); 
                   (*ptr)->n_listnext=newnode(LIST);
                   (*ptr)=(*ptr)->n_listnext;
                   gettipo2(ptr,tp->def.type);
                   break;
       case TLIST: (*ptr)->n_listvalue=getsym("LIST"); 
                   (*ptr)->n_listnext=newnode(LIST);
                   (*ptr)=(*ptr)->n_listnext;
                   gettipo2(ptr,tp->def.type);
                   break;
       case TREL : (*ptr)->n_listvalue=getsym("REL"); 
                   (*ptr)->n_listnext=newnode(LIST);
                   (*ptr)=(*ptr)->n_listnext;
                   gettipo2(ptr,tp->def.rel.dom);
                   (*ptr)->n_listnext=newnode(LIST);
                   (*ptr)=(*ptr)->n_listnext;
                   gettipo2(ptr,tp->def.rel.ran);
                   break;
       case TFF  : (*ptr)->n_listvalue=getsym("FF"); 
                   (*ptr)->n_listnext=newnode(LIST);
                   (*ptr)=(*ptr)->n_listnext;
                   gettipo2(ptr,tp->def.rel.dom);
                   (*ptr)->n_listnext=newnode(LIST);
                   (*ptr)=(*ptr)->n_listnext;
                   gettipo2(ptr,tp->def.rel.ran);
                   break;
     }
   else
     { ptipo=type_table;
       while (ptipo->tnumber!=p_nil->tnumber && strcmp(tp->tname,ptipo->tname))
         ptipo=ptipo->next;
       if (ptipo->tnumber==p_nil->tnumber)
         xlfail("Not a defined type!");
       (*ptr)->n_listvalue=newnode(SYM);
       (*ptr)->n_listvalue->n_symname=xmstralloc(strlen(tp->tname)+1);
       strcpy((*ptr)->n_listvalue->n_symname,tp->tname);
       (*ptr)->n_listvalue->n_symvalue=newnode(TYPE);
       (*ptr)->n_listvalue->n_symvalue->n_typedef=ptipo;
     }
}

static struct node *gettipo1(tp)
struct t_table *tp;
{ struct node *nlist, *nsym, *ptr, *ptr1;
  struct type *pa;
  struct lab_type *pt;
  struct t_table *ptipo;

  nlist=newnode(LIST);

  switch (tp->tnumber)
   { case TSYM : nlist->n_listvalue=getsym("SYM");
                 ptr=nlist;
                 break;
     case TANY : nlist->n_listvalue=getsym("ANY");
                 ptr=nlist;
                 break;
     case TNIL : nlist->n_listvalue=getsym("NIL");
                 ptr=nlist;
                 break;
     case TINT : nlist->n_listvalue=getsym("INT");
                 ptr=nlist;
                 break;
     case TSTR : nlist->n_listvalue=getsym("STR");
                 ptr=nlist;
                 break;
     case TSET : nlist->n_listvalue=getsym("SET");
                 ptr=newnode(LIST);
                 nlist->n_listnext=ptr;        
                 gettipo2(&ptr,tp->def.type);
                 break;
     case TLIST: nlist->n_listvalue=getsym("LIST");
                 ptr=newnode(LIST);
                 nlist->n_listnext=ptr;        
                 gettipo2(&ptr,tp->def.type);
                 break;
     case TEQ  : nlist->n_listvalue=getsym("EQ");
                 nlist->n_listnext=newnode(LIST);        
                 ptr=nlist->n_listnext;
                 ptipo=type_table;
                 while (ptipo->tnumber!=p_nil->tnumber && 
                                       strcmp(tp->def.type->tname,ptipo->tname))
                   ptipo=ptipo->next;
                 if (ptipo->tnumber==p_nil->tnumber)
                   xlfail("Not a defined type!");
                 ptr->n_listvalue=newnode(SYM);
                 ptr->n_listvalue->n_symname=xmstralloc(strlen(tp->def.type->tname)+1);
                 strcpy(ptr->n_listvalue->n_symname,tp->def.type->tname);
                 ptr->n_listvalue->n_symvalue=newnode(TYPE);
                 ptr->n_listvalue->n_symvalue->n_typedef=ptipo;
                 break;
     case TREL : nlist->n_listvalue=getsym("REL");
                 ptr=newnode(LIST);
                 nlist->n_listnext=ptr;
                 gettipo2(&ptr,tp->def.rel.dom);
                 ptr->n_listnext=newnode(LIST);
                 ptr=ptr->n_listnext;
                 gettipo2(&ptr,tp->def.rel.ran);
                 break;
     case TFF  : nlist->n_listvalue=getsym("FF");
                 ptr=newnode(LIST);
                 nlist->n_listnext=ptr;
                 gettipo2(&ptr,tp->def.rel.dom);
                 ptr->n_listnext=newnode(LIST);
                 ptr=ptr->n_listnext;
                 gettipo2(&ptr,tp->def.rel.ran);
                 break;
     case TALT : nlist->n_listvalue=getsym("ALT");
                 ptr=nlist;
                 for(pa=tp->def.acomp;pa!=NULL;pa=pa->next)
                  { ptr->n_listnext=newnode(LIST);
                    ptr=ptr->n_listnext;
                    gettipo2(&ptr,pa->type);
                  }
                 break;
     case TTUP : nlist->n_listvalue=getsym("TUP");
                 ptr=nlist;
                 for(pt=tp->def.tcomp;pt!=NULL;pt=pt->next)
                  { ptr->n_listnext=newnode(LIST);
                    ptr=ptr->n_listnext;
                    ptr->n_listvalue=newnode(LIST);
                    ptr->n_listvalue->n_listvalue=newnode(SYM);
                    ptr->n_listvalue->n_listvalue->n_symname=xmstralloc(strlen(pt->label)+1);
                    strcpy(ptr->n_listvalue->n_listvalue->n_symname,pt->label);
                    ptr->n_listvalue->n_listvalue->n_symvalue=newnode(LABEL);
                    ptr->n_listvalue->n_listvalue->n_symvalue->n_symname=xmstralloc(strlen(pt->label)+1);
                    strcpy(ptr->n_listvalue->n_listvalue->n_symvalue->n_symname,pt->label);
                    ptr1=ptr->n_listvalue->n_listnext=newnode(LIST);
                    gettipo2(&ptr1,pt->type);
                    ptr1->n_listnext=NULL;
                  }
                 break;
   }
  ptr->n_listnext=NULL;
  return(nlist);
}


struct node *gettipo( args )
struct node *args ;
{ struct node arg, arg1 ;
  struct node *oldstk, *val ;
  struct t_table *ptipo;
  int tp;

  /* create a new stack frame */
  oldstk = xlsave( &arg, &arg1, NULL ) ;

  /* initialize */
  arg.n_ptr = args ;

  arg1.n_ptr=xlevarg(&arg.n_ptr);

  if( (arg1.n_ptr == NULL)||(((tp = arg1.n_ptr->n_type) != SYM) &&(tp != TYPE)))
        xlfail( "bad argument type" ) ;  

  if(tp == SYM){
     ptipo=type_table;
     while (ptipo->tnumber!=p_nil->tnumber && 
                                    strcmp(arg1.n_ptr->n_symname,ptipo->tname))
        ptipo=ptipo->next;
  }
  else ptipo = arg1.n_ptr-> n_typedef;

  if (ptipo->tnumber!=p_nil->tnumber)
    val=gettipo1(ptipo);
   else
    xlfail("Not a defined type!");

  /* restore the previous stack frame */
  xlstack = oldstk ;

  return(val);
}


/* Alteracao de um elemento de um Tuplo (jcc) */
struct node *tplus( args )
struct node *args ;
{ struct node arg, *arg1, *arg2, *arg3;
  struct node *oldstk, *val ;
  struct t_table *ptipo;
  struct lab_type *ptrlab;
  struct node *ptrvalues;

  /* create a new stack frame */
  oldstk = xlsave( &arg, arg1, arg2, arg3, NULL ) ;

  /* initialize */
  arg.n_ptr = args ;

  arg1=xlevarg(&arg.n_ptr);
  arg2=xlevmatch(LABEL, &arg.n_ptr);
  arg3=xlevarg(&arg.n_ptr);
  /*xllastarg(&arg.n_ptr);*/

  if ((arg1 == NULL) || (arg1->n_type < USERDEFTYPE))
    xlfail("bad argument");
 
  ptipo=type_table;
  while (ptipo->tnumber!=p_nil->tnumber && ptipo->nnumber != arg1->n_type)
    ptipo=ptipo->next;
  if (ptipo->tnumber != TTUP)
    xlfail("bad type");
  val = newnode(arg1->n_type);

  for (ptrlab = ptipo->def.tcomp, 
         ptrvalues = val->n_symvalue=copy_list(arg1->n_symvalue);
       ptrlab != NULL && ptrlab->label != arg2->n_symname; 
       ptrlab = ptrlab->next)
    ptrvalues = ptrvalues->n_listnext;
  if (ptrlab == NULL)
    xlfail("bad selector");

  if (!type_match(arg3, ptrlab->type)) xlfail("bad new value"); 

  ptrvalues->n_listvalue = arg3;

  /* restore the previous stack frame */

  xlstack = oldstk ;

  return(val);
}


/* Alteracao de um elemento de um Tuplo (jcc) */
struct node *deffield( args )
struct node *args ;
{ struct node arg, *arg1, *arg2, *arg3;
  struct node *oldstk, *val ;
  struct t_table *ptipo;
  struct lab_type *ptrlab;
  struct node *ptrvalues;

  /* create a new stack frame */
  oldstk = xlsave( &arg, arg1, arg2, arg3, NULL ) ;

  /* initialize */
  arg.n_ptr = args ;

  arg1=xlevarg(&arg.n_ptr);
  arg2=xlevmatch(LABEL, &arg.n_ptr);
  arg3=xlevarg(&arg.n_ptr);
  /*xllastarg(&arg.n_ptr);*/

  if ((arg1 == NULL) || (arg1->n_type < USERDEFTYPE))
    xlfail("bad argument");
 
  ptipo=type_table;
  while (ptipo->tnumber!=p_nil->tnumber && ptipo->nnumber != arg1->n_type)
    ptipo=ptipo->next;
  if (ptipo->tnumber != TTUP)
    xlfail("bad type");

  for (ptrlab = ptipo->def.tcomp, ptrvalues = arg1->n_symvalue;
       ptrlab != NULL && ptrlab->label != arg2->n_symname; 
       ptrlab = ptrlab->next)
    ptrvalues = ptrvalues->n_listnext;
  if (ptrlab == NULL)
    xlfail("bad selector");

  if (!type_match(arg3, ptrlab->type)) xlfail("bad new value"); 

  ptrvalues->n_listvalue = arg3;

  /* restore the previous stack frame */

  xlstack = oldstk ;

  return(arg1);
}

#define xltype(p,na,nu,nx)\
    {if( (p = (struct t_table *)malloc( sizeof( struct t_table ) ))\
         == NULL ) xlfail( "out of heap space" ) ;\
    p->tname = na ; p->tnumber = nu ; p->next = nx;}

/*. xmutypeinit - initialize user type module .*/
void xmutypeinit()
{
    /* initialize null_str */
    null_str = "" ;

    /* insert atomic types into type_table */

   xltype(p_any,null_str,TANY,NULL)
   xltype(p_int,null_str,TINT,p_any)
   xltype(p_str,null_str,TSTR,p_int)
   xltype(p_fptr,null_str,TFPTR,p_str)
   xltype(p_sym,null_str,TSYM,p_fptr)

#ifdef COMM
   xltype(p_chan,null_str,TCHAN,p_sym)
#endif
   xltype(p_nil,null_str,TNIL,p_sym)

#ifdef COMM
    p_nil->next = p_chan ;
#endif

    type_table = p_nil ;
    name_number = USERDEFTYPE ;

    xlenter( "TUP" ) ;
    xlenter( "ALT" ) ;
    xlenter( "EQ" ) ;
    NIL = xlenter( "NIL" ) ;
    NIL->n_symvalue = NIL ;
    ANY = xlenter( "ANY" ) ;
    ANY->n_symvalue = ANY ;

    /* shell functions */
    xlsubr( "deftype", deftype ) ;
    xlsubr( "is", is ) ;
    xlsubr( "mactype", mactype ) ;
    xlsubr( "ignlab",ignlab ) ;
    xlsubr( "gettype",gettipo);
    xlsubr( "deffield",deffield);
    xlsubr( "tplus",tplus);
}

#endif

